@@ -2,7 +2,7 @@
     Title:     Write out a database as an ELF object file
     Author:    David Matthews.
 
-    Copyright (c) 2006-7, 2011, 2016 David C. J. Matthews
+    Copyright (c) 2006-7, 2011, 2016-18 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -241,6 +241,33 @@
 # define HOST_DIRECT_DATA_RELOC R_ALPHA_REFQUAD
 # define HOST_DIRECT_FPTR_RELOC R_ALPHA_REFQUAD
 # define USE_RELA 1
+#elif defined(HOSTARCHITECTURE_RISCV32) || defined(HOSTARCHITECTURE_RISCV64)
+# define HOST_E_MACHINE EM_RISCV
+# if defined(HOSTARCHITECTURE_RISCV32)
+#  define HOST_DIRECT_DATA_RELOC R_RISCV_32
+#  define HOST_DIRECT_FPTR_RELOC R_RISCV_32
+# else
+#  define HOST_DIRECT_DATA_RELOC R_RISCV_64
+#  define HOST_DIRECT_FPTR_RELOC R_RISCV_64
+# endif
+# if defined(__riscv_float_abi_soft)
+#  define HOST_E_FLAGS_FLOAT_ABI EF_RISCV_FLOAT_ABI_SOFT
+# elif defined(__riscv_float_abi_single)
+#  define HOST_E_FLAGS_FLOAT_ABI EF_RISCV_FLOAT_ABI_SINGLE
+# elif defined(__riscv_float_abi_double)
+#  define HOST_E_FLAGS_FLOAT_ABI EF_RISCV_FLOAT_ABI_DOUBLE
+# elif defined(__riscv_float_abi_quad)
+#  define HOST_E_FLAGS_FLOAT_ABI EF_RISCV_FLOAT_ABI_QUAD
+# else
+#  error "Unknown RISC-V float ABI"
+# endif
+# ifdef __riscv_32e
+#  define HOST_E_FLAGS_RVE __riscv_32e
+# else
+#  define HOST_E_FLAGS_RVE 0
+# endif
+# define HOST_E_FLAGS (HOST_E_FLAGS_FLOAT_ABI | HOST_E_FLAGS_RVE)
+# define USE_RELA 1
 #else
 # error "No support for exporting on this architecture"
 #endif
@@ -262,18 +289,18 @@ enum {
 };
 
 // Add an external reference to the RTS
-void ELFExport::addExternalReference(void *relocAddr, const char *name)
+void ELFExport::addExternalReference(void *relocAddr, const char *name, bool isFuncPtr)
 {
     externTable.makeEntry(name);
     // The symbol is added after the memory table entries and poly_exports
-    writeRelocation(0, relocAddr, symbolNum++, true);
+    writeRelocation(0, relocAddr, symbolNum++, isFuncPtr);
 }
 
 // Generate the address relative to the start of the segment.
 void ELFExport::setRelocationAddress(void *p, ElfXX_Addr *reloc)
 {
     unsigned area = findArea(p);
-    POLYUNSIGNED offset = (char*)p - (char*)memTable[area].mtAddr;
+    POLYUNSIGNED offset = (char*)p - (char*)memTable[area].mtOriginalAddr;
     *reloc = offset;
 }
 
@@ -282,7 +309,7 @@ PolyWord ELFExport::createRelocation(PolyWord p, void *relocAddr)
 {
     void *addr = p.AsAddress();
     unsigned addrArea = findArea(addr);
-    POLYUNSIGNED offset = (char*)addr - (char*)memTable[addrArea].mtAddr;
+    POLYUNSIGNED offset = (char*)addr - (char*)memTable[addrArea].mtOriginalAddr;
     return writeRelocation(offset, relocAddr, AreaToSym(addrArea), false);
 }
 
@@ -316,12 +343,13 @@ PolyWord ELFExport::writeRelocation(POLYUNSIGNED offset, void *relocAddr, unsign
    that the offset is saved in original word. */
 void ELFExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind code)
 {
-    PolyWord p = GetConstantValue(addr, code);
+#ifndef POLYML32IN64
+    PolyObject *p = GetConstantValue(addr, code);
 
-    if (IS_INT(p) || p == PolyWord::FromUnsigned(0))
+    if (p == 0)
         return;
 
-    void *a = p.AsAddress();
+    void *a = p;
     unsigned aArea = findArea(a);
 
     // We don't need a relocation if this is relative to the current segment
@@ -330,7 +358,7 @@ void ELFExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind co
         return;
 
     // Set the value at the address to the offset relative to the symbol.
-    POLYUNSIGNED offset = (char*)a - (char*)memTable[aArea].mtAddr;
+    POLYUNSIGNED offset = (char*)a - (char*)memTable[aArea].mtOriginalAddr;
 
     switch (code)
     {
@@ -383,6 +411,7 @@ void ELFExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind co
         default:
             ASSERT(0); // Wrong type of relocation for this architecture.
     }
+#endif
 }
 
 unsigned long ELFExport::makeStringTableEntry(const char *str, ExportStringTable *stab)
@@ -415,7 +444,7 @@ void ELFExport::alignFile(int align)
     fwrite(&pad, align - (offset % align), 1, exportFile);
 }
 
-void ELFExport::createStructsRelocation(unsigned sym, POLYUNSIGNED offset, POLYSIGNED addend)
+void ELFExport::createStructsRelocation(unsigned sym, size_t offset, size_t addend)
 {
 #if USE_RELA
     ElfXX_Rela reloc;
@@ -596,7 +625,7 @@ void ELFExport::exportStore(void)
         sections[relocSection].sh_offset = ftell(exportFile);
         relocationCount = 0;
         // Create the relocation table and turn all addresses into offsets.
-        char *start = (char*)memTable[i].mtAddr;
+        char *start = (char*)memTable[i].mtOriginalAddr;
         char *end = start + memTable[i].mtLength;
         for (p = (PolyWord*)start; p < (PolyWord*)end; )
         {
@@ -625,19 +654,19 @@ void ELFExport::exportStore(void)
 
     // Address of "memTable" within "exports". We can't use createRelocation because
     // the position of the relocation is not in either the mutable or the immutable area.
-    POLYSIGNED memTableOffset = (POLYSIGNED)sizeof(exportDescription); // It follows immediately after this.
+    size_t memTableOffset = sizeof(exportDescription); // It follows immediately after this.
     createStructsRelocation(AreaToSym(memTableEntries), offsetof(exportDescription, memTable), memTableOffset);
 
     // Address of "rootFunction" within "exports"
     unsigned rootAddrArea = findArea(rootFunction);
-    POLYSIGNED rootOffset = (char*)rootFunction - (char*)memTable[rootAddrArea].mtAddr;
+    size_t rootOffset = (char*)rootFunction - (char*)memTable[rootAddrArea].mtOriginalAddr;
     createStructsRelocation(AreaToSym(rootAddrArea), offsetof(exportDescription, rootFunction), rootOffset);
 
     // Addresses of the areas within memtable.
     for (i = 0; i < memTableEntries; i++)
     {
         createStructsRelocation(AreaToSym(i),
-            sizeof(exportDescription) + i * sizeof(memoryTableEntry) + offsetof(memoryTableEntry, mtAddr),
+            sizeof(exportDescription) + i * sizeof(memoryTableEntry) + offsetof(memoryTableEntry, mtCurrentAddr),
             0 /* No offset relative to base symbol*/);
     }
 
@@ -679,7 +708,7 @@ void ELFExport::exportStore(void)
         sections[dataSection].sh_size = memTable[i].mtLength;
         alignFile(sections[dataSection].sh_addralign);
         sections[dataSection].sh_offset = ftell(exportFile);
-        fwrite(memTable[i].mtAddr, 1, memTable[i].mtLength, exportFile);
+        fwrite(memTable[i].mtOriginalAddr, 1, memTable[i].mtLength, exportFile);
     }
 
     exportDescription exports;
@@ -694,11 +723,16 @@ void ELFExport::exportStore(void)
     exports.timeStamp = getBuildTime();
     exports.architecture = machineDependent->MachineArchitecture();
     exports.rtsVersion = POLY_version_number;
+#ifdef POLYML32IN64
+    exports.originalBaseAddr = globalHeapBase;
+#else
+    exports.originalBaseAddr = 0;
+#endif
 
     // Set the address values to zero before we write.  They will always
     // be relative to their base symbol.
     for (i = 0; i < memTableEntries; i++)
-        memTable[i].mtAddr = 0;
+        memTable[i].mtCurrentAddr = 0;
 
     // Now the binary data.
     alignFile(sections[sect_table_data].sh_addralign);
