@@ -1,7 +1,7 @@
 /*
     Title:  exporter.cpp - Export a function as an object or C file
 
-    Copyright (c) 2006-7, 2015, 2016-17 David C.J. Matthews
+    Copyright (c) 2006-7, 2015, 2016-19 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -50,7 +50,7 @@
 #include <stdlib.h>
 #endif
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #include <tchar.h>
 #else
 #define _T(x) x
@@ -83,7 +83,7 @@
 #include "machoexport.h"
 #endif
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #define NOMEMORY ERROR_NOT_ENOUGH_MEMORY
 #define ERRORNUMBER _doserrno
 #else
@@ -110,6 +110,26 @@ GraveYard::~GraveYard()
     free(graves);
 }
 
+// Used to calculate the space required for the ordinary mutables
+// and the no-overwrite mutables.  They are interspersed in local space.
+class MutSizes : public ScanAddress
+{
+public:
+    MutSizes() : mutSize(0), noOverSize(0) {}
+
+    virtual PolyObject *ScanObjectAddress(PolyObject *base) { return base; }// No Actually used
+
+    virtual void ScanAddressesInObject(PolyObject *base, POLYUNSIGNED lengthWord)
+    {
+        const POLYUNSIGNED words = OBJ_OBJECT_LENGTH(lengthWord) + 1; // Include length word
+        if (OBJ_IS_NO_OVERWRITE(lengthWord))
+            noOverSize += words;
+        else mutSize += words;
+    }
+
+    POLYUNSIGNED mutSize, noOverSize;
+};
+
 CopyScan::CopyScan(unsigned h/*=0*/): hierarchy(h)
 {
     defaultImmSize = defaultMutSize = defaultCodeSize = defaultNoOverSize = 0;
@@ -143,7 +163,7 @@ void CopyScan::initialise(bool isExport/*=true*/)
         if (space->hierarchy >= hierarchy) {
             // Include this if we're exporting (hierarchy=0) or if we're saving a state
             // and will include this in the new state.
-            POLYUNSIGNED size = (space->top-space->bottom)/4;
+            size_t size = (space->top-space->bottom)/4;
             if (space->noOverwrite)
                 defaultNoOverSize += size;
             else if (space->isMutable)
@@ -175,18 +195,24 @@ void CopyScan::initialise(bool isExport/*=true*/)
     for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
         LocalMemSpace *space = *i;
-        POLYUNSIGNED size = space->allocatedSpace();
+        uintptr_t size = space->allocatedSpace();
         // It looks as though the mutable size generally gets
         // overestimated while the immutable size is correct.
         if (space->isMutable)
-            defaultMutSize += size/4;
+        {
+            MutSizes sizeMut;
+            sizeMut.ScanAddressesInRegion(space->bottom, space->lowerAllocPtr);
+            sizeMut.ScanAddressesInRegion(space->upperAllocPtr, space->top);
+            defaultNoOverSize += sizeMut.noOverSize / 4;
+            defaultMutSize += sizeMut.mutSize / 4;
+        }
         else
             defaultImmSize += size/2;
     }
     for (std::vector<CodeSpace *>::iterator i = gMem.cSpaces.begin(); i < gMem.cSpaces.end(); i++)
     {
         CodeSpace *space = *i;
-        POLYUNSIGNED size = space->spaceSize();
+        uintptr_t size = space->spaceSize();
         defaultCodeSize += size/2;
     }
     if (isExport)
@@ -195,13 +221,22 @@ void CopyScan::initialise(bool isExport/*=true*/)
         if (defaultMutSize < 1024*1024) defaultMutSize = 1024*1024;
         if (defaultImmSize < 1024*1024) defaultImmSize = 1024*1024;
         if (defaultCodeSize < 1024*1024) defaultCodeSize = 1024*1024;
+#ifdef MACOSX
+        // Limit the segment size for Mac OS X.  The linker has a limit of 2^24 relocations
+        // in a segment so this is a crude way of ensuring the limit isn't exceeded.
+        // It's unlikely to be exceeded by the code itself.
+        // Actually, from trial-and-error, the limit seems to be around 6M.
+        if (defaultMutSize > 6 * 1024 * 1024) defaultMutSize = 6 * 1024 * 1024;
+        if (defaultImmSize > 6 * 1024 * 1024) defaultImmSize = 6 * 1024 * 1024;
+#endif
+        if (defaultNoOverSize < 4096) defaultNoOverSize = 4096; // Except for the no-overwrite area
     }
     else
     {
         // Much smaller minimum sizes for saved states.
         if (defaultMutSize < 1024) defaultMutSize = 1024;
         if (defaultImmSize < 4096) defaultImmSize = 4096;
-        if (defaultCodeSize < 4096) defaultImmSize = 4096;
+        if (defaultCodeSize < 4096) defaultCodeSize = 4096;
         if (defaultNoOverSize < 4096) defaultNoOverSize = 4096;
         // Set maximum sizes as well.  We may have insufficient contiguous space for
         // very large areas.
@@ -211,8 +246,8 @@ void CopyScan::initialise(bool isExport/*=true*/)
         if (defaultNoOverSize > 1024 * 1024) defaultNoOverSize = 1024 * 1024;
     }
     if (debugOptions & DEBUG_SAVING)
-        Log("SAVE: Copyscan default sizes: Immutable: %lu, Mutable: %lu, Code: %lu.\n",
-            defaultImmSize, defaultMutSize, defaultCodeSize);
+        Log("SAVE: Copyscan default sizes: Immutable: %" POLYUFMT ", Mutable: %" POLYUFMT ", Code: %" POLYUFMT ", No-overwrite %" POLYUFMT ".\n",
+            defaultImmSize, defaultMutSize, defaultCodeSize, defaultNoOverSize);
 }
 
 CopyScan::~CopyScan()
@@ -232,9 +267,20 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
     // Ignore integers.
     if (IS_INT(val) || val == PolyWord::FromUnsigned(0))
         return 0;
-    // Ignore pointers to the IO area.  They will be relocated
-    // when we write out the memory
-    MemSpace *space = gMem.SpaceForAddress(val.AsStackAddr()-1);
+
+    PolyObject *obj = val.AsObjPtr();
+    POLYUNSIGNED l = ScanAddress(&obj);
+    *pt = obj;
+    return l;
+}
+
+// This function is called for each address in an object
+// once it has been copied to its new location.  We copy first
+// then scan to update the addresses.
+POLYUNSIGNED CopyScan::ScanAddress(PolyObject **pt)
+{
+    PolyObject *obj = *pt;
+    MemSpace *space = gMem.SpaceForAddress((PolyWord*)obj - 1);
     ASSERT(space != 0);
     // We may sometimes get addresses that have already been updated
     // to point to the new area.  e.g. (only?) in the case of constants
@@ -251,14 +297,18 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
             return 0;
     }
 
-    ASSERT(OBJ_IS_DATAPTR(val));
-
     // Have we already scanned this?
-    PolyObject *obj = val.AsObjPtr();
     if (obj->ContainsForwardingPtr())
     {
         // Update the address to the new value.
+#ifdef POLYML32IN64
+        PolyObject *newAddr;
+        if (space->isCode)
+            newAddr = (PolyObject*)(globalCodeBase + ((obj->LengthWord() & ~_OBJ_TOMBSTONE_BIT) << 1));
+        else newAddr = obj->GetForwardingPtr();
+#else
         PolyObject *newAddr = obj->GetForwardingPtr();
+#endif
         *pt = newAddr;
         return 0; // No need to scan it again.
     }
@@ -268,13 +318,21 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
         for (unsigned i = 0; i < tombs; i++)
         {
             GraveYard *g = &graveYard[i];
-            if (val.AsStackAddr() >= g->startAddr && val.AsStackAddr() < g->endAddr)
+            if ((PolyWord*)obj >= g->startAddr && (PolyWord*)obj < g->endAddr)
             {
-                PolyWord *tombAddr = g->graves + (val.AsStackAddr() - g->startAddr);
+                PolyWord *tombAddr = g->graves + ((PolyWord*)obj - g->startAddr);
                 PolyObject *tombObject = (PolyObject*)tombAddr;
                 if (tombObject->ContainsForwardingPtr())
                 {
-                    *pt = tombObject->GetForwardingPtr();;
+#ifdef POLYML32IN64
+                    PolyObject *newAddr;
+                    if (space->isCode)
+                        newAddr = (PolyObject*)(globalCodeBase + ((tombObject->LengthWord() & ~_OBJ_TOMBSTONE_BIT) << 1));
+                    else newAddr = tombObject->GetForwardingPtr();
+#else
+                    PolyObject *newAddr = tombObject->GetForwardingPtr();
+#endif
+                    *pt = newAddr;
                     return 0;
                 }
                 break; // No need to look further
@@ -284,7 +342,7 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
 
     // No, we need to copy it.
     ASSERT(space->spaceType == ST_LOCAL || space->spaceType == ST_PERMANENT ||
-           space->spaceType == ST_CODE);
+        space->spaceType == ST_CODE);
     POLYUNSIGNED lengthWord = obj->LengthWord();
     POLYUNSIGNED words = OBJ_OBJECT_LENGTH(lengthWord);
 
@@ -309,11 +367,19 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
             isCodeObj == space->isCode)
         {
             ASSERT(space->topPointer <= space->top && space->topPointer >= space->bottom);
-            POLYUNSIGNED spaceLeft = space->top - space->topPointer;
+            size_t spaceLeft = space->top - space->topPointer;
             if (spaceLeft > words)
             {
-                newObj = (PolyObject*)(space->topPointer+1);
-                space->topPointer += words+1;
+                newObj = (PolyObject*)(space->topPointer + 1);
+                space->topPointer += words + 1;
+#ifdef POLYML32IN64
+                // Maintain the odd-word alignment of topPointer
+                if ((words & 1) == 0 && space->topPointer < space->top)
+                {
+                    *space->topPointer = PolyWord::FromUnsigned(0);
+                    space->topPointer++;
+                }
+#endif
                 break;
             }
         }
@@ -321,7 +387,7 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
     if (newObj == 0)
     {
         // Didn't find room in the existing spaces.  Create a new space.
-        POLYUNSIGNED spaceWords;
+        uintptr_t spaceWords;
         if (isMutableObj)
         {
             if (isNoOverwrite) spaceWords = defaultNoOverSize;
@@ -333,7 +399,7 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
             else spaceWords = defaultImmSize;
         }
         if (spaceWords <= words)
-            spaceWords = words+1; // Make sure there's space for this object.
+            spaceWords = words + 1; // Make sure there's space for this object.
         PermanentMemSpace *space = gMem.NewExportSpace(spaceWords, isMutableObj, isNoOverwrite, isCodeObj);
         if (isByteObj) space->byteOnly = true;
         if (space == 0)
@@ -343,14 +409,22 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
             // Unable to allocate this.
             throw MemoryException();
         }
-        newObj = (PolyObject*)(space->topPointer+1);
-        space->topPointer += words+1;
+        newObj = (PolyObject*)(space->topPointer + 1);
+        space->topPointer += words + 1;
+#ifdef POLYML32IN64
+        // Maintain the odd-word alignment of topPointer
+        if ((words & 1) == 0 && space->topPointer < space->top)
+        {
+            *space->topPointer = PolyWord::FromUnsigned(0);
+            space->topPointer++;
+        }
+#endif
         ASSERT(space->topPointer <= space->top && space->topPointer >= space->bottom);
     }
 
     newObj->SetLengthWord(lengthWord); // copy length word
 
-    memcpy(newObj, obj, words*sizeof(PolyWord));
+    memcpy(newObj, obj, words * sizeof(PolyWord));
 
     if (space->spaceType == ST_PERMANENT && !space->isMutable && ((PermanentMemSpace*)space)->hierarchy == 0)
     {
@@ -359,16 +433,34 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
         for (m = 0; m < tombs; m++)
         {
             GraveYard *g = &graveYard[m];
-            if (val.AsStackAddr() >= g->startAddr && val.AsStackAddr() < g->endAddr)
+            if ((PolyWord*)obj >= g->startAddr && (PolyWord*)obj < g->endAddr)
             {
-                PolyWord *tombAddr = g->graves + (val.AsStackAddr() - g->startAddr);
+                PolyWord *tombAddr = g->graves + ((PolyWord*)obj - g->startAddr);
                 PolyObject *tombObject = (PolyObject*)tombAddr;
+#ifdef POLYML32IN64
+                if (isCodeObj)
+                {
+                    POLYUNSIGNED ll = (POLYUNSIGNED)(((PolyWord*)newObj - globalCodeBase) >> 1 | _OBJ_TOMBSTONE_BIT);
+                    tombObject->SetLengthWord(ll);
+                }
+                else tombObject->SetForwardingPtr(newObj);
+#else
                 tombObject->SetForwardingPtr(newObj);
+#endif
                 break; // No need to look further
             }
         }
         ASSERT(m < tombs); // Should be there.
     }
+#ifdef POLYML32IN64
+    // If this is a code address we can't use the usual forwarding pointer format.
+    // Instead we have to compute the offset relative to the base of the code.
+    else if (isCodeObj)
+    {
+        POLYUNSIGNED ll = (POLYUNSIGNED)(((PolyWord*)newObj-globalCodeBase) >> 1 | _OBJ_TOMBSTONE_BIT);
+        obj->SetLengthWord(ll);
+    }
+#endif
     else obj->SetForwardingPtr(newObj); // Put forwarding pointer in old object.
 
     if (OBJ_IS_CODE_OBJECT(lengthWord))
@@ -385,6 +477,15 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
     return lengthWord;  // This new object needs to be scanned.
 }
 
+// The address of code in the code area.  We treat this as a normal heap cell.
+// We will probably need to copy this and to process addresses within it.
+POLYUNSIGNED CopyScan::ScanCodeAddressAt(PolyObject **pt)
+{
+    POLYUNSIGNED lengthWord = ScanAddress(pt);
+    if (lengthWord)
+        ScanAddressesInObject(*pt, lengthWord);
+    return 0;
+}
 
 PolyObject *CopyScan::ScanObjectAddress(PolyObject *base)
 {
@@ -411,9 +512,19 @@ static POLYUNSIGNED GetObjLength(PolyObject *obj)
 {
     if (obj->ContainsForwardingPtr())
     {
-        PolyObject *forwardedTo = obj->GetForwardingPtr();
+        PolyObject *forwardedTo;
+#ifdef POLYML32IN64
+        {
+            MemSpace *space = gMem.SpaceForAddress((PolyWord*)obj - 1);
+            if (space->isCode)
+                forwardedTo = (PolyObject*)(globalCodeBase + ((obj->LengthWord() & ~_OBJ_TOMBSTONE_BIT) << 1));
+            else forwardedTo = obj->GetForwardingPtr();
+        }
+#else
+        forwardedTo = obj->GetForwardingPtr();
+#endif
         POLYUNSIGNED length = GetObjLength(forwardedTo);
-        MemSpace *space = gMem.SpaceForAddress(forwardedTo-1);
+        MemSpace *space = gMem.SpaceForAddress((PolyWord*)forwardedTo-1);
         if (space->spaceType == ST_EXPORT)
             obj->SetLengthWord(length);
         return length;
@@ -424,13 +535,21 @@ static POLYUNSIGNED GetObjLength(PolyObject *obj)
     }
 }
 
-static void FixForwarding(PolyWord *pt, POLYUNSIGNED space)
+static void FixForwarding(PolyWord *pt, size_t space)
 {
     while (space)
     {
         pt++;
         PolyObject *obj = (PolyObject*)pt;
-        POLYUNSIGNED length = OBJ_OBJECT_LENGTH(GetObjLength(obj));
+#ifdef POLYML32IN64
+        if ((uintptr_t)obj & 4)
+        {
+            // Skip filler words needed to align to an even word
+            space--;
+            continue; // We've added 1 to pt so just loop.
+        }
+#endif
+        size_t length = OBJ_OBJECT_LENGTH(GetObjLength(obj));
         pt += length;
         ASSERT(space > length);
         space -= length+1;
@@ -539,7 +658,7 @@ void Exporter::RunExport(PolyObject *rootFunction)
             if (space->hierarchy < hierarchy)
             {
                 memoryTableEntry *entry = &exports->memTable[memEntry++];
-                entry->mtAddr = space->bottom;
+                entry->mtOriginalAddr = entry->mtCurrentAddr = space->bottom;
                 entry->mtLength = (space->topPointer-space->bottom)*sizeof(PolyWord);
                 entry->mtIndex = space->index;
                 entry->mtFlags = 0;
@@ -554,7 +673,7 @@ void Exporter::RunExport(PolyObject *rootFunction)
     {
         memoryTableEntry *entry = &exports->memTable[memEntry++];
         PermanentMemSpace *space = *i;
-        entry->mtAddr = space->bottom;
+        entry->mtOriginalAddr = entry->mtCurrentAddr = space->bottom;
         entry->mtLength = (space->topPointer-space->bottom)*sizeof(PolyWord);
         entry->mtIndex = hierarchy == 0 ? memEntry-1 : space->index;
         entry->mtFlags = 0;
@@ -570,8 +689,13 @@ void Exporter::RunExport(PolyObject *rootFunction)
     ASSERT(memEntry == tableEntries);
     exports->memTableEntries = memEntry;
     exports->rootFunction = copiedRoot;
-    exports->exportStore();
-    return;
+    try {
+        // This can raise MemoryException at least in PExport::exportStore. 
+        exports->exportStore();
+    }
+    catch (MemoryException &) {
+        exports->errorMessage = "Insufficient Memory";
+    }
 }
 
 // Functions called via the RTS call.
@@ -579,7 +703,7 @@ Handle exportNative(TaskData *taskData, Handle args)
 {
 #ifdef HAVE_PECOFF
     // Windows including Cygwin
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     const TCHAR *extension = _T(".obj"); // Windows
 #else
     const char *extension = ".o"; // Cygwin
@@ -625,7 +749,7 @@ POLYUNSIGNED PolyExport(PolyObject *threadId, PolyWord fileName, PolyWord root)
     try {
 #ifdef HAVE_PECOFF
         // Windows including Cygwin
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
         const TCHAR *extension = _T(".obj"); // Windows
 #else
         const char *extension = ".o"; // Cygwin
@@ -687,9 +811,11 @@ Exporter::~Exporter()
 
 void Exporter::relocateValue(PolyWord *pt)
 {
+#ifndef POLYML32IN64
     PolyWord q = *pt;
     if (IS_INT(q) || q == PolyWord::FromUnsigned(0)) {}
     else createRelocation(pt);
+#endif
 }
 
 // Check through the areas to see where the address is.  It must be
@@ -698,8 +824,8 @@ unsigned Exporter::findArea(void *p)
 {
     for (unsigned i = 0; i < memTableEntries; i++)
     {
-        if (p > memTable[i].mtAddr &&
-            p <= (char*)memTable[i].mtAddr + memTable[i].mtLength)
+        if (p > memTable[i].mtOriginalAddr &&
+            p <= (char*)memTable[i].mtOriginalAddr + memTable[i].mtLength)
             return i;
     }
     { ASSERT(0); }
@@ -714,11 +840,12 @@ void Exporter::relocateObject(PolyObject *p)
         {
             // Weak mutable byte refs are used for external references and
             // also in the FFI for non-persistent values.
-            const char *entryName = getEntryPointName(p);
-            if (entryName != 0) addExternalReference(p, entryName);
+            bool isFuncPtr = true;
+            const char *entryName = getEntryPointName(p, &isFuncPtr);
+            if (entryName != 0) addExternalReference(p, entryName, isFuncPtr);
             // Clear the first word of the data.
-            ASSERT(p->Length() > 0);
-            p->Set(0, PolyWord::FromSigned(0));
+            ASSERT(p->Length() >= sizeof(uintptr_t)/sizeof(PolyWord));
+            *(uintptr_t*)p = 0;
         }
     }
     else if (p->IsCodeObject())
@@ -731,6 +858,13 @@ void Exporter::relocateObject(PolyObject *p)
         for (POLYUNSIGNED i = 0; i < constCount; i++) relocateValue(&(cp[i]));
 
     }
+    else if (p->IsClosureObject())
+    {
+#ifndef POLYML32IN64
+        ASSERT(0);
+#endif
+        // This should only be used in 32-in-64 where we don't use relocations.
+    }
     else /* Ordinary objects, essentially tuples. */
     {
         POLYUNSIGNED length = p->Length();
@@ -757,13 +891,14 @@ unsigned long ExportStringTable::makeEntry(const char *str)
         stringAvailable = stringAvailable+stringAvailable/2;
         if (stringAvailable < stringSize + len + 1)
             stringAvailable = stringSize + len + 1 + 500;
-        strings = (char*)realloc(strings, stringAvailable);
-        if (strings == 0)
+        char* newStrings = (char*)realloc(strings, stringAvailable);
+        if (newStrings == 0)
         {
             if (debugOptions & DEBUG_SAVING)
                 Log("SAVE: Unable to realloc string table, size: %lu.\n", stringAvailable);
             throw MemoryException();
         }
+        else strings = newStrings;
      }
     strcpy(strings + stringSize, str);
     stringSize += len + 1;
