@@ -1,7 +1,7 @@
 /*
     Title:  exporter.h - Export a function as an object or C file
 
-    Copyright (c) 2006, 2015-16 David C.J. Matthews
+    Copyright (c) 2006, 2015-17 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -53,7 +53,7 @@ protected:
     void relocateObject(PolyObject *p);
     void createRelocation(PolyWord *pt) { *pt = createRelocation(*pt, pt); }
     unsigned findArea(void *p); // Find index of area that address is in.
-    virtual void addExternalReference(void *p, const char *entryPoint) {}
+    virtual void addExternalReference(void *p, const char *entryPoint, bool isFuncPtr) {}
 
 public:
     FILE     *exportFile;
@@ -100,11 +100,15 @@ public:
     ~CopyScan();
 protected:
     virtual POLYUNSIGNED ScanAddressAt(PolyWord *pt);
+    // Have to follow pointers from closures into code.
+    virtual POLYUNSIGNED ScanCodeAddressAt(PolyObject **pt);
+    POLYUNSIGNED ScanAddress(PolyObject **pt);
+
 public:
     virtual PolyObject *ScanObjectAddress(PolyObject *base);
 
     // Default sizes of the segments.
-    POLYUNSIGNED defaultImmSize, defaultCodeSize, defaultMutSize, defaultNoOverSize;
+    uintptr_t defaultImmSize, defaultCodeSize, defaultMutSize, defaultNoOverSize;
     unsigned hierarchy;
 
     GraveYard *graveYard;
