@@ -1,7 +1,7 @@
 /*
     Title:      Multi-Threaded Garbage Collector - Mark phase
 
-    Copyright (c) 2010-12, 2015-16 David C. J. Matthews
+    Copyright (c) 2010-12, 2015-16, 2019 David C. J. Matthews
 
     Based on the original garbage collector code
         Copyright 2000-2008
@@ -97,6 +97,8 @@ public:
         { ScanAddressesInObject(base, base->LengthWord()); }
 
     virtual void ScanConstant(PolyObject *base, byte *addressOfConstant, ScanRelocationKind code);
+    // ScanCodeAddressAt should never be called.
+    POLYUNSIGNED ScanCodeAddressAt(PolyObject **pt) { ASSERT(0); return 0; }
 
     static void MarkPointersTask(GCTaskId *, void *arg1, void *arg2);
 
@@ -115,7 +117,7 @@ private:
     void MarkAndTestForScan(PolyWord *pt);
     void Reset();
 
-    void PushToStack(PolyObject *obj, PolyWord *currentPtr = 0, POLYUNSIGNED originalLength = 0)
+    void PushToStack(PolyObject *obj, PolyWord *currentPtr = 0)
     {
         // If we don't have all the threads running we start a new one but
         // only once we have several items on the stack.  Otherwise we
@@ -207,7 +209,7 @@ void MTGCProcessMarkPointers::Reset()
 // in the range to be rescanned.
 void MTGCProcessMarkPointers::StackOverflow(PolyObject *obj)
 {
-    MarkableSpace *space = (MarkableSpace*)gMem.SpaceForAddress(obj-1);
+    MarkableSpace *space = (MarkableSpace*)gMem.SpaceForAddress((PolyWord*)obj-1);
     ASSERT(space != 0 && (space->spaceType == ST_LOCAL || space->spaceType == ST_CODE));
     PLocker lock(&space->spaceLock);
     // Have to include this in the range to rescan.
@@ -312,7 +314,7 @@ bool MTGCProcessMarkPointers::TestForScan(PolyWord *pt)
         *pt = obj;
     }
 
-    MemSpace *sp = gMem.SpaceForAddress(obj-1);
+    MemSpace *sp = gMem.SpaceForAddress((PolyWord*)obj-1);
     if (sp == 0 || (sp->spaceType != ST_LOCAL && sp->spaceType != ST_CODE))
         return false; // Ignore it if it points to a permanent area
 
@@ -346,18 +348,15 @@ void MTGCProcessMarkPointers::MarkAndTestForScan(PolyWord *pt)
 // updated address of an object.
 PolyObject *MTGCProcessMarkPointers::ScanObjectAddress(PolyObject *obj)
 {
-    PolyWord val = obj;
-    MemSpace *sp = gMem.SpaceForAddress(val.AsStackAddr()-1);
+    MemSpace *sp = gMem.SpaceForAddress((PolyWord*)obj-1);
+
     if (!(sp->spaceType == ST_LOCAL || sp->spaceType == ST_CODE))
         return obj; // Ignore it if it points to a permanent area
 
     // We may have a forwarding pointer if this has been moved by the
     // minor GC.
     if (obj->ContainsForwardingPtr())
-    {
         obj = FollowForwarding(obj);
-        val = obj;
-    }
 
     ASSERT(obj->ContainsNormalLengthWord());
 
@@ -407,6 +406,7 @@ void MTGCProcessMarkPointers::ScanRuntimeAddress(PolyObject **pt, RtsStrength we
 // This is called via ScanAddressesInRegion to process the permanent mutables.  It is
 // also called from ScanObjectAddress to process root addresses.
 // It processes all the addresses reachable from the object.
+// This is almost the same as RecursiveScan::ScanAddressesInObject. 
 void MTGCProcessMarkPointers::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord)
 {
     if (OBJ_IS_BYTE_OBJECT(lengthWord))
@@ -416,9 +416,9 @@ void MTGCProcessMarkPointers::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNE
     {
         ASSERT (OBJ_IS_LENGTH(lengthWord));
 
-        // Get the length and base address.  N.B.  If this is a code segment
-        // these will be side-effected by GetConstSegmentForCode.
         POLYUNSIGNED length = OBJ_OBJECT_LENGTH(lengthWord);
+        PolyWord *baseAddr = (PolyWord*)obj;
+        PolyWord *endWord = baseAddr + length;
 
         if (OBJ_IS_WEAKREF_OBJECT(lengthWord))
         {
@@ -427,22 +427,33 @@ void MTGCProcessMarkPointers::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNE
             ASSERT(OBJ_IS_WORD_OBJECT(lengthWord)); // Should be a plain object.
             // We need to mark the "SOME" values in this object but we don't mark
             // the references contained within the "SOME".
-            PolyWord *baseAddr = (PolyWord*)obj;
             // Mark every word but ignore the result.
             for (POLYUNSIGNED i = 0; i < length; i++)
                 (void)MarkAndTestForScan(baseAddr+i);
             // We've finished with this.
-            length = 0;
+            endWord = baseAddr;
         }
 
         else if (OBJ_IS_CODE_OBJECT(lengthWord))
         {
+            // Legacy: The code-generator now uses PolyCopyByteVecToClosure to allocate mutable
+            // code cells in the code area.  Previously they were allocated in the heap and copied
+            // into the code area only when they were locked.
             // It's better to process the whole code object in one go.
             ScanAddress::ScanAddressesInObject(obj, lengthWord);
-            length = 0; // Finished
+            endWord = baseAddr; // Finished
         }
 
-        // else it's a normal object,
+        else if (OBJ_IS_CLOSURE_OBJECT(lengthWord))
+        {
+            // The first word is the absolute address of the code ...
+            PolyObject *codeAddr = *(PolyObject**)obj;
+            // except that it is possible we haven't yet set it.
+            if (((uintptr_t)codeAddr & 1) == 0)
+                ScanObjectAddress(codeAddr);
+            // The rest is a normal tuple.
+            baseAddr += sizeof(PolyObject*) / sizeof(PolyWord);
+        }
 
         // If there are only two addresses in this cell that need to be
         // followed we follow them immediately and treat this cell as done.
@@ -450,8 +461,6 @@ void MTGCProcessMarkPointers::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNE
         // the stack, follow the first address and then rescan it.  That way
         // list cells are processed once only but we don't overflow the
         // stack by pushing all the addresses in a very large vector.
-        PolyWord *baseAddr = (PolyWord*)obj;
-        PolyWord *endWord = baseAddr + length;
         PolyObject *firstWord = 0;
         PolyObject *secondWord = 0;
         PolyWord *restartAddr = 0;
@@ -459,8 +468,8 @@ void MTGCProcessMarkPointers::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNE
         if (obj == largeObjectCache[locPtr].base)
         {
             baseAddr = largeObjectCache[locPtr].current;
-            ASSERT(baseAddr > (PolyWord*)obj && baseAddr < ((PolyWord*)obj)+length);
-            if (locPtr == 0) locPtr = LARGECACHE_SIZE-1; else locPtr--;
+            ASSERT(baseAddr > (PolyWord*)obj && baseAddr < endWord);
+            if (locPtr == 0) locPtr = LARGECACHE_SIZE - 1; else locPtr--;
         }
 
         while (baseAddr != endWord)
@@ -492,7 +501,7 @@ void MTGCProcessMarkPointers::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNE
 
         if (baseAddr != endWord)
             // Put this back on the stack while we process the first word
-            PushToStack(obj, length < largeObjectSize ? 0 : restartAddr, length);
+            PushToStack(obj, length < largeObjectSize ? 0 : restartAddr);
         else if (secondWord != 0)
         {
             // Mark it now because we will process it.
@@ -547,19 +556,18 @@ void MTGCProcessMarkPointers::ScanConstant(PolyObject *base, byte *addressOfCons
 
     if (lock != 0)
         lock->Lock();
-    PolyWord p = GetConstantValue(addressOfConstant, code);
+    PolyObject *p = GetConstantValue(addressOfConstant, code);
     if (lock != 0)
         lock->Unlock();
 
-    if (! IS_INT(p))
+    if (p != 0)
     {
-        PolyWord oldValue = p;
-        ScanAddress::ScanAddressAt(&p);
-        if (p != oldValue) // Update it if it has changed.
+        PolyObject *newVal = ScanObjectAddress(p);
+        if (newVal != p) // Update it if it has changed.
         {
             if (lock != 0)
                 lock->Lock();
-            SetConstantValue(addressOfConstant, p, code);
+            SetConstantValue(addressOfConstant, newVal, code);
             if (lock != 0)
                 lock->Unlock();
         }
@@ -613,6 +621,7 @@ public:
 
     // Have to define this.
     virtual PolyObject *ScanObjectAddress(PolyObject *base) { ASSERT(false); return 0; }
+    virtual POLYUNSIGNED ScanCodeAddressAt(PolyObject **pt) { ASSERT(false); return 0; }
 
     bool ScanSpace(MarkableSpace *space);
 private:
@@ -678,6 +687,13 @@ static void SetBitmaps(LocalMemSpace *space, PolyWord *pt, PolyWord *top)
 {
     while (pt < top)
     {
+#ifdef POLYML32IN64
+        if ((((uintptr_t)pt) & 4) == 0)
+        {
+            pt++;
+            continue;
+        }
+#endif
         PolyObject *obj = (PolyObject*)++pt;
         // If it has been copied by a minor collection skip it
         if (obj->ContainsForwardingPtr())
@@ -693,7 +709,7 @@ static void SetBitmaps(LocalMemSpace *space, PolyWord *pt, PolyWord *top)
             if (L & _OBJ_GC_MARK)
             {
                 obj->SetLengthWord(L & ~(_OBJ_GC_MARK));
-                POLYUNSIGNED bitno = space->wordNo(pt);
+                uintptr_t bitno = space->wordNo(pt);
                 space->bitmap.SetBits(bitno - 1, n + 1);
 
                 if (OBJ_IS_MUTABLE_OBJECT(L))
@@ -731,7 +747,11 @@ static void CreateBitmapsTask(GCTaskId *, void *arg1, void *arg2)
 static void CheckMarksOnCodeTask(GCTaskId *, void *arg1, void *arg2)
 {
     CodeSpace *space = (CodeSpace*)arg1;
+#ifdef POLYML32IN64
+    PolyWord *pt = space->bottom+1;
+#else
     PolyWord *pt = space->bottom;
+#endif
     PolyWord *lastFree = 0;
     POLYUNSIGNED lastFreeSpace = 0;
     space->largestFree = 0;
@@ -751,6 +771,20 @@ static void CheckMarksOnCodeTask(GCTaskId *, void *arg1, void *arg2)
             lastFree = 0;
             lastFreeSpace = 0;
         }
+#ifdef POLYML32IN64
+        else if (length == 0)
+        {
+            // We may have zero filler words to set the correct alignment.
+            // Merge them into a previously free area otherwise leave
+            // them if they're after something allocated.
+            if (lastFree + lastFreeSpace == pt)
+            {
+                lastFreeSpace += length + 1;
+                PolyObject *freeSpace = (PolyObject*)(lastFree + 1);
+                freeSpace->SetLengthWord(lastFreeSpace - 1, F_BYTE_OBJ);
+            }
+        }
+#endif
         else { // Turn it into a byte area i.e. free.  It may already be free.
             if (space->firstFree == 0) space->firstFree = pt;
             space->headerMap.ClearBit(pt-space->bottom); // Remove the "header" bit
@@ -815,7 +849,7 @@ void GCMarkPhase(void)
 
     gHeapSizeParameters.RecordGCTime(HeapSizeParameters::GCTimeIntermediate, "Bitmap");
 
-    POLYUNSIGNED totalLive = 0;
+    uintptr_t totalLive = 0;
     for(std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
         LocalMemSpace *lSpace = *i;
