@@ -1,12 +1,11 @@
 /*
     Title:      Task farm for Multi-Threaded Garbage Collector
 
-    Copyright (c) 2010 David C. J. Matthews
+    Copyright (c) 2010, 2019 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -59,18 +58,14 @@ GCTaskFarm::GCTaskFarm(): workLock("GC task farm work")
     workQueue = 0;
     terminate = false;
     threadCount = activeThreadCount = 0;
-#if (defined(HAVE_PTHREAD_H) || defined(HAVE_WINDOWS_H))
     threadHandles = 0;
-#endif
 }
 
 GCTaskFarm::~GCTaskFarm()
 {
     Terminate();
     free(workQueue);
-#if (defined(HAVE_PTHREAD_H) || defined(HAVE_WINDOWS_H))
     free(threadHandles);
-#endif
 }
 
 
@@ -80,28 +75,26 @@ bool GCTaskFarm::Initialise(unsigned thrdCount, unsigned qSize)
     if (!waitForWork.Init(0, thrdCount)) return false;
     workQueue = (queue_entry*)calloc(qSize, sizeof(queue_entry));
     if (workQueue == 0) return false;
-#if ((!defined(_WIN32) || defined(__CYGWIN__)) && defined(HAVE_PTHREAD_H))
+#if (!defined(_WIN32))
     queueSize = qSize;
     threadHandles = (pthread_t*)calloc(thrdCount, sizeof(pthread_t));
     if (threadHandles == 0) return false;
-#elif defined(HAVE_WINDOWS_H)
+#else
     queueSize = qSize;
     threadHandles = (HANDLE*)calloc(thrdCount, sizeof(HANDLE));
     if (threadHandles == 0) return false;
-#else
-    queueSize = 0;
 #endif
     // Create the worker threads.
     for (unsigned i = 0; i < thrdCount; i++) {
         // Fork a thread
-#if ((!defined(_WIN32) || defined(__CYGWIN__)) && defined(HAVE_PTHREAD_H))
+#if (!defined(_WIN32))
         // Create a thread that isn't joinable since we don't want to wait
         // for it to finish.
         pthread_t pthreadId;
         bool isError = pthread_create(&pthreadId, NULL, WorkerThreadFunction, this) != 0;
         if (isError) break;
         threadHandles[threadCount++] = pthreadId;
-#elif defined(HAVE_WINDOWS_H)
+#else
         DWORD dwThrdId; // Have to provide this although we don't use it.
         HANDLE threadHandle =
             CreateThread(NULL, 0, WorkerThreadFunction, this, 0, &dwThrdId);
@@ -119,10 +112,10 @@ void GCTaskFarm::Terminate()
     // Increment the semaphore by the number of threads to release them all.
     for (unsigned i = 0; i < threadCount; i++) waitForWork.Signal();
     // Wait for the threads to terminate.
-#if ((!defined(_WIN32) || defined(__CYGWIN__)) && defined(HAVE_PTHREAD_H))
+#if (!defined(_WIN32))
     for (unsigned j = 0; j < threadCount; j++)
         pthread_join(threadHandles[j], NULL);
-#elif defined(HAVE_WINDOWS_H)
+#else
     if (threadCount != 0)
         WaitForMultipleObjects(threadCount, threadHandles, TRUE, 10000);
 #endif
@@ -134,7 +127,8 @@ bool GCTaskFarm::AddWork(gctask work, void *arg1, void *arg2)
     bool wantSignal = false;
     {
         PLocker l(&workLock);
-        if (queuedItems == queueSize) return false; // Queue is full
+        if (queuedItems == queueSize)
+			return false; // Queue is full
         workQueue[queueIn].task = work;
         workQueue[queueIn].arg1 = arg1;
         workQueue[queueIn].arg2 = arg2;
@@ -157,7 +151,7 @@ void GCTaskFarm::AddWorkOrRunNow(gctask work, void *arg1, void *arg2)
 void GCTaskFarm::ThreadFunction()
 {
     GCTaskId myTaskId;
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     DWORD startActive = GetTickCount();
 #else
     struct timeval startTime;
@@ -198,7 +192,7 @@ void GCTaskFarm::ThreadFunction()
 
             if (debugOptions & DEBUG_GCTASKS)
             {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
                 Log("GCTask: Thread %p blocking after %u milliseconds\n", &myTaskId,
                      GetTickCount() - startActive);
 #else
@@ -216,7 +210,7 @@ void GCTaskFarm::ThreadFunction()
             // We've been woken up
             if (debugOptions & DEBUG_GCTASKS)
             {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
                 startActive = GetTickCount();
 #else
                 gettimeofday(&startTime, NULL);
@@ -231,14 +225,14 @@ void GCTaskFarm::ThreadFunction()
     workLock.Unlock();
 }
 
-#if ((!defined(_WIN32) || defined(__CYGWIN__)) && defined(HAVE_PTHREAD_H))
+#if (!defined(_WIN32))
 void *GCTaskFarm::WorkerThreadFunction(void *parameter)
 {
     GCTaskFarm *t = (GCTaskFarm *)parameter;
     t->ThreadFunction();
     return 0;
 }
-#elif defined(HAVE_WINDOWS_H)
+#else
 DWORD WINAPI GCTaskFarm::WorkerThreadFunction(void *parameter)
 {
     GCTaskFarm *t = (GCTaskFarm *)parameter;
@@ -250,7 +244,7 @@ DWORD WINAPI GCTaskFarm::WorkerThreadFunction(void *parameter)
 // Wait until the queue is empty.
 void GCTaskFarm::WaitForCompletion(void)
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     DWORD startWait;
     if (debugOptions & DEBUG_GCTASKS)
         startWait = GetTickCount();
@@ -266,7 +260,7 @@ void GCTaskFarm::WaitForCompletion(void)
 
     if (debugOptions & DEBUG_GCTASKS)
     {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
         Log("GCTask: Threads completed after %u milliseconds\n", GetTickCount()-startWait);
 #else
         struct timeval endWait;
