@@ -2,13 +2,16 @@
     Title:  Globals for the system.
     Author:     Dave Matthews, Cambridge University Computer Laboratory
 
+    Copyright David C. J. Matthews 2017-19
+
     Copyright (c) 2000-7
         Cambridge University Technical Services Limited
 
+     Further work copyright David C.J. Matthews 2006-18
+
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -25,41 +28,23 @@
 #define _GLOBALS_H
 
 /*
-    Poly words, pointers and objects.
+    Poly words, pointers and cells (objects).
 
     The garbage collector needs to be able to distinguish different uses of a
     memory word.  We need to be able find which words are pointers to other
     objects and which are simple integers.  The simple distinction is between
     integers, which are tagged by having the bottom bit set, and Addresses
-    which are word aligned (bottom 2 bits set on a 32 bit machine, bottom 3
-    bits on a 64 bit machine).  There is a third case that can arise where
-    we have a return address to a piece of code.  These
-    are generally aligned on a word+two-byte boundary except that they can
-    appear with any alignment in the saved PC field of a stack.  Normally
-    these will only appear on the stack or in a register saved to the stack
-    but they can be passed as an argument to SetCodeConstant.
+    which are word aligned (bottom 2 bits zero on a 32 bit machine, bottom 3
+    bits on a 64 bit machine, bottom bit in 32-in-64).  
 
-    Addresses always point to the start of objects.  There is one exception:
-    a stack object can contain addresses within the current stack.  The
-    preceding word of a stack object is the length word.  This contains the
-    length of the object in words in the low-order 3 (7 on a 64-bit machine)
+    Addresses always point to the start of cells.  The preceding word of a
+    cell is the length word.  This contains the
+    length of the cell in words in the low-order 3 (7 in native 64-bits)
     bytes and a flag byte in the top byte.  The flags give information about
     the type of the object.  The length word is also used by the garbage
     collector and other object processors.
 */
 
-/* Generally we use a tagged integer representation with 31 bits of
-   precision and the low-order bit as a tag bit.  one represents a tagged
-   integer, zero a pointer.  Code pointers are always aligned onto a
-   word + 2 byte boundary.
-   The exception is the Sparc which has special tagged add and subtract
-   instructions.  These cause a trap if given a value which is not a
-   multiple of 4.  To make use of these instructions we represent tagged
-   integers using 30 bits of precision and with the low order bits
-   containing 01. 
-   DCJM 27/9/00.
- */
-
 #if HAVE_STDINT_H
 #  include <stdint.h>
 #endif
@@ -69,30 +54,49 @@
 #    define __STDC_FORMAT_MACROS
 #  endif
 #  include <inttypes.h>
+
+#elif (defined(_MSC_VER) && (_MSC_VER >= 1900))
+// In VS 2015 and later we need to use <cinttypes>
+#  include <cinttypes>  
 #endif
 
 #ifdef HAVE_STDDEF_H
 #  include <stddef.h>
 #endif
 
-// Integers (non-pointers) have the bottom bit set and the rest of the word
-// contains the significant bits.  The original native code Sparc version
-// shifted by two so that it could use the Taddcc/Tsubcc instructions.
 #define POLY_TAGSHIFT    1
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #  include <windows.h>
 #endif
 
+#ifdef POLYML32IN64
+typedef int32_t         POLYSIGNED;
+typedef uint32_t        POLYUNSIGNED;
+#define SIZEOF_POLYWORD 4
+#else
 typedef intptr_t        POLYSIGNED;
 typedef uintptr_t       POLYUNSIGNED;
+#define SIZEOF_POLYWORD SIZEOF_VOIDP
+#endif
 
 // libpolyml uses printf-style I/O instead of C++ standard IOstreams,
 // so we need specifier to format POLYUNSIGNED/POLYSIGNED values.
-#ifdef PRIuPTR
+#ifdef POLYML32IN64
+#if (defined(PRIu32))
+#  define POLYUFMT PRIu32
+#  define POLYSFMT PRId32
+#elif (defined(_MSC_VER))
+#  define POLYUFMT "lu"
+#  define POLYSFMT "ld"
+#else
+#  define POLYUFMT "u"
+#  define POLYSFMT "d"
+#endif
+#elif (defined(PRIuPTR))
 #  define POLYUFMT PRIuPTR
 #  define POLYSFMT PRIdPTR
-#elif (defined(_MSC_VER) && (SIZEOF_VOIDP == 8))
+#elif (defined(_MSC_VER) && (SIZEOF_POLYWORD == 8))
 #  define POLYUFMT "llu"
 #  define POLYSFMT "lld"
 #else
@@ -100,18 +104,39 @@ typedef uintptr_t       POLYUNSIGNED;
 #  define POLYSFMT "ld"         // idem.
 #endif
 
+// We can use the C99 %zu in most cases except MingW since it uses
+// the old msvcrt and that only supports C89.
+#if (defined(_WIN32) && (! defined(_MSC_VER) || _MSC_VER < 1800))
+#  if (SIZEOF_VOIDP == 8)
+#    define PRI_SIZET PRIu64
+#  else
+#    define PRI_SIZET PRIu32
+#  endif
+#else
+#   define PRI_SIZET "zu"
+#endif
+
 typedef unsigned char   byte;
 
 class PolyObject;
 typedef PolyObject *POLYOBJPTR;
 
+#ifdef POLYML32IN64
+class PolyWord;
+extern PolyWord *globalHeapBase, *globalCodeBase;
+typedef uint32_t POLYOBJECTPTR; // This is an index into globalHeapBase
+
+// If a 64-bit value if in the range of the object pointers.
+inline bool IsHeapAddress(void *addr) { return (uintptr_t)addr <= 0xffffffff; }
+#else
+typedef POLYOBJPTR POLYOBJECTPTR;
+inline bool IsHeapAddress(void *) { return true; }
+#endif
+
 typedef byte *POLYCODEPTR;
 
 class PolyWord {
 public:
-    // An object pointer can become a word directly.
-    PolyWord(POLYOBJPTR p) { contents.objectPtr = p; }
-
     // Initialise to TAGGED(0).  This is very rarely used.
     PolyWord() { contents.unsignedInt = 1; }
 
@@ -124,15 +149,34 @@ public:
 
     // Tests for the various cases.
     bool IsTagged(void) const { return (contents.unsignedInt & 1) != 0; }
-    bool IsDataPtr(void) const { return (contents.unsignedInt & (sizeof(PolyWord)-1)) == 0; }
+#ifndef POLYML32IN64
+    // In native 32-bit and 64-bit addresses are on word boundaries
+    bool IsDataPtr(void) const { return (contents.unsignedInt & (sizeof(PolyWord) - 1)) == 0; }
+#else
+    // In 32-in-64 addresses are anything that isn't tagged.
+    bool IsDataPtr(void) const { return (contents.unsignedInt & 1) == 0; }
+#ifdef POLYML32IN64DEBUG
+    static POLYOBJECTPTR AddressToObjectPtr(void *address);
+#else
+    static POLYOBJECTPTR AddressToObjectPtr(void *address)
+        { return (POLYOBJECTPTR)((PolyWord*)address - globalHeapBase); }
+#endif
+#endif
 
     // Extract the various cases.
     POLYSIGNED UnTagged(void) const { return contents.signedInt >> POLY_TAGSHIFT; }
     POLYUNSIGNED UnTaggedUnsigned(void) const { return contents.unsignedInt >> POLY_TAGSHIFT; }
-    POLYOBJPTR AsObjPtr(void) const { return contents.objectPtr; }
-    PolyWord *AsStackAddr(void) const { return contents.stackAddr; }
-    POLYCODEPTR AsCodePtr(void) const { return contents.codePtr; }
-
+#ifdef POLYML32IN64
+    PolyWord(POLYOBJPTR p) { contents.objectPtr = AddressToObjectPtr(p); }
+    PolyWord *AsStackAddr(PolyWord *base = globalHeapBase) const { return base + contents.objectPtr; }
+    POLYOBJPTR AsObjPtr(PolyWord *base = globalHeapBase) const { return (POLYOBJPTR)AsStackAddr(base); }
+#else
+    // An object pointer can become a word directly.
+    PolyWord(POLYOBJPTR p) { contents.objectPtr = p; }
+    POLYOBJPTR AsObjPtr(PolyWord *base = 0) const { return contents.objectPtr; }
+    PolyWord *AsStackAddr(PolyWord *base=0) const { return (PolyWord *)contents.objectPtr; }
+#endif
+    POLYCODEPTR AsCodePtr(void) const { return (POLYCODEPTR)AsObjPtr(); }
     void *AsAddress(void)const { return AsCodePtr(); }
 
     // There are a few cases where we need to store and extract untagged values
@@ -150,14 +194,17 @@ public:
     bool operator != (PolyWord b) const { return contents.unsignedInt != b.contents.unsignedInt; }
 
 protected:
-    PolyWord(PolyWord *sp) { contents.stackAddr = sp; }
-    PolyWord(POLYCODEPTR p) { contents.codePtr = p; }
+#ifdef POLYML32IN64
+    PolyWord(PolyWord *sp) { contents.objectPtr = AddressToObjectPtr(sp); }
+    PolyWord(POLYCODEPTR p) { contents.objectPtr = AddressToObjectPtr(p); }
+#else
+    PolyWord(PolyWord *sp) { contents.objectPtr = (PolyObject*)sp; }
+    PolyWord(POLYCODEPTR p) { contents.objectPtr = (PolyObject*)p; }
+#endif
     union {
         POLYSIGNED      signedInt;      // A tagged integer - lowest bit set
         POLYUNSIGNED    unsignedInt;    // A tagged integer - lowest bit set
-        POLYOBJPTR      objectPtr;      // Object pointer   - two lowest bits clear.
-        POLYCODEPTR     codePtr;        // Address within code - lowest bits contain 10.
-        PolyWord        *stackAddr;     // Address within current stack - two lowest bits clear.
+        POLYOBJECTPTR   objectPtr;      // Object pointer   - lowest bit clear.
     } contents;
 };
 
@@ -175,31 +222,7 @@ inline POLYUNSIGNED UNTAGGED_UNSIGNED(PolyWord a)   { return a.UnTaggedUnsigned(
 
 #define IS_INT(x) ((x).IsTagged())
 
-/************************************************************************
- *
- * Low level definitions.
- *
- ************************************************************************/
-
-
 /* length word flags */
-
-/*
-The following encodings are used:
-
-(1) OBJ_PRIVATE_GC_BIT = 0
-       A normal length word i.e. 7 more flags + 24 bit length.
-       
-(2) OBJ_PRIVATE_GC_BIT = 1, OBJ_PRIVATE_MUTABLE_BIT = 0
-       A "normal" tombstone. Remaining 30 bits are (word-aligned) pointer >> 2.
-
-(3) OBJ_PRIVATE_GC_BIT = 1, OBJ_PRIVATE_MUTABLE_BIT = 1
-       A "depth" tombstone; used by share program. Remaining 30 bits are
-       an unsigned integer representing the (approximate) depth of the object
-       in the data-structure. 
-
-SPF 24/1/95
-*/
 #define OBJ_PRIVATE_FLAGS_SHIFT     (8 * (sizeof(PolyWord) - 1))
 
 #define _TOP_BYTE(x)                ((POLYUNSIGNED)(x) << OBJ_PRIVATE_FLAGS_SHIFT)
@@ -208,6 +231,7 @@ SPF 24/1/95
 // Zero bits mean ordinary word object containing addresses or tagged integers.
 #define F_BYTE_OBJ                  0x01  /* byte object (contains no pointers) */
 #define F_CODE_OBJ                  0x02  /* code object (mixed bytes and words) */
+#define F_CLOSURE_OBJ               0x03  /* closure (32-in-64 only).  First word is code addr. */
 #define F_GC_MARK                   0x04  // Used during the GC marking phase
 #define F_NO_OVERWRITE              0x08  /* don't overwrite when loading - mutables only. */
 // This bit is overloaded and has different meanings depending on what other bits are set.
@@ -217,27 +241,29 @@ SPF 24/1/95
 #define F_NEGATIVE_BIT              0x10  // Sign bit for arbitrary precision ints (byte segs only)
 #define F_PROFILE_BIT               0x10  // Object has a profile pointer (word segs only)
 #define F_WEAK_BIT                  0x20  /* object contains weak references to option values. */
+// The Weak bit is only used on mutables.  The data sharing (sharedata.cpp) uses this with
+// immutables to indicate that the length field is being used to store the "depth".
 #define F_MUTABLE_BIT               0x40  /* object is mutable */
-#define F_PRIVATE_GC_BIT            0x80  /* object is (pointer or depth) tombstone */
+#define F_TOMBSTONE_BIT             0x80  // Object is a forwarding pointer
 #define F_PRIVATE_FLAGS_MASK        0xFF
 
-#define _OBJ_BYTE_OBJ                _TOP_BYTE(0x01)  /* byte object (contains no pointers) */
-#define _OBJ_CODE_OBJ                _TOP_BYTE(0x02)  /* code object (mixed bytes and words) */
-#define _OBJ_GC_MARK                 _TOP_BYTE(0x04)  // Mark bit
-#define _OBJ_NO_OVERWRITE            _TOP_BYTE(0x08)  /* don't overwrite when loading - mutables only. */
-#define _OBJ_NEGATIVE_BIT            _TOP_BYTE(0x10)  /* sign bit for arbitrary precision ints */
-#define _OBJ_PROFILE_BIT             _TOP_BYTE(0x10)  /* sign bit for arbitrary precision ints */
-#define _OBJ_WEAK_BIT                _TOP_BYTE(0x20)
-#define _OBJ_MUTABLE_BIT             _TOP_BYTE(0x40)  /* object is mutable */
-#define _OBJ_PRIVATE_GC_BIT          _TOP_BYTE(0x80)  /* object is (pointer or depth) tombstone */
-#define _OBJ_PRIVATE_FLAGS_MASK      _TOP_BYTE(0xFF)
+// Shifted bits
+#define _OBJ_BYTE_OBJ                _TOP_BYTE(F_BYTE_OBJ)  /* byte object (contains no pointers) */
+#define _OBJ_CODE_OBJ                _TOP_BYTE(F_CODE_OBJ)  /* code object (mixed bytes and words) */
+#define _OBJ_CLOSURE_OBJ             _TOP_BYTE(F_CLOSURE_OBJ)  // closure (32-in-64 only).  First word is code addr.
+#define _OBJ_GC_MARK                 _TOP_BYTE(F_GC_MARK)  // Mark bit
+#define _OBJ_NO_OVERWRITE            _TOP_BYTE(F_NO_OVERWRITE)  /* don't overwrite when loading - mutables only. */
+#define _OBJ_NEGATIVE_BIT            _TOP_BYTE(F_NEGATIVE_BIT)  /* sign bit for arbitrary precision ints */
+#define _OBJ_PROFILE_BIT             _TOP_BYTE(F_PROFILE_BIT)  /* sign bit for arbitrary precision ints */
+#define _OBJ_WEAK_BIT                _TOP_BYTE(F_WEAK_BIT)
+#define _OBJ_MUTABLE_BIT             _TOP_BYTE(F_MUTABLE_BIT)  /* object is mutable */
+#define _OBJ_TOMBSTONE_BIT           _TOP_BYTE(F_TOMBSTONE_BIT)  // object is a tombstone.
+#define _OBJ_PRIVATE_FLAGS_MASK      _TOP_BYTE(F_PRIVATE_FLAGS_MASK)
 #define _OBJ_PRIVATE_LENGTH_MASK     ((-1) ^ _OBJ_PRIVATE_FLAGS_MASK)
 #define MAX_OBJECT_SIZE              _OBJ_PRIVATE_LENGTH_MASK
 
-#define _OBJ_PRIVATE_DEPTH_MASK      (_OBJ_PRIVATE_GC_BIT|_OBJ_MUTABLE_BIT)
-
-/* case 1 - proper length words */
-inline bool OBJ_IS_LENGTH(POLYUNSIGNED L)               { return ((L & _OBJ_PRIVATE_GC_BIT) == 0); }
+// 
+inline bool OBJ_IS_LENGTH(POLYUNSIGNED L)               { return ((L & _OBJ_TOMBSTONE_BIT) == 0); }
 
 /* these should only be applied to proper length words */
 /* discards GC flag, mutable bit and weak bit. */
@@ -246,6 +272,7 @@ inline byte GetTypeBits(POLYUNSIGNED L)             { return (byte)(L >> OBJ_PRI
 inline POLYUNSIGNED OBJ_OBJECT_LENGTH(POLYUNSIGNED L)   { return L & _OBJ_PRIVATE_LENGTH_MASK; }
 inline bool OBJ_IS_BYTE_OBJECT(POLYUNSIGNED L)          { return (GetTypeBits(L) == F_BYTE_OBJ); }
 inline bool OBJ_IS_CODE_OBJECT(POLYUNSIGNED L)          { return (GetTypeBits(L) == F_CODE_OBJ); }
+inline bool OBJ_IS_CLOSURE_OBJECT(POLYUNSIGNED L)       { return (GetTypeBits(L) == F_CLOSURE_OBJ); }
 inline bool OBJ_IS_NO_OVERWRITE(POLYUNSIGNED L)         { return ((L & _OBJ_NO_OVERWRITE) != 0); }
 inline bool OBJ_IS_NEGATIVE(POLYUNSIGNED L)             { return ((L & _OBJ_NEGATIVE_BIT) != 0); }
 inline bool OBJ_HAS_PROFILE(POLYUNSIGNED L)             { return ((L & _OBJ_PROFILE_BIT) != 0); }
@@ -261,18 +288,14 @@ inline bool OBJ_IS_WEAKREF_OBJECT(POLYUNSIGNED L)       { return ((L & _OBJ_WEAK
 #define OBJ_IS_WORD_OBJECT(L)           (GetTypeBits(L) == 0)
 
 /* case 2 - forwarding pointer */
-inline bool OBJ_IS_POINTER(POLYUNSIGNED L)  { return (L & _OBJ_PRIVATE_DEPTH_MASK) == _OBJ_PRIVATE_GC_BIT; }
-inline PolyObject *OBJ_GET_POINTER(POLYUNSIGNED L) { return (PolyObject*)(( L & ~_OBJ_PRIVATE_DEPTH_MASK) <<2); }
-inline POLYUNSIGNED OBJ_SET_POINTER(PolyObject *pt) { return ((POLYUNSIGNED)pt >> 2) | _OBJ_PRIVATE_GC_BIT; }
-
-/* case 3 - depth */
-inline bool OBJ_IS_DEPTH(POLYUNSIGNED L)  { return (L & _OBJ_PRIVATE_DEPTH_MASK) == _OBJ_PRIVATE_DEPTH_MASK; }
-inline POLYUNSIGNED OBJ_GET_DEPTH(POLYUNSIGNED L) { return L & ~_OBJ_PRIVATE_DEPTH_MASK; }
-inline POLYUNSIGNED OBJ_SET_DEPTH(POLYUNSIGNED n) { return n | _OBJ_PRIVATE_DEPTH_MASK; }
-// or sharing chain
-inline bool OBJ_IS_CHAINED(POLYUNSIGNED L)  { return (L & _OBJ_PRIVATE_DEPTH_MASK) == _OBJ_PRIVATE_DEPTH_MASK; }
-inline PolyObject *OBJ_GET_CHAIN(POLYUNSIGNED L) { return (PolyObject*)(( L & ~_OBJ_PRIVATE_DEPTH_MASK) <<2); }
-inline POLYUNSIGNED OBJ_SET_CHAIN(PolyObject *pt) { return ((POLYUNSIGNED)pt >> 2) | _OBJ_PRIVATE_DEPTH_MASK; }
+inline bool OBJ_IS_POINTER(POLYUNSIGNED L)  { return (L & _OBJ_TOMBSTONE_BIT) != 0; }
+#ifdef POLYML32IN64
+inline PolyObject *OBJ_GET_POINTER(POLYUNSIGNED L) { return (PolyObject*)(globalHeapBase + ((L & ~_OBJ_TOMBSTONE_BIT) << 1)); }
+inline POLYUNSIGNED OBJ_SET_POINTER(PolyObject *pt) { return PolyWord::AddressToObjectPtr(pt) >> 1 | _OBJ_TOMBSTONE_BIT; }
+#else
+inline PolyObject *OBJ_GET_POINTER(POLYUNSIGNED L) { return (PolyObject*)(( L & ~_OBJ_TOMBSTONE_BIT) <<2); }
+inline POLYUNSIGNED OBJ_SET_POINTER(PolyObject *pt) { return ((POLYUNSIGNED)pt >> 2) | _OBJ_TOMBSTONE_BIT; }
+#endif
 
 // An object i.e. a piece of allocated memory in the heap.  In the simplest case this is a
 // tuple, a list cons cell, a string or a ref.  Every object has a length word in the word before
@@ -296,6 +319,7 @@ public:
 
     bool IsByteObject(void) const { return OBJ_IS_BYTE_OBJECT(LengthWord()); }
     bool IsCodeObject(void) const { return OBJ_IS_CODE_OBJECT(LengthWord()); }
+    bool IsClosureObject(void) const { return OBJ_IS_CLOSURE_OBJECT(LengthWord()); }
     bool IsWordObject(void) const { return OBJ_IS_WORD_OBJECT(LengthWord()); }
     bool IsMutable(void) const { return OBJ_IS_MUTABLE_OBJECT(LengthWord()); }
     bool IsWeakRefObject(void) const { return OBJ_IS_WEAKREF_OBJECT(LengthWord()); }
@@ -305,10 +329,6 @@ public:
     PolyObject *GetForwardingPtr(void) const { return OBJ_GET_POINTER(LengthWord()); }
     void SetForwardingPtr(PolyObject *newp) { ((PolyWord*)this)[-1] = PolyWord::FromUnsigned(OBJ_SET_POINTER(newp)); }
 
-    bool ContainsShareChain(void) const { return OBJ_IS_CHAINED(LengthWord()); }
-    PolyObject *GetShareChain(void) const { return OBJ_GET_CHAIN(LengthWord()); }
-    void SetShareChain(PolyObject *newp) { ((PolyWord*)this)[-1] = PolyWord::FromUnsigned(OBJ_SET_CHAIN(newp)); }
-
     bool ContainsNormalLengthWord(void) const { return OBJ_IS_LENGTH(LengthWord()); }
 
     // Find the start of the constant section for a piece of code.
@@ -374,16 +394,6 @@ public:
 
 typedef PolyException poly_exn;
 
-/* 
- * Stream tokens are pointers to UNTAGGED C integers 
- * (stored in byte objects) 
- */
-class StreamToken: public PolyObject
-{
-public:
-    POLYUNSIGNED streamNo;
-};
-
 /* Macro to round a number of bytes up to a number of words. */
 #define WORDS(s) ((s+sizeof(PolyWord)-1)/sizeof(PolyWord))
 
