@@ -124,27 +124,30 @@ HeapSizeParameters::HeapSizeParameters()
 #define K_to_words(k) ROUNDUP((k) * (1024 / sizeof(PolyWord)),BITSPERWORD)
 
 // Returns physical memory size in bytes
-static POLYUNSIGNED GetPhysicalMemorySize(void);
+static size_t GetPhysicalMemorySize(void);
 
 // These are the maximum values for the number of words.
 #if (SIZEOF_VOIDP == 4)
-#   define MAXIMUMADDRESS   0x3fffffff
+#   define MAXIMUMADDRESS   0x3fffffff /* 4Gbytes as words */
+#elif defined(POLYML32IN64)
+#   define MAXIMUMADDRESS   0xffffffff /* 16Gbytes as words */
 #else
 #   define MAXIMUMADDRESS   0x1fffffffffffffff
 #endif
 
 // Set the initial size based on any parameters specified on the command line.
 // Any of these can be zero indicating they should default.
-void HeapSizeParameters::SetHeapParameters(POLYUNSIGNED minsize, POLYUNSIGNED maxsize, POLYUNSIGNED initialsize, unsigned percent)
+void HeapSizeParameters::SetHeapParameters(uintptr_t minsize, uintptr_t maxsize, uintptr_t initialsize, unsigned percent)
 {
     minHeapSize = K_to_words(minsize); // If these overflow assume the result will be zero
     maxHeapSize = K_to_words(maxsize);
-    POLYUNSIGNED initialSize = K_to_words(initialsize);
+    uintptr_t initialSize = K_to_words(initialsize);
 
-    POLYUNSIGNED memsize = GetPhysicalMemorySize() / sizeof(PolyWord);
+    uintptr_t memsize = GetPhysicalMemorySize() / sizeof(PolyWord);
 
     // If no maximum is given default it to 80% of the physical memory.
     // This allows some space for the OS and other things.
+    // We now check maxsize so it should never exceed the maximum.
     if (maxHeapSize == 0 || maxHeapSize > MAXIMUMADDRESS)
     {
         if (memsize != 0)
@@ -200,7 +203,7 @@ void HeapSizeParameters::SetHeapParameters(POLYUNSIGNED minsize, POLYUNSIGNED ma
     }
 }
 
-void HeapSizeParameters::SetReservation(POLYUNSIGNED rsize)
+void HeapSizeParameters::SetReservation(uintptr_t rsize)
 {
     gMem.SetReservation(K_to_words(rsize));
 }
@@ -208,14 +211,21 @@ void HeapSizeParameters::SetReservation(POLYUNSIGNED rsize)
 // Called in the minor GC if a GC thread needs to grow the heap.
 // Returns zero if the heap cannot be grown. "space" is the space required for the
 // object (and length field) in case this is larger than the default size.
-LocalMemSpace *HeapSizeParameters::AddSpaceInMinorGC(POLYUNSIGNED space, bool isMutable)
+LocalMemSpace *HeapSizeParameters::AddSpaceInMinorGC(uintptr_t space, bool isMutable)
 {
     // See how much space is allocated to the major heap.
-    POLYUNSIGNED spaceAllocated = gMem.CurrentHeapSize() - gMem.CurrentAllocSpace();
+    uintptr_t spaceAllocated = gMem.CurrentHeapSize() - gMem.CurrentAllocSpace();
 
     // The new segment is either the default size or as large as
     // necessary for the object.
-    POLYUNSIGNED spaceSize = gMem.DefaultSpaceSize();
+    uintptr_t spaceSize = gMem.DefaultSpaceSize();
+#ifdef POLYML32IN64
+    // When we allocate a space in NewLocalSpace we take one word to ensure
+    // the that the first length word is on an odd-word boundary.
+    // We need to add one here to ensure there is sufficient space to do that.
+    // See AllocHeapSpace
+    space++;
+#endif
     if (space > spaceSize) spaceSize = space;
 
     // We allow for extension if the total heap size after extending it
@@ -260,7 +270,7 @@ LocalMemSpace *HeapSizeParameters::AddSpaceBeforeCopyPhase(bool isMutable)
 // Growing the heap is just a matter of adjusting the limits.  We
 // don't actually need to allocate the space here.
 // See also adjustHeapSizeAfterMinorGC for adjustments after a minor GC.
-void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
+void HeapSizeParameters::AdjustSizeAfterMajorGC(uintptr_t wordsRequired)
 {
     // Cumulative times since the last major GC
     TIMEDATA gc, nonGc;
@@ -271,7 +281,7 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
 
     if (highWaterMark < heapSizeAtStart) highWaterMark = heapSizeAtStart;
 
-    POLYUNSIGNED heapSpace = gMem.SpaceForHeap() < highWaterMark ? gMem.SpaceForHeap() : highWaterMark;
+    uintptr_t heapSpace = gMem.SpaceForHeap() < highWaterMark ? gMem.SpaceForHeap() : highWaterMark;
     currentSpaceUsed = wordsRequired;
     for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
@@ -290,7 +300,7 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
     if (performSharingPass)
     {
         // We ran the sharing pass last time: calculate the actual recovery rate.
-        POLYUNSIGNED originalSpaceUsed = currentSpaceUsed + sharingWordsRecovered;
+        uintptr_t originalSpaceUsed = currentSpaceUsed + sharingWordsRecovered;
         sharingRecoveryRate = (double)sharingWordsRecovered / (double)originalSpaceUsed;
         if (debugOptions & DEBUG_HEAPSIZE)
             Log("Heap: Sharing recovery rate was %0.3f and cost %0.3f seconds (%0.3f%% of total).\n",
@@ -307,7 +317,7 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
 
     if (debugOptions & DEBUG_HEAPSIZE)
     {
-        POLYUNSIGNED currentFreeSpace = currentSpaceUsed < heapSpace ? 0: heapSpace - currentSpaceUsed;
+        uintptr_t currentFreeSpace = currentSpaceUsed < heapSpace ? 0: heapSpace - currentSpaceUsed;
         Log("Heap: GC cpu time %2.3f non-gc time %2.3f ratio %0.3f for free space ",
             gc.toSeconds(), nonGc.toSeconds(), lastMajorGCRatio);
         LogSize((lastFreeSpace + currentFreeSpace)/2);
@@ -325,7 +335,7 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
         // compute the paging contribution in the cost function.
         double scaleFactor = 1.0 + log((double)majorGCPageFaults / PAGINGCOUNTFACTOR) / PAGINGCOSTSTEEPNESS;
         ASSERT(scaleFactor > 0.0);
-        POLYUNSIGNED newLimit = (POLYUNSIGNED)((double)heapSpace / scaleFactor);
+        uintptr_t newLimit = (uintptr_t)((double)heapSpace / scaleFactor);
         if (pagingLimitSize == 0)
             pagingLimitSize = newLimit;
         else 
@@ -349,7 +359,7 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
     }
 
     // Calculate the new heap size and the predicted cost.
-    POLYUNSIGNED newHeapSize;
+    uintptr_t newHeapSize;
     double cost;
     bool atTarget = getCostAndSize(newHeapSize, cost, false);
     // If we have been unable to allocate any more memory we may already
@@ -368,7 +378,7 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
     }
     else
     {
-        POLYUNSIGNED newHeapSizeWithSharing;
+        uintptr_t newHeapSizeWithSharing;
         double costWithSharing;
         // Get the cost and heap size if sharing was enabled.  If we are at the
         // limit, though, we need to work using the size we can achieve.
@@ -428,7 +438,7 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
     // Also, if we use the whole of the heap we may not then be able to allocate
     // new areas in the major heap without going over the limit.  Restrict it to
     // half of the available heap.
-    POLYUNSIGNED nextLimit = highWaterMark + highWaterMark / 32;
+    uintptr_t nextLimit = highWaterMark + highWaterMark / 32;
     if (nextLimit > newHeapSize) nextLimit = newHeapSize;
     // gMem.CurrentHeapSize() is the live space size.
     if (gMem.CurrentHeapSize() > nextLimit)
@@ -441,18 +451,19 @@ void HeapSizeParameters::AdjustSizeAfterMajorGC(POLYUNSIGNED wordsRequired)
 
 // Called after a minor GC.  Currently does nothing.
 // See also adjustHeapSize for adjustments after a major GC.
-bool HeapSizeParameters::AdjustSizeAfterMinorGC(POLYUNSIGNED spaceAfterGC, POLYUNSIGNED spaceBeforeGC)
+bool HeapSizeParameters::AdjustSizeAfterMinorGC(uintptr_t spaceAfterGC, uintptr_t spaceBeforeGC)
 {
-    POLYUNSIGNED spaceCopiedOut = spaceAfterGC-spaceBeforeGC;
-    TIMEDATA gc, total;
+    uintptr_t spaceCopiedOut = spaceAfterGC-spaceBeforeGC;
+    TIMEDATA gc, nonGC;
     minorGCsSinceMajor++;
-    // The major costs are cumulative so we use those
-    gc.add(majorGCSystemCPU);
-    gc.add(majorGCUserCPU);
-    total.add(gc);
-    total.add(majorNonGCSystemCPU);
-    total.add(majorNonGCUserCPU);
-    float g = gc.toSeconds() / total.toSeconds();
+    // Work out the ratio of GC to non-GC time since the last major GC. We can only adjust the heap size
+    // at a major GC.  If we're now spending too much time in minor GCs we may need a major GC to adjust
+    // the size.
+    gc.add(minorGCSystemCPU);
+    gc.add(minorGCUserCPU);
+    nonGC.add(minorNonGCSystemCPU);
+    nonGC.add(minorNonGCUserCPU);
+    float g = gc.toSeconds() / nonGC.toSeconds();
 
     if (debugOptions & DEBUG_HEAPSIZE)
     {
@@ -467,7 +478,7 @@ bool HeapSizeParameters::AdjustSizeAfterMinorGC(POLYUNSIGNED spaceAfterGC, POLYU
 
     if (highWaterMark < gMem.CurrentHeapSize()) highWaterMark = gMem.CurrentHeapSize();
 
-    POLYUNSIGNED nextLimit = highWaterMark + highWaterMark / 32;
+    uintptr_t nextLimit = highWaterMark + highWaterMark / 32;
     if (nextLimit > gMem.SpaceForHeap()) nextLimit = gMem.SpaceForHeap();
 
     // Set the space available for the allocation area to be the difference between the
@@ -475,15 +486,15 @@ bool HeapSizeParameters::AdjustSizeAfterMinorGC(POLYUNSIGNED spaceAfterGC, POLYU
     // on this GC.  That allows for the next minor GC to copy the same amount without
     // extending the heap.  If the next minor GC adds more than this the heap will be
     // extended and a corresponding amount deducted so that the heap shrinks again.
-    POLYUNSIGNED currHeap = gMem.CurrentHeapSize();
-    POLYUNSIGNED currAlloc = gMem.CurrentAllocSpace();
-    POLYUNSIGNED nonAlloc = currHeap - currAlloc + spaceCopiedOut;
+    uintptr_t currHeap = gMem.CurrentHeapSize();
+    uintptr_t currAlloc = gMem.CurrentAllocSpace();
+    uintptr_t nonAlloc = currHeap - currAlloc + spaceCopiedOut;
     // TODO: If we have limited the space to the high water mark + 1/32 but that is less
     // than we really need we should increase it further.
-    POLYUNSIGNED allowedAlloc = nonAlloc >= nextLimit ? 0 : nextLimit - nonAlloc;
+    uintptr_t allowedAlloc = nonAlloc >= nextLimit ? 0 : nextLimit - nonAlloc;
     // Normally the allocation area will be empty but if we've failed to copy
     // everything out, especially a big object, it may not be.
-    POLYUNSIGNED allocatedInAlloc = gMem.AllocatedInAlloc();
+    uintptr_t allocatedInAlloc = gMem.AllocatedInAlloc();
 
     // If we hit the limit at the last major GC we have to be much more careful.
     // If the minor GC cannot allocate a major GC space when it needs it the minor
@@ -517,18 +528,18 @@ bool HeapSizeParameters::AdjustSizeAfterMinorGC(POLYUNSIGNED spaceAfterGC, POLYU
 // Estimate the GC cost for a given heap size.  The result is the ratio of
 // GC time to application time.
 // This is really guesswork.
-double HeapSizeParameters::costFunction(POLYUNSIGNED heapSize, bool withSharing, bool withSharingCost)
+double HeapSizeParameters::costFunction(uintptr_t heapSize, bool withSharing, bool withSharingCost)
 {
-    POLYUNSIGNED heapSpace = gMem.SpaceForHeap() < highWaterMark ? gMem.SpaceForHeap() : highWaterMark;
-    POLYUNSIGNED currentFreeSpace = heapSpace < currentSpaceUsed ? 0: heapSpace - currentSpaceUsed;
-    POLYUNSIGNED averageFree = (lastFreeSpace + currentFreeSpace) / 2;
-    POLYUNSIGNED spaceUsed = currentSpaceUsed; // N.B.  currentSpaceUsed includes the new space we want
+    uintptr_t heapSpace = gMem.SpaceForHeap() < highWaterMark ? gMem.SpaceForHeap() : highWaterMark;
+    uintptr_t currentFreeSpace = heapSpace < currentSpaceUsed ? 0: heapSpace - currentSpaceUsed;
+    uintptr_t averageFree = (lastFreeSpace + currentFreeSpace) / 2;
+    uintptr_t spaceUsed = currentSpaceUsed; // N.B.  currentSpaceUsed includes the new space we want
     if (heapSize <= currentSpaceUsed)
         return 1.0E6;
     // If we run the sharing pass the live space will be smaller.
     if (withSharing)
-        spaceUsed -= (POLYUNSIGNED)((double)currentSpaceUsed * sharingRecoveryRate);
-    POLYUNSIGNED estimatedFree = heapSize - spaceUsed;
+        spaceUsed -= (uintptr_t)((double)currentSpaceUsed * sharingRecoveryRate);
+    uintptr_t estimatedFree = heapSize - spaceUsed;
     // The cost scales as the inverse of the amount of free space.
     double result = lastMajorGCRatio * (double)averageFree / (double)estimatedFree;
     // If we run the sharing pass the GC cost will increase.
@@ -557,21 +568,21 @@ double HeapSizeParameters::costFunction(POLYUNSIGNED heapSize, bool withSharing,
 // the user GC ratio and false if we minimised the cost
 // TODO: This could definitely be improved although it's not likely to contribute much to
 // the overall cost of a GC.
-bool HeapSizeParameters::getCostAndSize(POLYUNSIGNED &heapSize, double &cost, bool withSharing)
+bool HeapSizeParameters::getCostAndSize(uintptr_t &heapSize, double &cost, bool withSharing)
 {
     bool isBounded = false;
-    POLYUNSIGNED heapSpace = gMem.SpaceForHeap() < highWaterMark ? gMem.SpaceForHeap() : highWaterMark;
+    uintptr_t heapSpace = gMem.SpaceForHeap() < highWaterMark ? gMem.SpaceForHeap() : highWaterMark;
     // Calculate a new heap size.  We allow a maximum doubling or halving of size.
     // It's probably more important to limit the increase in case we hit paging.
-    POLYUNSIGNED sizeMax = heapSpace * 2;
+    uintptr_t sizeMax = heapSpace * 2;
     if (sizeMax > maxHeapSize) sizeMax = maxHeapSize;
-    POLYUNSIGNED sizeMin = heapSpace / 2;
+    uintptr_t sizeMin = heapSpace / 2;
     if (sizeMin < minHeapSize) sizeMin = minHeapSize;
     // We mustn't reduce the heap size too far.  If the application does a lot
     // of work with few allocations and particularly if it calls PolyML.fullGC
     // explicitly we could attempt to shrink the heap below the current live data size.
     // Add 3*space size here.  We require 2* after a minor GC. Add 1 for rounding.
-    POLYUNSIGNED minForAllocation = gMem.CurrentHeapSize() + gMem.DefaultSpaceSize() * 3;
+    uintptr_t minForAllocation = gMem.CurrentHeapSize() + gMem.DefaultSpaceSize() * 3;
     if (minForAllocation > maxHeapSize) minForAllocation = maxHeapSize;
     if (sizeMin < minForAllocation) sizeMin = minForAllocation;
 
@@ -585,7 +596,7 @@ bool HeapSizeParameters::getCostAndSize(POLYUNSIGNED &heapSize, double &cost, bo
         double costMax = costFunction(sizeMax, withSharing, true);
         while (sizeMax > sizeMin + gMem.DefaultSpaceSize())
         {
-            POLYUNSIGNED sizeNext = (sizeMin + sizeMax) / 2;
+            uintptr_t sizeNext = (sizeMin + sizeMax) / 2;
             double cost = costFunction(sizeNext, withSharing, true);
             if (cost < userGCRatio)
                 isBounded = true;
@@ -623,7 +634,7 @@ bool HeapSizeParameters::RunMajorGCImmediately()
 
 static bool GetLastStats(TIMEDATA &userTime, TIMEDATA &systemTime, TIMEDATA &realTime, long &pageCount)
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     FILETIME kt, ut;
     FILETIME ct, et; // Unused
     FILETIME rt;
@@ -752,7 +763,7 @@ void HeapSizeParameters::RecordGCTime(gcTime isEnd, const char *stage)
             minorGCPageFaults += pageCount - startPF;
             majorGCPageFaults += pageCount - startPF;
             startPF = pageCount;
-            globalStats.copyGCTimes(totalGCUserCPU, totalGCSystemCPU);
+            globalStats.copyGCTimes(totalGCUserCPU, totalGCSystemCPU, totalGCReal);
         }
         break;
     }
@@ -776,7 +787,7 @@ void HeapSizeParameters::RecordSharingData(POLYUNSIGNED recovery)
 
 Handle HeapSizeParameters::getGCUtime(TaskData *taskData) const
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     return Make_arb_from_Filetime(taskData, totalGCUserCPU);
 #else
     return Make_arb_from_pair_scaled(taskData, ((struct timeval)totalGCUserCPU).tv_sec, ((struct timeval)totalGCUserCPU).tv_usec, 1000000);
@@ -785,7 +796,7 @@ Handle HeapSizeParameters::getGCUtime(TaskData *taskData) const
 
 Handle HeapSizeParameters::getGCStime(TaskData *taskData) const
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     return Make_arb_from_Filetime(taskData, totalGCSystemCPU);
 #else
     return Make_arb_from_pair_scaled(taskData, ((struct timeval)totalGCSystemCPU).tv_sec, ((struct timeval)totalGCSystemCPU).tv_usec, 1000000);
@@ -794,7 +805,7 @@ Handle HeapSizeParameters::getGCStime(TaskData *taskData) const
 
 void HeapSizeParameters::Init()
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     // Record an initial time of day to use as the basis of real timing
     FILETIME s;
     GetSystemTimeAsFileTime(&s);
@@ -806,7 +817,7 @@ void HeapSizeParameters::Init()
     startRTime = startTime; // Start of this non-gc phase
 
     resetMajorTimingData();
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     startPF = GetPaging(0);
 #else
     startPF = GetPaging(0);
@@ -819,7 +830,7 @@ void HeapSizeParameters::Final()
     if (debugOptions & (DEBUG_GC|DEBUG_HEAPSIZE))
     {
         TIMEDATA userTime, systemTime, realTime;
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
         FILETIME kt, ut;
         FILETIME ct, et; // Unused
         FILETIME rt;
@@ -907,10 +918,10 @@ void HeapSizing::Stop()
     gHeapSizeParameters.Final();
 }
 
-static POLYUNSIGNED GetPhysicalMemorySize(void)
+static size_t GetPhysicalMemorySize(void)
 {
-    POLYUNSIGNED maxMem = 0-1; // Maximum unsigned value.
-#if defined(HAVE_WINDOWS_H)
+    size_t maxMem = (size_t)0-1; // Maximum unsigned value.
+#if defined(HAVE_WINDOWS_H) // Windows including Cygwin
     {
         MEMORYSTATUSEX memStatEx;
         memset(&memStatEx, 0, sizeof(memStatEx));
@@ -923,7 +934,7 @@ static POLYUNSIGNED GetPhysicalMemorySize(void)
             if (memStatEx.ullTotalPhys > dwlMax)
                 return maxMem;
             else
-                return (POLYUNSIGNED)memStatEx.ullTotalPhys;
+                return (size_t)memStatEx.ullTotalPhys;
         }
     }
 
@@ -956,7 +967,7 @@ static POLYUNSIGNED GetPhysicalMemorySize(void)
             if (physMem > (uint64_t)maxMem)
                 return maxMem;
             else
-                return (POLYUNSIGNED)physMem;
+                return (size_t)physMem;
         }
     }
 #endif
