@@ -2,7 +2,7 @@
     Title:     Write out a database as a Mach object file
     Author:    David Matthews.
 
-    Copyright (c) 2006-7, 2011-2, 2016 David C. J. Matthews
+    Copyright (c) 2006-7, 2011-2, 2016-18 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -55,7 +55,6 @@
 #include <mach-o/loader.h>
 #include <mach-o/reloc.h>
 #include <mach-o/nlist.h>
-#include <mach-o/ppc/reloc.h>
 #include <mach-o/x86_64/reloc.h>
 
 #ifdef HAVE_STRING_H
@@ -81,7 +80,7 @@
 
 // Mach-O seems to require each section to have a discrete virtual address range
 // so we have to adjust various offsets to fit.
-void MachoExport::adjustOffset(unsigned area, POLYUNSIGNED &offset)
+void MachoExport::adjustOffset(unsigned area, size_t &offset)
 {
     // Add in the offset.  If sect is memTableEntries it's actually the
     // descriptors so doesn't have any additional offset.
@@ -93,7 +92,7 @@ void MachoExport::adjustOffset(unsigned area, POLYUNSIGNED &offset)
     }
 }
 
-void MachoExport::addExternalReference(void *relocAddr, const char *name)
+void MachoExport::addExternalReference(void *relocAddr, const char *name, bool /*isFuncPtr*/)
 {
     externTable.makeEntry(name);
     writeRelocation(0, relocAddr, symbolNum++, true);
@@ -103,7 +102,7 @@ void MachoExport::addExternalReference(void *relocAddr, const char *name)
 void MachoExport::setRelocationAddress(void *p, int32_t *reloc)
 {
     unsigned area = findArea(p);
-    POLYUNSIGNED offset = (char*)p - (char*)memTable[area].mtAddr;
+    size_t offset = (char*)p - (char*)memTable[area].mtOriginalAddr;
     *reloc = offset;
 }
 
@@ -112,7 +111,7 @@ PolyWord MachoExport::createRelocation(PolyWord p, void *relocAddr)
 {
     void *addr = p.AsAddress();
     unsigned addrArea = findArea(addr);
-    POLYUNSIGNED offset = (char*)addr - (char*)memTable[addrArea].mtAddr;
+    size_t offset = (char*)addr - (char*)memTable[addrArea].mtOriginalAddr;
     adjustOffset(addrArea, offset);
     return writeRelocation(offset, relocAddr, addrArea+1 /* Sections count from 1 */, false);
 }
@@ -144,16 +143,17 @@ PolyWord MachoExport::writeRelocation(POLYUNSIGNED offset, void *relocAddr, unsi
    that the offset is saved in original word. */
 void MachoExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind code)
 {
-    PolyWord p = GetConstantValue(addr, code);
+#ifndef POLYML32IN64
+    PolyObject *p = GetConstantValue(addr, code);
 
-    if (IS_INT(p) || p == PolyWord::FromUnsigned(0))
+    if (p == 0)
         return;
 
-    void *a = p.AsAddress();
+    void *a = p;
     unsigned aArea = findArea(a);
 
     // Set the value at the address to the offset relative to the symbol.
-    POLYUNSIGNED offset = (char*)a - (char*)memTable[aArea].mtAddr;
+    size_t offset = (char*)a - (char*)memTable[aArea].mtOriginalAddr;
     adjustOffset(aArea, offset);
 
     switch (code)
@@ -204,7 +204,7 @@ void MachoExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind
                 fwrite(&reloc, sizeof(reloc), 1, exportFile);
                 relocationCount++;
 
-                POLYUNSIGNED addrOffset = (char*)addr - (char*)memTable[addrArea].mtAddr;
+                size_t addrOffset = (char*)addr - (char*)memTable[addrArea].mtOriginalAddr;
                 adjustOffset(addrArea, addrOffset);
                 offset -= addrOffset + 4;
 
@@ -220,6 +220,7 @@ void MachoExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind
         default:
             ASSERT(0); // Wrong type of relocation for this architecture.
     }
+#endif
 }
 
 // Set the file alignment.
@@ -231,7 +232,7 @@ void MachoExport::alignFile(int align)
     fwrite(&pad, align - (offset % align), 1, exportFile);
 }
 
-void MachoExport::createStructsRelocation(unsigned sect, POLYUNSIGNED offset)
+void MachoExport::createStructsRelocation(unsigned sect, size_t offset)
 {
     struct relocation_info reloc;
     reloc.r_address = offset;
@@ -386,7 +387,7 @@ void MachoExport::exportStore(void)
         sections[i].reloff = ftell(exportFile);
         relocationCount = 0;
         // Create the relocation table and turn all addresses into offsets.
-        char *start = (char*)memTable[i].mtAddr;
+        char *start = (char*)memTable[i].mtOriginalAddr;
         char *end = start + memTable[i].mtLength;
         for (p = (PolyWord*)start; p < (PolyWord*)end; )
         {
@@ -411,7 +412,7 @@ void MachoExport::exportStore(void)
 
     // Address of "rootFunction" within "exports"
     unsigned rootAddrArea = findArea(rootFunction);
-    POLYUNSIGNED rootOffset = (char*)rootFunction - (char*)memTable[rootAddrArea].mtAddr;
+    size_t rootOffset = (char*)rootFunction - (char*)memTable[rootAddrArea].mtOriginalAddr;
     adjustOffset(rootAddrArea, rootOffset);
     createStructsRelocation(rootAddrArea, offsetof(exportDescription, rootFunction));
 
@@ -419,7 +420,7 @@ void MachoExport::exportStore(void)
     for (i = 0; i < memTableEntries; i++)
     {
         createStructsRelocation(i,
-            sizeof(exportDescription) + i * sizeof(memoryTableEntry) + offsetof(memoryTableEntry, mtAddr));
+            sizeof(exportDescription) + i * sizeof(memoryTableEntry) + offsetof(memoryTableEntry, mtCurrentAddr));
     }
     sections[memTableEntries].nreloc = relocationCount;
 
@@ -483,17 +484,22 @@ void MachoExport::exportStore(void)
     exports.timeStamp = getBuildTime();
     exports.architecture = machineDependent->MachineArchitecture();
     exports.rtsVersion = POLY_version_number;
+#ifdef POLYML32IN64
+    exports.originalBaseAddr = globalHeapBase;
+#else
+    exports.originalBaseAddr = 0;
+#endif
 
     sections[memTableEntries].offset = ftell(exportFile);
     fwrite(&exports, sizeof(exports), 1, exportFile);
-    POLYUNSIGNED addrOffset = sizeof(exports)+sizeof(memoryTableEntry)*memTableEntries;
+    size_t addrOffset = sizeof(exports)+sizeof(memoryTableEntry)*memTableEntries;
     for (i = 0; i < memTableEntries; i++)
     {
-        void *save = memTable[i].mtAddr;
-        memTable[i].mtAddr = (void*)addrOffset; // Set this to the relative address.
+        void *save = memTable[i].mtCurrentAddr;
+        memTable[i].mtCurrentAddr = (void*)addrOffset; // Set this to the relative address.
         addrOffset += memTable[i].mtLength;
         fwrite(&memTable[i], sizeof(memoryTableEntry), 1, exportFile);
-        memTable[i].mtAddr = save;
+        memTable[i].mtCurrentAddr = save;
     }
 
     // Now the binary data.
@@ -501,7 +507,7 @@ void MachoExport::exportStore(void)
     {
         alignFile(4);
         sections[i].offset = ftell(exportFile);
-        fwrite(memTable[i].mtAddr, 1, memTable[i].mtLength, exportFile);
+        fwrite(memTable[i].mtOriginalAddr, 1, memTable[i].mtLength, exportFile);
     }
     // Rewind to rewrite the headers with the actual offsets.
     rewind(exportFile);
