@@ -2,7 +2,7 @@
     Title:     Export memory as a Mach object file
     Author:    David C. J. Matthews.
 
-    Copyright (c) 2006,2016 David C. J. Matthews
+    Copyright (c) 2006,2016-18 David C. J. Matthews
 
 
     This library is free software; you can redistribute it and/or
@@ -41,15 +41,15 @@ private:
     virtual void ScanConstant(PolyObject *base, byte *addrOfConst, ScanRelocationKind code);
     // At the moment we should only get calls to ScanConstant.
     virtual PolyObject *ScanObjectAddress(PolyObject *base) { return base; }
-    virtual void addExternalReference(void *addr, const char *name);
+    virtual void addExternalReference(void *addr, const char *name, bool isFuncPtr);
 
 private:
     void setRelocationAddress(void *p, int32_t *reloc);
     PolyWord createRelocation(PolyWord p, void *relocAddr);
     PolyWord writeRelocation(POLYUNSIGNED offset, void *relocAddr, unsigned symbolNumber, bool isExtern);
     void alignFile(int align);
-    void createStructsRelocation(unsigned area, POLYUNSIGNED offset);
-    void adjustOffset(unsigned area, POLYUNSIGNED &offset);
+    void createStructsRelocation(unsigned area, size_t offset);
+    void adjustOffset(unsigned area, size_t &offset);
 
     unsigned relocationCount;
     ExportStringTable stringTable;
