@@ -1,7 +1,7 @@
 /*
     Title:  memmgr.cpp   Memory segment manager
 
-    Copyright (c) 2006-7, 2011-12, 2016-17 David C. J. Matthews
+    Copyright (c) 2006-7, 2011-12, 2016-18 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -32,6 +32,8 @@
 #define ASSERT(x)
 #endif
 
+#include <stdio.h>
+
 #include <new>
 
 #include "globals.h"
@@ -44,30 +46,36 @@
 #include "statistics.h"
 #include "processes.h"
 
+
+#ifdef POLYML32IN64
+// This contains the address of the base of the heap.
+PolyWord *globalHeapBase, *globalCodeBase;
+#endif
+
 // heap resizing policy option requested on command line
 unsigned heapsizingOption = 0;
 
-MemSpace::MemSpace(): SpaceTree(true)
+MemSpace::MemSpace(OSMem *alloc): SpaceTree(true)
 {
     spaceType = ST_PERMANENT;
     isMutable = false;
     bottom = 0;
     top = 0;
-    isOwnSpace = false;
     isCode = false;
+    allocator = alloc;
 }
 
 MemSpace::~MemSpace()
 {
-    if (isOwnSpace && bottom != 0)
-        osMemoryManager->Free(bottom, (char*)top - (char*)bottom);
+    if (allocator != 0 && bottom != 0)
+        allocator->Free(bottom, (char*)top - (char*)bottom);
 }
 
-MarkableSpace::MarkableSpace(): spaceLock("Local space")
+MarkableSpace::MarkableSpace(OSMem *alloc): MemSpace(alloc), spaceLock("Local space")
 {
 }
 
-LocalMemSpace::LocalMemSpace()
+LocalMemSpace::LocalMemSpace(OSMem *alloc): MarkableSpace(alloc)
 {
     spaceType = ST_LOCAL;
     upperAllocPtr = lowerAllocPtr = 0;
@@ -78,27 +86,21 @@ LocalMemSpace::LocalMemSpace()
     allocationSpace = false;
 }
 
-bool LocalMemSpace::InitSpace(POLYUNSIGNED size, bool mut)
+bool LocalMemSpace::InitSpace(PolyWord *heapSpace, uintptr_t size, bool mut)
 {
     isMutable = mut;
-
-    // Allocate the heap itself.
-    size_t iSpace = size*sizeof(PolyWord);
-    bottom  =
-        (PolyWord*)osMemoryManager->Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE);
-
-    if (bottom == 0)
-        return false;
-    isOwnSpace = true; // Deallocate when we're finished.
-
-    // The size may have been rounded up to a block boundary.
-    size = iSpace/sizeof(PolyWord);
-
+    bottom = heapSpace;
     top = bottom + size;
     // Initialise all the fields.  The partial GC in particular relies on this.
     upperAllocPtr = partialGCTop = fullGCRescanStart = fullGCLowerLimit = lowestWeak = top;
     lowerAllocPtr = partialGCScan = partialGCRootBase = partialGCRootTop =
         fullGCRescanEnd = highestWeak = bottom;
+#ifdef POLYML32IN64
+    // The address must be on an odd-word boundary so that after the length
+    // word is put in the actual cell address is on an even-word boundary.
+    lowerAllocPtr[0] = PolyWord::FromUnsigned(0);
+    lowerAllocPtr = bottom + 1;
+#endif
     spaceOwner = 0;
 
     allocationSpace = false;
@@ -135,11 +137,37 @@ MemMgr::~MemMgr()
         delete(*i);
 }
 
+bool MemMgr::Initialise()
+{
+#ifdef POLYML32IN64
+    // Allocate a single 16G area but with no access.
+    void *heapBase;
+    if (!osHeapAlloc.Initialise((size_t)16 * 1024 * 1024 * 1024, &heapBase))
+        return false;
+    globalHeapBase = (PolyWord*)heapBase;
+
+    // Allocate a 4 gbyte area for the stacks.
+    // It's important that the stack and code areas have addresses with
+    // non-zero top 32-bits.
+    if (!osStackAlloc.Initialise((size_t)4 * 1024 * 1024 * 1024))
+        return false;
+
+    // Allocate a 2G area for the code.
+    void *codeBase;
+    if (!osCodeAlloc.Initialise((size_t)2 * 1024 * 1024 * 1024, &codeBase))
+        return false;
+    globalCodeBase = (PolyWord*)codeBase;
+    return true;
+#else
+    return osHeapAlloc.Initialise() && osStackAlloc.Initialise() && osCodeAlloc.Initialise();
+#endif
+}
+
 // Create and initialise a new local space and add it to the table.
-LocalMemSpace* MemMgr::NewLocalSpace(POLYUNSIGNED size, bool mut)
+LocalMemSpace* MemMgr::NewLocalSpace(uintptr_t size, bool mut)
 {
     try {
-        LocalMemSpace *space = new LocalMemSpace;
+        LocalMemSpace *space = new LocalMemSpace(&osHeapAlloc);
         // Before trying to allocate the heap temporarily allocate the
         // reserved space.  This ensures that this much space will always
         // be available for C stacks and the C++ heap.
@@ -147,7 +175,7 @@ LocalMemSpace* MemMgr::NewLocalSpace(POLYUNSIGNED size, bool mut)
         size_t rSpace = reservedSpace*sizeof(PolyWord);
 
         if (reservedSpace != 0) {
-            reservation = osMemoryManager->Allocate(rSpace, PERMISSION_READ);
+            reservation = osHeapAlloc.Allocate(rSpace, PERMISSION_READ);
             if (reservation == 0) {
                 // Insufficient space for the reservation.  Can't allocate this local space.
                 if (debugOptions & DEBUG_MEMMGR)
@@ -157,8 +185,15 @@ LocalMemSpace* MemMgr::NewLocalSpace(POLYUNSIGNED size, bool mut)
             }
         }
 
-        bool success = space->InitSpace(size, mut) && AddLocalSpace(space);
-        if (reservation != 0) osMemoryManager->Free(reservation, rSpace);
+        // Allocate the heap itself.
+        size_t iSpace = size * sizeof(PolyWord);
+        PolyWord *heapSpace =
+            (PolyWord*)osHeapAlloc.Allocate(iSpace, PERMISSION_READ | PERMISSION_WRITE);
+        // The size may have been rounded up to a block boundary.
+        size = iSpace / sizeof(PolyWord);
+        bool success = heapSpace != 0 && space->InitSpace(heapSpace, size, mut) && AddLocalSpace(space);
+
+        if (reservation != 0) osHeapAlloc.Free(reservation, rSpace);
         if (success)
         {
             if (debugOptions & DEBUG_MEMMGR)
@@ -183,7 +218,7 @@ LocalMemSpace* MemMgr::NewLocalSpace(POLYUNSIGNED size, bool mut)
 }
 
 // Create a local space for initial allocation.
-LocalMemSpace *MemMgr::CreateAllocationSpace(POLYUNSIGNED size)
+LocalMemSpace *MemMgr::CreateAllocationSpace(uintptr_t size)
 {
     LocalMemSpace *result = NewLocalSpace(size, true);
     if (result) 
@@ -242,13 +277,12 @@ bool MemMgr::AddLocalSpace(LocalMemSpace *space)
     return true;
 }
 
-
 // Create an entry for a permanent space.
-PermanentMemSpace* MemMgr::NewPermanentSpace(PolyWord *base, POLYUNSIGNED words,
+PermanentMemSpace* MemMgr::NewPermanentSpace(PolyWord *base, uintptr_t words,
                                              unsigned flags, unsigned index, unsigned hierarchy /*= 0*/)
 {
     try {
-        PermanentMemSpace *space = new PermanentMemSpace;
+        PermanentMemSpace *space = new PermanentMemSpace(0/* Not freed */);
         space->bottom = base;
         space->topPointer = space->top = space->bottom + words;
         space->spaceType = ST_PERMANENT;
@@ -277,12 +311,68 @@ PermanentMemSpace* MemMgr::NewPermanentSpace(PolyWord *base, POLYUNSIGNED words,
     }
 }
 
+PermanentMemSpace *MemMgr::AllocateNewPermanentSpace(uintptr_t byteSize, unsigned flags, unsigned index, unsigned hierarchy)
+{
+    try {
+        OSMem *alloc = flags & MTF_EXECUTABLE ? &osCodeAlloc : &osHeapAlloc;
+        PermanentMemSpace *space = new PermanentMemSpace(alloc);
+        unsigned int perms = PERMISSION_READ | PERMISSION_WRITE;
+        if (flags & MTF_EXECUTABLE) perms |= PERMISSION_EXEC;
+        size_t actualSize = byteSize;
+        PolyWord *base = (PolyWord*)alloc->Allocate(actualSize, perms);
+        if (base == 0)
+        {
+            delete(space);
+            return 0;
+        }
+        space->bottom = base;
+        space->topPointer = space->top = space->bottom + actualSize/sizeof(PolyWord);
+        space->spaceType = ST_PERMANENT;
+        space->isMutable = flags & MTF_WRITEABLE ? true : false;
+        space->noOverwrite = flags & MTF_NO_OVERWRITE ? true : false;
+        space->byteOnly = flags & MTF_BYTES ? true : false;
+        space->isCode = flags & MTF_EXECUTABLE ? true : false;
+        space->index = index;
+        space->hierarchy = hierarchy;
+        if (index >= nextIndex) nextIndex = index + 1;
+
+        // Extend the permanent memory table and add this space to it.
+        try {
+            AddTree(space);
+            pSpaces.push_back(space);
+        }
+        catch (std::exception&) {
+            RemoveTree(space);
+            delete space;
+            return 0;
+        }
+        return space;
+    }
+    catch (std::bad_alloc&) {
+        return 0;
+    }
+}
+
+bool MemMgr::CompletePermanentSpaceAllocation(PermanentMemSpace *space)
+{
+    // Remove write access unless it is mutable.
+    // Don't remove write access unless this is top-level. Share-data assumes only hierarchy 0 is write-protected.
+    if (!space->isMutable && space->hierarchy == 0)
+    {
+        if (space->isCode)
+            osCodeAlloc.SetPermissions(space->bottom, (char*)space->top - (char*)space->bottom,
+                PERMISSION_READ | PERMISSION_EXEC);
+        else osHeapAlloc.SetPermissions(space->bottom, (char*)space->top - (char*)space->bottom, PERMISSION_READ);
+    }
+    return true;
+}
+
 // Delete a local space and remove it from the table.
 void MemMgr::DeleteLocalSpace(std::vector<LocalMemSpace*>::iterator &iter)
 {
     LocalMemSpace *sp = *iter;
     if (debugOptions & DEBUG_MEMMGR)
-        Log("MMGR: Deleted local %s space %p\n", sp->spaceTypeString(), sp);
+        Log("MMGR: Deleted local %s space %p at %p size %zu\n", sp->spaceTypeString(), sp, sp->bottom, sp->spaceSize());
     currentHeapSize -= sp->spaceSize();
     globalStats.setSize(PSS_TOTAL_HEAP, currentHeapSize * sizeof(PolyWord));
     if (sp->allocationSpace) currentAllocSpace -= sp->spaceSize();
@@ -298,17 +388,18 @@ void MemMgr::RemoveEmptyLocals()
     for (std::vector<LocalMemSpace*>::iterator i = lSpaces.begin(); i < lSpaces.end(); )
     {
         LocalMemSpace *space = *i;
-        if (space->allocatedSpace() == 0)
+        if (space->isEmpty())
             DeleteLocalSpace(i);
         else i++;
     }
 }
 
 // Create and initialise a new export space and add it to the table.
-PermanentMemSpace* MemMgr::NewExportSpace(POLYUNSIGNED size, bool mut, bool noOv, bool code)
+PermanentMemSpace* MemMgr::NewExportSpace(uintptr_t size, bool mut, bool noOv, bool code)
 {
     try {
-        PermanentMemSpace *space = new PermanentMemSpace;
+        OSMem *alloc = code ? &osCodeAlloc : &osHeapAlloc;
+        PermanentMemSpace *space = new PermanentMemSpace(alloc);
         space->spaceType = ST_EXPORT;
         space->isMutable = mut;
         space->noOverwrite = noOv;
@@ -317,7 +408,7 @@ PermanentMemSpace* MemMgr::NewExportSpace(POLYUNSIGNED size, bool mut, bool noOv
         // Allocate the memory itself.
         size_t iSpace = size*sizeof(PolyWord);
         space->bottom  =
-            (PolyWord*)osMemoryManager->Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE|PERMISSION_EXEC);
+            (PolyWord*)alloc->Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE|PERMISSION_EXEC);
 
         if (space->bottom == 0)
         {
@@ -326,12 +417,17 @@ PermanentMemSpace* MemMgr::NewExportSpace(POLYUNSIGNED size, bool mut, bool noOv
                 Log("MMGR: New export %smutable space: insufficient space\n", mut ? "" : "im");
             return 0;
         }
-        space->isOwnSpace = true;
- 
+
         // The size may have been rounded up to a block boundary.
         size = iSpace/sizeof(PolyWord);
         space->top = space->bottom + size;
         space->topPointer = space->bottom;
+#ifdef POLYML32IN64
+        // The address must be on an odd-word boundary so that after the length
+        // word is put in the actual cell address is on an even-word boundary.
+        space->topPointer[0] = PolyWord::FromUnsigned(0);
+        space->topPointer = space->bottom + 1;
+#endif
 
         if (debugOptions & DEBUG_MEMMGR)
             Log("MMGR: New export %smutable %s%sspace %p, size=%luk words, bottom=%p, top=%p\n", mut ? "" : "im",
@@ -393,7 +489,10 @@ bool MemMgr::PromoteExportSpaces(unsigned hierarchy)
 
                 if (pSpace->isCode)
                 {
-                    CodeSpace *space = new CodeSpace(pSpace->bottom, pSpace->spaceSize());
+                    // Enable write access.  Permanent spaces are read-only.
+                    osCodeAlloc.SetPermissions(pSpace->bottom, (char*)pSpace->top - (char*)pSpace->bottom,
+                        PERMISSION_READ | PERMISSION_WRITE | PERMISSION_EXEC);
+                    CodeSpace *space = new CodeSpace(pSpace->bottom, pSpace->spaceSize(), &osCodeAlloc);
                     if (! space->headerMap.Create(space->spaceSize()))
                     {
                         if (debugOptions & DEBUG_MEMMGR)
@@ -416,24 +515,35 @@ bool MemMgr::PromoteExportSpaces(unsigned hierarchy)
                         // copied to the exported area. Restore the original length word.
                         if (obj->ContainsForwardingPtr())
                         {
+#ifdef POLYML32IN64
+                            PolyObject *forwardedTo = obj;
+                            // This is relative to globalCodeBase not globalHeapBase
+                            while (forwardedTo->ContainsForwardingPtr())
+                                forwardedTo = (PolyObject*)(globalCodeBase + ((forwardedTo->LengthWord() & ~_OBJ_TOMBSTONE_BIT) << 1));
+#else
                             PolyObject *forwardedTo = obj->FollowForwardingChain();
+#endif
                             obj->SetLengthWord(forwardedTo->LengthWord());
                         }
+                        // Set the "start" bit if this is allocated.  It will be a byte seg if not.
                         if (obj->IsCodeObject())
                             space->headerMap.SetBit(ptr-space->bottom);
+                        ASSERT(!obj->IsClosureObject());
                         ptr += obj->Length() + 1;
                     }
                 }
                 else
                 {
-                    LocalMemSpace *space = new LocalMemSpace;
+                    // Enable write access.  Permanent spaces are read-only.
+                    osHeapAlloc.SetPermissions(pSpace->bottom, (char*)pSpace->top - (char*)pSpace->bottom,
+                        PERMISSION_READ | PERMISSION_WRITE);
+                    LocalMemSpace *space = new LocalMemSpace(&osHeapAlloc);
                     space->top = pSpace->top;
                     // Space is allocated in local areas from the top down.  This area is full and
                     // all data is in the old generation.  The area can be recovered by a full GC.
                     space->bottom = space->upperAllocPtr = space->lowerAllocPtr =
                         space->fullGCLowerLimit = pSpace->bottom;
                     space->isMutable = pSpace->isMutable;
-                    space->isOwnSpace = true;
                     space->isCode = false;
                     if (! space->bitmap.Create(space->top-space->bottom) || ! AddLocalSpace(space))
                     {
@@ -441,6 +551,7 @@ bool MemMgr::PromoteExportSpaces(unsigned hierarchy)
                             Log("MMGR: Unable to convert saved state space %p into local space\n", pSpace);
                         return false;
                     }
+
                     if (debugOptions & DEBUG_MEMMGR)
                         Log("MMGR: Converted saved state space %p into local %smutable space %p\n",
                                 pSpace, pSpace->isMutable ? "im": "", space);
@@ -493,16 +604,25 @@ PermanentMemSpace *MemMgr::SpaceForIndex(unsigned index)
 
 // In several places we assume that segments are filled with valid
 // objects.  This fills unused memory with one or more "byte" objects.
-void MemMgr::FillUnusedSpace(PolyWord *base, POLYUNSIGNED words)
+void MemMgr::FillUnusedSpace(PolyWord *base, uintptr_t words)
 {
     PolyWord *pDummy = base+1;
     while (words > 0)
     {
-        POLYUNSIGNED oSize = words;
+#ifdef POLYML32IN64
+        // Make sure that any dummy object we insert is properly aligned.
+        if (((uintptr_t)pDummy) & 4)
+        {
+            *pDummy++ = PolyWord::FromUnsigned(0);
+            words--;
+            continue;
+        }
+#endif
+        POLYUNSIGNED oSize;
         // If the space is larger than the maximum object size
         // we will need several objects.
         if (words > MAX_OBJECT_SIZE) oSize = MAX_OBJECT_SIZE;
-        else oSize = words-1;
+        else oSize = (POLYUNSIGNED)(words-1);
         // Make this a byte object so it's always skipped.
         ((PolyObject*)pDummy)->SetLengthWord(oSize, F_BYTE_OBJ);
         words -= oSize+1;
@@ -514,7 +634,7 @@ void MemMgr::FillUnusedSpace(PolyWord *base, POLYUNSIGNED words)
 // This is used both when allocating single objects (when minWords and maxWords
 // are the same) and when allocating heap segments.  If there is insufficient
 // space to satisfy the minimum it will return 0.
-PolyWord *MemMgr::AllocHeapSpace(POLYUNSIGNED minWords, POLYUNSIGNED &maxWords, bool doAllocation)
+PolyWord *MemMgr::AllocHeapSpace(uintptr_t minWords, uintptr_t &maxWords, bool doAllocation)
 {
     PLocker locker(&allocLock);
     // We try to distribute the allocations between the memory spaces
@@ -531,15 +651,25 @@ PolyWord *MemMgr::AllocHeapSpace(POLYUNSIGNED minWords, POLYUNSIGNED &maxWords,
         LocalMemSpace *space = gMem.lSpaces[j++];
         if (space->allocationSpace)
         {
-            POLYUNSIGNED available = space->freeSpace();
+            uintptr_t available = space->freeSpace();
             if (available > 0 && available >= minWords)
             {
                 // Reduce the maximum value if we had less than that.
-                if (available < maxWords)
-                    maxWords = available;
+                if (available < maxWords) maxWords = available;
+#ifdef POLYML32IN64
+                // If necessary round down to an even boundary
+                if (maxWords & 1)
+                {
+                    maxWords--;
+                    space->lowerAllocPtr[maxWords] = PolyWord::FromUnsigned(0);
+                }
+#endif
                 PolyWord *result = space->lowerAllocPtr; // Return the address.
                 if (doAllocation)
                     space->lowerAllocPtr += maxWords; // Allocate it.
+#ifdef POLYML32IN64
+                ASSERT((uintptr_t)result & 4); // Must be odd-word aligned
+#endif
                 return result;
             }
         }
@@ -557,49 +687,74 @@ PolyWord *MemMgr::AllocHeapSpace(POLYUNSIGNED minWords, POLYUNSIGNED &maxWords,
         // but it may be that allocating this object will take us over the limit.  We allow
         // that to happen so that we can successfully allocate very large objects even if
         // we have a new GC very shortly.
-        POLYUNSIGNED spaceSize = defaultSpaceSize;
+        uintptr_t spaceSize = defaultSpaceSize;
+#ifdef POLYML32IN64
+        // When we create the allocation space we take one word so that the first
+        // length word is on an odd-word boundary.  We need to allow for that otherwise
+        // we may have available < minWords.
+        if (minWords >= spaceSize) spaceSize = minWords+1; // If we really want a large space.
+#else
         if (minWords > spaceSize) spaceSize = minWords; // If we really want a large space.
+#endif
         LocalMemSpace *space = CreateAllocationSpace(spaceSize);
         if (space == 0) return 0; // Can't allocate it
         // Allocate our space in this new area.
-        POLYUNSIGNED available = space->freeSpace();
+        uintptr_t available = space->freeSpace();
         ASSERT(available >= minWords);
         if (available < maxWords)
+        {
             maxWords = available;
+#ifdef POLYML32IN64
+            // If necessary round down to an even boundary
+            if (maxWords & 1)
+            {
+                maxWords--;
+                space->lowerAllocPtr[maxWords] = PolyWord::FromUnsigned(0);
+            }
+#endif
+        }
         PolyWord *result = space->lowerAllocPtr; // Return the address.
         if (doAllocation)
             space->lowerAllocPtr += maxWords; // Allocate it.
+#ifdef POLYML32IN64
+        ASSERT((uintptr_t)result & 4); // Must be odd-word aligned
+#endif
         return result;
     }
     return 0; // There isn't space even for the minimum.
 }
 
-CodeSpace::CodeSpace(PolyWord *start, POLYUNSIGNED spaceSize)
+CodeSpace::CodeSpace(PolyWord *start, uintptr_t spaceSize, OSMem *alloc): MarkableSpace(alloc)
 {
-    isOwnSpace = true;
     bottom = start;
     top = start+spaceSize;
     isMutable = true; // Make it mutable just in case.  This will cause it to be scanned.
-    isOwnSpace = true;
     isCode = true;
     spaceType = ST_CODE;
-    largestFree = spaceSize-1;
+#ifdef POLYML32IN64
+    // Dummy word so that the cell itself, after the length word, is on an 8-byte boundary.
+    *start = PolyWord::FromUnsigned(0);
+    largestFree = spaceSize - 2;
+    firstFree = start+1;
+#else
+    largestFree = spaceSize - 1;
     firstFree = start;
+#endif
 }
 
-CodeSpace *MemMgr::NewCodeSpace(POLYUNSIGNED size)
+CodeSpace *MemMgr::NewCodeSpace(uintptr_t size)
 {
     // Allocate a new area and add it at the end of the table.
     CodeSpace *allocSpace = 0;
     // Allocate a new mutable, code space. N.B.  This may round up "actualSize".
     size_t actualSize = size * sizeof(PolyWord);
     PolyWord *mem =
-        (PolyWord*)osMemoryManager->Allocate(actualSize,
+        (PolyWord*)osCodeAlloc.Allocate(actualSize,
             PERMISSION_READ | PERMISSION_WRITE | PERMISSION_EXEC);
     if (mem != 0)
     {
         try {
-            allocSpace = new CodeSpace(mem, actualSize / sizeof(PolyWord));
+            allocSpace = new CodeSpace(mem, actualSize / sizeof(PolyWord), &osCodeAlloc);
             if (!allocSpace->headerMap.Create(allocSpace->spaceSize()))
             {
                 delete allocSpace;
@@ -613,14 +768,14 @@ CodeSpace *MemMgr::NewCodeSpace(POLYUNSIGNED size)
             else if (debugOptions & DEBUG_MEMMGR)
                 Log("MMGR: New code space %p allocated at %p size %lu\n", allocSpace, allocSpace->bottom, allocSpace->spaceSize());
             // Put in a byte cell to mark the area as unallocated.
-            FillUnusedSpace(allocSpace->bottom, allocSpace->spaceSize());
+            FillUnusedSpace(allocSpace->firstFree, allocSpace->top- allocSpace->firstFree);
         }
         catch (std::bad_alloc&)
         {
         }
         if (allocSpace == 0)
         {
-            osMemoryManager->Free(mem, actualSize);
+            osCodeAlloc.Free(mem, actualSize);
             mem = 0;
         }
     }
@@ -632,12 +787,11 @@ CodeSpace *MemMgr::NewCodeSpace(POLYUNSIGNED size)
 // area executable.  It will not be executed until the mutable bit has been cleared.
 // Once code is allocated it is not GCed or moved.
 // initCell is a byte cell that is copied into the new code area.
-PolyObject*MemMgr::AllocCodeSpace(PolyObject *initCell)
+PolyObject* MemMgr::AllocCodeSpace(POLYUNSIGNED requiredSize)
 {
     PLocker locker(&codeSpaceLock);
     // Search the code spaces until we find a free area big enough.
     size_t i = 0;
-    POLYUNSIGNED requiredSize = initCell->Length();
     while (true)
     {
         if (i != cSpaces.size())
@@ -665,15 +819,23 @@ PolyObject*MemMgr::AllocCodeSpace(PolyObject *initCell)
                         {
                             // Free and large enough
                             PolyWord *next = pt+requiredSize+1;
-                            if (requiredSize < length)
-                                FillUnusedSpace(next, length-requiredSize);
+                            POLYUNSIGNED spare = length - requiredSize;
+#ifdef POLYML32IN64
+                            // Maintain alignment.
+                            if (((requiredSize + 1) & 1) && spare != 0)
+                            {
+                                *next++ = PolyWord::FromUnsigned(0);
+                                spare--;
+                            }
+#endif
+                            if (spare != 0)
+                                FillUnusedSpace(next, spare);
                             space->isMutable = true; // Set this - it ensures the area is scanned on GC.
                             space->headerMap.SetBit(pt-space->bottom); // Set the "header" bit
                             // Set the length word of the code area and copy the byte cell in.
                             // The code bit must be set before the lock is released to ensure
                             // another thread doesn't reuse this.
                             obj->SetLengthWord(requiredSize,  F_CODE_OBJ|F_MUTABLE_BIT);
-                            memcpy(obj, initCell, requiredSize * sizeof(PolyWord));
                             return obj;
                         }
                         else if (length >= actualLargest) actualLargest = length+1;
@@ -688,9 +850,16 @@ PolyObject*MemMgr::AllocCodeSpace(PolyObject *initCell)
         else
         {
             // Allocate a new area and add it at the end of the table.
-            CodeSpace *allocSpace = NewCodeSpace(requiredSize + 1);
+            uintptr_t spaceSize = requiredSize + 1;
+#ifdef POLYML32IN64
+            // We need to allow for the extra alignment word otherwise we
+            // may allocate less than we need.
+            spaceSize += 1;
+#endif
+            CodeSpace *allocSpace = NewCodeSpace(spaceSize);
             if (allocSpace == 0)
                 return 0; // Try a GC.
+            globalStats.incSize(PSS_CODE_SPACE, allocSpace->spaceSize() * sizeof(PolyWord));
         }
     }
 }
@@ -707,7 +876,8 @@ void MemMgr::RemoveEmptyCodeAreas()
         if (start->IsByteObject() && start->Length() == space->spaceSize()-1)
         {
             if (debugOptions & DEBUG_MEMMGR)
-                Log("MMGR: Deleted code space %p\n", space);
+                Log("MMGR: Deleted code space %p at %p size %zu\n", space, space->bottom, space->spaceSize());
+            globalStats.decSize(PSS_CODE_SPACE, space->spaceSize() * sizeof(PolyWord));
             // We have an empty cell that fills the whole space.
             RemoveTree(space);
             delete(space);
@@ -734,9 +904,9 @@ bool MemMgr::AddCodeSpace(CodeSpace *space)
 // Check that we have sufficient space for an allocation to succeed.
 // Called from the GC to ensure that we will not get into an infinite
 // loop trying to allocate, failing and garbage-collecting again.
-bool MemMgr::CheckForAllocation(POLYUNSIGNED words)
+bool MemMgr::CheckForAllocation(uintptr_t words)
 {
-    POLYUNSIGNED allocated = 0;
+    uintptr_t allocated = 0;
     return AllocHeapSpace(words, allocated, false) != 0;
 }
 
@@ -744,13 +914,13 @@ bool MemMgr::CheckForAllocation(POLYUNSIGNED words)
 // size of the allocation area is less than the required value.  This
 // is used after the quick GC and also if we need to allocate a large
 // object.
-void MemMgr::RemoveExcessAllocation(POLYUNSIGNED words)
+void MemMgr::RemoveExcessAllocation(uintptr_t words)
 {
     // First remove any non-standard allocation areas.
     for (std::vector<LocalMemSpace*>::iterator i = lSpaces.begin(); i < lSpaces.end();)
     {
         LocalMemSpace *space = *i;
-        if (space->allocationSpace && space->allocatedSpace() == 0 &&
+        if (space->allocationSpace && space->isEmpty() &&
                 space->spaceSize() != defaultSpaceSize)
             DeleteLocalSpace(i);
         else i++;
@@ -758,16 +928,16 @@ void MemMgr::RemoveExcessAllocation(POLYUNSIGNED words)
     for (std::vector<LocalMemSpace*>::iterator i = lSpaces.begin(); currentAllocSpace > words && i < lSpaces.end(); )
     {
         LocalMemSpace *space = *i;
-        if (space->allocationSpace && space->allocatedSpace() == 0)
+        if (space->allocationSpace && space->isEmpty())
             DeleteLocalSpace(i);
         else i++;
     }
 }
 
 // Return number of words free in all allocation spaces.
-POLYUNSIGNED MemMgr::GetFreeAllocSpace()
+uintptr_t MemMgr::GetFreeAllocSpace()
 {
-    POLYUNSIGNED freeSpace = 0;
+    uintptr_t freeSpace = 0;
     PLocker lock(&allocLock);
     for (std::vector<LocalMemSpace*>::iterator i = lSpaces.begin(); i < lSpaces.end(); i++)
     {
@@ -778,15 +948,15 @@ POLYUNSIGNED MemMgr::GetFreeAllocSpace()
     return freeSpace;
 }
 
-StackSpace *MemMgr::NewStackSpace(POLYUNSIGNED size)
+StackSpace *MemMgr::NewStackSpace(uintptr_t size)
 {
     PLocker lock(&stackSpaceLock);
 
     try {
-        StackSpace *space = new StackSpace;
+        StackSpace *space = new StackSpace(&osStackAlloc);
         size_t iSpace = size*sizeof(PolyWord);
         space->bottom =
-            (PolyWord*)osMemoryManager->Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE);
+            (PolyWord*)osStackAlloc.Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE);
         if (space->bottom == 0)
         {
             if (debugOptions & DEBUG_MEMMGR)
@@ -816,6 +986,7 @@ StackSpace *MemMgr::NewStackSpace(POLYUNSIGNED size)
         }
         if (debugOptions & DEBUG_MEMMGR)
             Log("MMGR: New stack space %p allocated at %p size %lu\n", space, space->bottom, space->spaceSize());
+        globalStats.incSize(PSS_STACK_SPACE, space->spaceSize() * sizeof(PolyWord));
         return space;
     }
     catch (std::bad_alloc&) {
@@ -833,18 +1004,23 @@ void MemMgr::ProtectImmutable(bool on)
         for (std::vector<LocalMemSpace*>::iterator i = lSpaces.begin(); i < lSpaces.end(); i++)
         {
             LocalMemSpace *space = *i;
-            if (! space->isMutable)
-                osMemoryManager->SetPermissions(space->bottom, (char*)space->top - (char*)space->bottom,
-                    on ? PERMISSION_READ|PERMISSION_EXEC : PERMISSION_READ|PERMISSION_EXEC|PERMISSION_WRITE);
+            if (!space->isMutable)
+            {
+                if (space->isCode)
+                    osCodeAlloc.SetPermissions(space->bottom, (char*)space->top - (char*)space->bottom,
+                        on ? PERMISSION_READ | PERMISSION_EXEC : PERMISSION_READ | PERMISSION_EXEC | PERMISSION_WRITE);
+                else osHeapAlloc.SetPermissions(space->bottom, (char*)space->top - (char*)space->bottom,
+                    on ? PERMISSION_READ : PERMISSION_READ | PERMISSION_WRITE);
+            }
         }
     }
 }
 
-bool MemMgr::GrowOrShrinkStack(TaskData *taskData, POLYUNSIGNED newSize)
+bool MemMgr::GrowOrShrinkStack(TaskData *taskData, uintptr_t newSize)
 {
     StackSpace *space = taskData->stack;
     size_t iSpace = newSize*sizeof(PolyWord);
-    PolyWord *newSpace = (PolyWord*)osMemoryManager->Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE);
+    PolyWord *newSpace = (PolyWord*)osStackAlloc.Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE);
     if (newSpace == 0)
     {
         if (debugOptions & DEBUG_MEMMGR)
@@ -865,12 +1041,13 @@ bool MemMgr::GrowOrShrinkStack(TaskData *taskData, POLYUNSIGNED newSize)
     taskData->CopyStackFrame(space->stack(), space->spaceSize(), (StackObject*)newSpace, newSize);
     if (debugOptions & DEBUG_MEMMGR)
         Log("MMGR: Size of stack %p changed from %lu to %lu at %p\n", space, space->spaceSize(), newSize, newSpace);
+    globalStats.incSize(PSS_STACK_SPACE, (newSize - space->spaceSize()) * sizeof(PolyWord));
     RemoveTree(space); // Remove it BEFORE freeing the space - another thread may allocate it
     PolyWord *oldBottom = space->bottom;
     size_t oldSize = (char*)space->top - (char*)space->bottom;
     space->bottom = newSpace; // Switch this before freeing - We could get a profile trap during the free
     space->top = newSpace+newSize;
-    osMemoryManager->Free(oldBottom, oldSize);
+    osStackAlloc.Free(oldBottom, oldSize);
     return true;
 }
 
@@ -885,11 +1062,12 @@ bool MemMgr::DeleteStackSpace(StackSpace *space)
     {
         if (*i == space)
         {
+            globalStats.decSize(PSS_STACK_SPACE, space->spaceSize() * sizeof(PolyWord));
             RemoveTree(space);
             delete space;
             sSpaces.erase(i);
             if (debugOptions & DEBUG_MEMMGR)
-                Log("MMGR: Deleted stack space %p\n", space);
+                Log("MMGR: Deleted stack space %p at %p size %zu\n", space, space->bottom, space->spaceSize());
             return true;
         }
     }
@@ -1009,9 +1187,9 @@ void MemMgr::RemoveTreeRange(SpaceTree **tt, MemSpace *space, uintptr_t startS,
     *tt = 0;
 }
 
-POLYUNSIGNED MemMgr::AllocatedInAlloc()
+uintptr_t MemMgr::AllocatedInAlloc()
 {
-    POLYUNSIGNED inAlloc = 0;
+    uintptr_t inAlloc = 0;
     for (std::vector<LocalMemSpace*>::iterator i = lSpaces.begin(); i < lSpaces.end(); i++)
     {
         LocalMemSpace *sp = *i;
@@ -1023,7 +1201,7 @@ POLYUNSIGNED MemMgr::AllocatedInAlloc()
 // Report heap sizes and occupancy before and after GC
 void MemMgr::ReportHeapSizes(const char *phase)
 {
-    POLYUNSIGNED alloc = 0, nonAlloc = 0, inAlloc = 0, inNonAlloc = 0;
+    uintptr_t alloc = 0, nonAlloc = 0, inAlloc = 0, inNonAlloc = 0;
     for (std::vector<LocalMemSpace*>::iterator i = lSpaces.begin(); i < lSpaces.end(); i++)
     {
         LocalMemSpace *sp = *i;
@@ -1047,9 +1225,9 @@ void MemMgr::ReportHeapSizes(const char *phase)
     Log(" (%1.0f%%). Total space ", (float)inAlloc / (float)alloc * 100.0F);
     LogSize(spaceForHeap);
     Log(" %1.0f%% full.\n", (float)(inAlloc + inNonAlloc) / (float)spaceForHeap * 100.0F);
-    Log("Heap: Local spaces %u, permanent spaces %u, code spaces %u, stack spaces %u\n",
+    Log("Heap: Local spaces %" PRI_SIZET ", permanent spaces %" PRI_SIZET ", code spaces %" PRI_SIZET ", stack spaces %" PRI_SIZET "\n",
         lSpaces.size(), pSpaces.size(), cSpaces.size(), sSpaces.size());
-    POLYUNSIGNED cTotal = 0, cOccupied = 0;
+    uintptr_t cTotal = 0, cOccupied = 0;
     for (std::vector<CodeSpace*>::iterator c = cSpaces.begin(); c != cSpaces.end(); c++)
     {
         cTotal += (*c)->spaceSize();
@@ -1060,7 +1238,13 @@ void MemMgr::ReportHeapSizes(const char *phase)
             PolyObject *obj = (PolyObject*)pt;
             if (obj->ContainsForwardingPtr())
             {
+#ifdef POLYML32IN64
+                // This is relative to globalCodeBase not globalHeapBase
+                while (obj->ContainsForwardingPtr())
+                    obj = (PolyObject*)(globalCodeBase + ((obj->LengthWord() & ~_OBJ_TOMBSTONE_BIT) << 1));
+#else
                 obj = obj->FollowForwardingChain();
+#endif
                 pt += obj->Length();
             }
             else
@@ -1072,7 +1256,7 @@ void MemMgr::ReportHeapSizes(const char *phase)
         }
     }
     Log("Heap: Code area: total "); LogSize(cTotal); Log(" occupied: "); LogSize(cOccupied); Log("\n");
-    POLYUNSIGNED stackSpace = 0;
+    uintptr_t stackSpace = 0;
     for (std::vector<StackSpace*>::iterator s = sSpaces.begin(); s != sSpaces.end(); s++)
     {
         stackSpace += (*s)->spaceSize();
@@ -1115,12 +1299,12 @@ PolyObject *MemMgr::FindCodeObject(const byte *addr)
     }
 
     // A bit is set if it is a length word.
-    while ((POLYUNSIGNED)addr & (sizeof(POLYUNSIGNED)-1)) addr--; // Make it word aligned
+    while ((uintptr_t)addr & (sizeof(POLYUNSIGNED)-1)) addr--; // Make it word aligned
     PolyWord *wordAddr = (PolyWord*)addr;
     // Work back to find the first set bit before this.
     // Normally we will find one but if we're looking up a value that
     // is actually an integer it might be in a piece of code that is now free.
-    POLYUNSIGNED bitOffset = profMap->FindLastSet(wordAddr - space->bottom);
+    uintptr_t bitOffset = profMap->FindLastSet(wordAddr - space->bottom);
     if (space->spaceType == ST_CODE)
     {
         PolyWord *ptr = space->bottom+bitOffset;
@@ -1128,7 +1312,14 @@ PolyObject *MemMgr::FindCodeObject(const byte *addr)
         // This will find the last non-free code cell or the first cell.
         // Return zero if the value was not actually in the cell or it wasn't code.
         PolyObject *obj = (PolyObject*)(ptr+1);
+#ifdef POLYML32IN64
+        PolyObject *lastObj = obj;
+        // This is relative to globalCodeBase not globalHeapBase.
+        while (lastObj->ContainsForwardingPtr())
+            lastObj = (PolyObject*)(globalCodeBase + ((lastObj->LengthWord() & ~_OBJ_TOMBSTONE_BIT) << 1));
+#else
         PolyObject *lastObj = obj->FollowForwardingChain();
+#endif
         // We normally replace forwarding pointers but when scanning to update
         // addresses after a saved state we may not have yet done that.
         if (wordAddr > ptr && wordAddr < ptr + 1 + lastObj->Length() && lastObj->IsCodeObject())
@@ -1159,5 +1350,16 @@ void MemMgr::RemoveProfilingBitmaps()
         (*i)->profileCode.Destroy();
 }
 
+#ifdef POLYML32IN64DEBUG
+POLYOBJECTPTR PolyWord::AddressToObjectPtr(void *address)
+{
+    ASSERT(address >= globalHeapBase);
+    uintptr_t offset = (PolyWord*)address - globalHeapBase;
+    ASSERT(offset <= 0x7fffffff); // Currently limited to 8Gbytes
+    ASSERT((offset & 1) == 0);
+    return (POLYOBJECTPTR)offset;
+}
+#endif
+
 MemMgr gMem; // The one and only memory manager object
 
