@@ -4,7 +4,7 @@
     Copyright (c) 2000
         Cambridge University Technical Services Limited
 
-    Further development copyright David C.J. Matthews 2001-12, 2015
+    Further development copyright David C.J. Matthews 2001-12, 2015, 2017-19
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -52,7 +52,7 @@
 #define ASSERT(x) 0
 #endif
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #include <tchar.h>
 #else
 #define _T(x) x
@@ -80,16 +80,17 @@
 #include "polystring.h"
 #include "statistics.h"
 #include "noreturn.h"
+#include "savestate.h"
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
-#include "Console.h"
+#if (defined(_WIN32))
+#include "winstartup.h"
+#include "winguiconsole.h"
 
 static const TCHAR *lpszServiceName = 0; // DDE service name
 #endif
 
 FILE *polyStdout, *polyStderr; // Redirected in the Windows GUI
 
-static void  InitHeaderFromExport(exportDescription *exports);
 NORETURNFN(static void Usage(const char *message, ...));
 
 
@@ -125,7 +126,7 @@ static struct __argtab {
     { _T("--gcthreads"),    "Number of threads to use for garbage collection",      OPT_GCTHREADS },
     { _T("--debug"),        "Debug options: checkmem, gc, x",                       OPT_DEBUGOPTS },
     { _T("--logfile"),      "Logging file (default is to log to stdout)",           OPT_DEBUGFILE },
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #ifdef UNICODE
     { _T("--codepage"),     "Code-page to use for file-names etc in Windows",       OPT_CODEPAGE },
 #endif
@@ -195,10 +196,18 @@ POLYUNSIGNED parseSize(const TCHAR *p, const TCHAR *arg)
     }
     if (*p != 0)
         Usage("Malformed %s option\n", arg);
-    // Check that the number of kbytes is less than the address space.
-    // The value could overflow when converted to bytes.
-    if (result >= ((POLYUNSIGNED)1 << (SIZEOF_VOIDP*8 - 10)))
-        Usage("Value of %s option is too large\n", arg);
+    // The sizes must not exceed the possible heap size.
+#ifdef POLYML32IN64
+    if (result > 16 * 1024 * 1024)
+        Usage("Value of %s option must not exceeed 16Gbytes\n", arg);
+#elif (SIZEOF_VOIDP == 4)
+    if (result > 4 * 1024 * 1024)
+        Usage("Value of %s option must not exceeed 4Gbytes\n", arg);
+#else
+    // For completion only!
+    if (result > (POLYUNSIGNED)8 * 1024 * 1024 * 1024 * 1024 * 1024)
+        Usage("Value of %s option must not exceeed 8Ebytes\n", arg);
+#endif
     return result;
 }
 
@@ -315,7 +324,7 @@ int polymain(int argc, TCHAR **argv, exportDescription *exports)
                     case OPT_DEBUGFILE:
                         SetLogFile(p);
                         break;
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
                     case OPT_DDESERVICE:
                         // Set the name for the DDE service.  This allows the caller to specify the
                         // service name to be used to send Interrupt "signals".
@@ -346,6 +355,9 @@ int polymain(int argc, TCHAR **argv, exportDescription *exports)
             userOptions.user_arg_strings[userOptions.user_arg_count++] = argv[i];
     }
 
+    if (!gMem.Initialise())
+        Usage("Unable to initialise memory allocator\n");
+
     if (exports == 0 && importFileName == 0)
         Usage("Missing import file name\n");
 
@@ -374,7 +386,7 @@ int polymain(int argc, TCHAR **argv, exportDescription *exports)
     // Set the heap size if it has been provided otherwise use the default.
     gHeapSizeParameters.SetHeapParameters(minsize, maxsize, initsize, gcpercent);
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     SetupDDEHandler(lpszServiceName); // Windows: Start the DDE handler now we processed any service name.
 #endif
 
@@ -385,10 +397,7 @@ int polymain(int argc, TCHAR **argv, exportDescription *exports)
     PolyObject *rootFunction = 0;
 
     if (exports != 0)
-    {
-        InitHeaderFromExport(exports);
-        rootFunction = (PolyObject *)exports->rootFunction;
-    }
+        rootFunction = InitHeaderFromExport(exports);
     else
     {
         if (importFileName != 0)
@@ -419,7 +428,7 @@ void finish (int n)
     // Make sure we don't get any interrupts once the destructors are
     // applied to globals or statics.
     Uninitialise();
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     ExitThread(n);
 #else
     exit (n);
@@ -454,7 +463,7 @@ void Usage(const char *message, ...)
     }
     fflush(polyStdout);
     
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     if (useConsole)
     {
         MessageBox(hMainWindow, _T("Poly/ML has exited"), _T("Poly/ML"), MB_OK);
@@ -495,38 +504,3 @@ char *RTSArgHelp(void)
     ASSERT((unsigned)(p - buff) < (unsigned)sizeof(buff));
     return buff;
 }
-
-void InitHeaderFromExport(exportDescription *exports)
-{
-    // Check the structure sizes stored in the export structure match the versions
-    // used in this library.
-    if (exports->structLength != sizeof(exportDescription) ||
-        exports->memTableSize != sizeof(memoryTableEntry) ||
-        exports->rtsVersion < FIRST_supported_version ||
-        exports->rtsVersion > LAST_supported_version)
-    {
-#if (FIRST_supported_version == LAST_supported_version)
-        Exit("The exported object file has version %0.2f but this library supports %0.2f",
-            ((float)exports->rtsVersion) / 100.0,
-            ((float)FIRST_supported_version) / 100.0);
-#else
-        Exit("The exported object file has version %0.2f but this library supports %0.2f-%0.2f",
-            ((float)exports->rtsVersion) / 100.0,
-            ((float)FIRST_supported_version) / 100.0,
-            ((float)LAST_supported_version) / 100.0);
-#endif
-    }
-    // We could also check the RTS version and the architecture.
-    exportTimeStamp = exports->timeStamp; // Needed for load and save.
-
-    memoryTableEntry *memTable = exports->memTable;
-    for (unsigned i = 0; i < exports->memTableEntries; i++)
-    {
-        // Construct a new space for each of the entries.
-        if (gMem.NewPermanentSpace(
-                (PolyWord*)memTable[i].mtAddr,
-                memTable[i].mtLength/sizeof(PolyWord), (unsigned)memTable[i].mtFlags,
-                (unsigned)memTable[i].mtIndex) == 0)
-            Exit("Unable to initialise a permanent memory space");
-    }
-}
