@@ -4,10 +4,11 @@
     Copyright (c) 2000
         Cambridge University Technical Services Limited
 
+    Further development David C.J. Matthews 2016
+
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -30,4 +31,7 @@ class TaskData;
 extern Handle ObjSize(TaskData *taskData, Handle obj);
 extern Handle ShowSize(TaskData *taskData, Handle obj);
 extern Handle ObjProfile(TaskData *taskData, Handle obj);
+
+extern struct _entrypts objSizeEPT[];
+
 #endif
