@@ -2,7 +2,7 @@
     Title:     Export memory as a PE/COFF object
     Author:    David C. J. Matthews.
 
-    Copyright (c) 2006, 2011, 2016 David C. J. Matthews
+    Copyright (c) 2006, 2011, 2016-18 David C. J. Matthews
 
 
     This library is free software; you can redistribute it and/or
@@ -76,7 +76,7 @@
 #define RELATIVE_32BIT_RELOCATION   IMAGE_REL_I386_REL32
 #endif
 
-void PECOFFExport::addExternalReference(void *relocAddr, const char *name)
+void PECOFFExport::addExternalReference(void *relocAddr, const char *name, bool/* isFuncPtr*/)
 {
     externTable.makeEntry(name);
     IMAGE_RELOCATION reloc;
@@ -92,7 +92,7 @@ void PECOFFExport::addExternalReference(void *relocAddr, const char *name)
 void PECOFFExport::setRelocationAddress(void *p, DWORD *reloc)
 {
     unsigned area = findArea(p);
-    DWORD offset = (DWORD)((char*)p - (char*)memTable[area].mtAddr);
+    DWORD offset = (DWORD)((char*)p - (char*)memTable[area].mtOriginalAddr);
     *reloc = offset;
 }
 
@@ -104,7 +104,7 @@ PolyWord PECOFFExport::createRelocation(PolyWord p, void *relocAddr)
     setRelocationAddress(relocAddr, &reloc.VirtualAddress);
     void *addr = p.AsAddress();
     unsigned addrArea = findArea(addr);
-    POLYUNSIGNED offset = (char*)addr - (char*)memTable[addrArea].mtAddr;
+    POLYUNSIGNED offset = (POLYUNSIGNED)((char*)addr - (char*)memTable[addrArea].mtOriginalAddr);
     reloc.SymbolTableIndex = addrArea;
     reloc.Type = DIRECT_WORD_RELOCATION;
     fwrite(&reloc, sizeof(reloc), 1, exportFile);
@@ -147,13 +147,14 @@ void PECOFFExport::writeSymbol(const char *symbolName, __int32 value, int sectio
    that the offset is saved in original word. */
 void PECOFFExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind code)
 {
+#ifndef POLYML32IN64
     IMAGE_RELOCATION reloc;
-    PolyWord p = GetConstantValue(addr, code);
+    PolyObject *p = GetConstantValue(addr, code);
 
-    if (IS_INT(p) || p == PolyWord::FromUnsigned(0))
+    if (p == 0)
         return;
 
-    void *a = p.AsAddress();
+    void *a = p;
     unsigned aArea = findArea(a);
 
     // We don't need a relocation if this is relative to the current segment
@@ -163,7 +164,7 @@ void PECOFFExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind
 
     setRelocationAddress(addr, &reloc.VirtualAddress);
     // Set the value at the address to the offset relative to the symbol.
-    POLYUNSIGNED offset = (char*)a - (char*)memTable[aArea].mtAddr;
+    uintptr_t offset = (char*)a - (char*)memTable[aArea].mtOriginalAddr;
     reloc.SymbolTableIndex = aArea;
 
     // The value we store here is the offset whichever relocation method
@@ -182,6 +183,7 @@ void PECOFFExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind
 
     fwrite(&reloc, sizeof(reloc), 1, exportFile);
     relocationCount++;
+#endif
 }
 
 // Set the file alignment.
@@ -270,7 +272,7 @@ void PECOFFExport::exportStore(void)
         relocationCount = 1;
 
         // Create the relocation table and turn all addresses into offsets.
-        char *start = (char*)memTable[i].mtAddr;
+        char *start = (char*)memTable[i].mtOriginalAddr;
         char *end = start + memTable[i].mtLength;
         for (p = (PolyWord*)start; p < (PolyWord*)end; )
         {
@@ -326,7 +328,7 @@ void PECOFFExport::exportStore(void)
         reloc.Type = DIRECT_WORD_RELOCATION;
         reloc.SymbolTableIndex = i; // Relative to base symbol
         reloc.VirtualAddress =
-            sizeof(exportDescription) + i * sizeof(memoryTableEntry) + offsetof(memoryTableEntry, mtAddr);
+            sizeof(exportDescription) + i * sizeof(memoryTableEntry) + offsetof(memoryTableEntry, mtCurrentAddr);
         fwrite(&reloc, sizeof(reloc), 1, exportFile);
         relocationCount++;
     }
@@ -338,7 +340,7 @@ void PECOFFExport::exportStore(void)
     for (i = 0; i < memTableEntries; i++)
     {
         sections[i].PointerToRawData = ftell(exportFile);
-        fwrite(memTable[i].mtAddr, 1, memTable[i].mtLength, exportFile);
+        fwrite(memTable[i].mtOriginalAddr, 1, memTable[i].mtLength, exportFile);
     }
 
     sections[memTableEntries].PointerToRawData = ftell(exportFile);
@@ -347,15 +349,20 @@ void PECOFFExport::exportStore(void)
     exports.memTableSize = sizeof(memoryTableEntry);
     exports.memTableEntries = memTableEntries;
     exports.memTable = (memoryTableEntry *)sizeof(exports); // It follows immediately after this.
-    exports.rootFunction = (void*)((char*)rootFunction - (char*)memTable[rootAddrArea].mtAddr);
+    exports.rootFunction = (void*)((char*)rootFunction - (char*)memTable[rootAddrArea].mtOriginalAddr);
     exports.timeStamp = now;
     exports.architecture = machineDependent->MachineArchitecture();
     exports.rtsVersion = POLY_version_number;
+#ifdef POLYML32IN64
+    exports.originalBaseAddr = globalHeapBase;
+#else
+    exports.originalBaseAddr = 0; 
+#endif
 
     // Set the address values to zero before we write.  They will always
     // be relative to their base symbol.
     for (i = 0; i < memTableEntries; i++)
-        memTable[i].mtAddr = 0;
+        memTable[i].mtCurrentAddr = 0;
 
     fwrite(&exports, sizeof(exports), 1, exportFile);
     fwrite(memTable, sizeof(memoryTableEntry), memTableEntries, exportFile);
