@@ -2,7 +2,7 @@
     Title:     Export and import memory in a portable format
     Author:    David C. J. Matthews.
 
-    Copyright (c) 2006-7, 2015-7 David C. J. Matthews
+    Copyright (c) 2006-7, 2015-8 David C. J. Matthews
 
 
     This library is free software; you can redistribute it and/or
@@ -50,10 +50,9 @@
 #include "version.h"
 #include "sys.h"
 #include "polystring.h"
-#include "processes.h" // For IO_SPACING
 #include "memmgr.h"
-#include "osmem.h"
 #include "rtsentry.h"
+#include "mpoly.h" // For polyStderr
 
 /*
 This file contains the code both to export the file and to import it
@@ -62,29 +61,23 @@ in a new session.
 
 PExport::PExport()
 {
-    pMap = 0;
-    nMapSize = 0;
-    nObjects = 0;
-    indexOrder = 0;
 }
 
 PExport::~PExport()
 {
-    free(pMap);
-    free(indexOrder);
 }
 
 
 // Get the index corresponding to an address.
-unsigned long PExport::getIndex(PolyObject *p)
+size_t PExport::getIndex(PolyObject *p)
 {
     // Binary chop to find the index from the address.
-    unsigned long lower = 0, upper = nObjects;
+    size_t lower = 0, upper = pMap.size();
     while (1)
     {
         ASSERT(lower < upper);
-        unsigned long middle = (lower+upper)/2;
-        ASSERT(middle < nObjects);
+        size_t middle = (lower+upper)/2;
+        ASSERT(middle < pMap.size());
         if (p < pMap[middle])
         {
             // Use lower to middle
@@ -103,7 +96,7 @@ unsigned long PExport::getIndex(PolyObject *p)
 /* Get the index corresponding to an address. */
 void PExport::printAddress(void *p)
 {
-    fprintf(exportFile, "@%lu", getIndex((PolyObject*)p));
+    fprintf(exportFile, "@%" PRI_SIZET "", getIndex((PolyObject*)p));
 }
 
 void PExport::printValue(PolyWord q)
@@ -119,9 +112,9 @@ void PExport::printObject(PolyObject *p)
     POLYUNSIGNED length = p->Length();
     POLYUNSIGNED i;
 
-    unsigned long myIndex = getIndex(p);
+    size_t myIndex = getIndex(p);
 
-    fprintf(exportFile, "%lu:", myIndex);
+    fprintf(exportFile, "%" PRI_SIZET ":", myIndex);
 
     if (p->IsMutable())
         putc('M', exportFile);
@@ -138,7 +131,10 @@ void PExport::printObject(PolyObject *p)
         {
             // This is either an entry point or a weak ref used in the FFI.
             // Clear the first word
-            if (p->Length() >= 1) p->Set(0, PolyWord::FromSigned(0));
+            if (p->Length() == 1)
+                p->Set(0, PolyWord::FromSigned(0)); // Weak ref
+            else if (p->Length() > 1)
+                *(uintptr_t*)p = 0; // Entry point
         }
         /* May be a string, a long format arbitrary precision
            number or a real number. */
@@ -177,7 +173,7 @@ void PExport::printObject(PolyObject *p)
             */
             byte *u = (byte*)p;
             putc('B', exportFile);
-            fprintf(exportFile, "%" POLYUFMT "|", length*sizeof(PolyWord));
+            fprintf(exportFile, "%" PRI_SIZET "|", length*sizeof(PolyWord));
             for (unsigned i = 0; i < (unsigned)(length*sizeof(PolyWord)); i++)
             {
                 fprintf(exportFile, "%02x", u[i]);
@@ -217,14 +213,24 @@ void PExport::printObject(PolyObject *p)
         // Finally any constants in the code object.
         machineDependent->ScanConstantsWithinCode(p, this);
     }
-    else /* Ordinary objects, essentially tuples. */
+    else // Ordinary objects, essentially tuples, or closures.
     {
-        fprintf(exportFile, "O%" POLYUFMT "|", length);
-        for (i = 0; i < length; i++)
+        fprintf(exportFile, "%c%" POLYUFMT "|", p->IsClosureObject() ? 'L' : 'O', length);
+        if (p->IsClosureObject())
+        {
+            // The first word is always a code address.
+            printAddress(*(PolyObject**)p);
+            i = sizeof(PolyObject*)/sizeof(PolyWord);
+            if (i < length)
+                putc(',', exportFile);
+        }
+        else i = 0;
+        while (i < length)
         {
             printValue(p->Get(i));
             if (i < length-1)
                 putc(',', exportFile);
+            i++;
         }
     }
     fprintf(exportFile, "\n");
@@ -235,97 +241,71 @@ void PExport::printObject(PolyObject *p)
    that the offset is saved in original word. */
 void PExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind code)
 {
-    PolyWord p = GetConstantValue(addr, code);
-    // We put in all the values including tagged constants.
+    PolyObject *p = GetConstantValue(addr, code);
+    if (p == 0) return; // Don't put in tagged constants
+
     // Put in the byte offset and the relocation type code.
     POLYUNSIGNED offset = (POLYUNSIGNED)(addr - (byte*)base);
     ASSERT (offset < base->Length() * sizeof(POLYUNSIGNED));
     fprintf(exportFile, "%" POLYUFMT ",%d,", (POLYUNSIGNED)(addr - (byte*)base), code);
-    printValue(p); // The value to plug in.
+    printAddress(p); // The value to plug in.
     fprintf(exportFile, " ");
 }
 
 void PExport::exportStore(void)
 {
-    unsigned i;
-
-    // Calculate a first guess for the map size based on the space size
-    totalBytes = 0;
-    void *startAddr = 0;
-    for (i = 0; i < memTableEntries; i++)
-    {
-        totalBytes += (unsigned long)memTable[i].mtLength;
-        // Get the lowest address.
-        if (startAddr == 0 || memTable[i].mtAddr < startAddr)
-            startAddr = memTable[i].mtAddr;
-    }
-    // Create a map entry for each entry.  Allow five words per object.
-    nMapSize = totalBytes/(sizeof(PolyWord)*5);
-    pMap = (PolyObject **)malloc(sizeof(PolyObject*)*nMapSize);
-
-    if (pMap == 0)
-        throw MemoryException();
-
     // We want the entries in pMap to be in ascending
     // order of address to make searching easy so we need to process the areas
     // in order of increasing address, which may not be the order in memTable.
-    indexOrder = (unsigned*)calloc(sizeof(unsigned), memTableEntries);
-    if (indexOrder == 0)
-        throw MemoryException();
+    std::vector<size_t> indexOrder;
+    indexOrder.reserve(memTableEntries);
 
-    unsigned items = 0;
-    for (i = 0; i < memTableEntries; i++)
+    for (size_t i = 0; i < memTableEntries; i++)
     {
-        unsigned j = items;
-        while (j > 0 && memTable[i].mtAddr < memTable[indexOrder[j-1]].mtAddr)
-        {
-            indexOrder[j] = indexOrder[j-1];
-            j--;
+        std::vector<size_t>::iterator it;
+        for (it = indexOrder.begin(); it != indexOrder.end(); it++) {
+            if (memTable[*it].mtOriginalAddr >= memTable[i].mtOriginalAddr)
+                break;
         }
-        indexOrder[j] = i;
-        items++;
+        indexOrder.insert(it, i);
     }
-    ASSERT(items == memTableEntries);
 
     // Process the area in order of ascending address.
-    for (i = 0; i < items; i++)
+    for (std::vector<size_t>::iterator i = indexOrder.begin(); i != indexOrder.end(); i++)
     {
-        unsigned index = indexOrder[i];
-        char *start = (char*)memTable[index].mtAddr;
+        size_t index = *i;
+        char *start = (char*)memTable[index].mtOriginalAddr;
         char *end = start + memTable[index].mtLength;
         for (PolyWord *p = (PolyWord*)start; p < (PolyWord*)end; )
         {
             p++;
             PolyObject *obj = (PolyObject*)p;
-            if (nObjects == nMapSize)
-            {
-                // Need to expand the array.
-                PolyObject **newMap =
-                    (PolyObject **)realloc(pMap, (nMapSize + nMapSize/2)*sizeof(PolyObject*));
-                if (newMap == 0)
-                    throw MemoryException();
-                pMap = newMap;
-            }
             POLYUNSIGNED length = obj->Length();
-            pMap[nObjects++] = obj;
+            pMap.push_back(obj);
             p += length;
         }
     }
 
     /* Start writing the information. */
-    fprintf(exportFile, "Objects\t%lu\n", nObjects);
-    fprintf(exportFile, "Root\t%lu\n", getIndex(rootFunction));
+    fprintf(exportFile, "Objects\t%" PRI_SIZET "\n", pMap.size());
+    fprintf(exportFile, "Root\t%" PRI_SIZET "\n", getIndex(rootFunction));
 
     // Generate each of the areas.
-    for (i = 0; i < memTableEntries; i++)
+    for (size_t i = 0; i < memTableEntries; i++)
     {
-        char *start = (char*)memTable[i].mtAddr;
+        char *start = (char*)memTable[i].mtOriginalAddr;
         char *end = start + memTable[i].mtLength;
         for (PolyWord *p = (PolyWord*)start; p < (PolyWord*)end; )
         {
             p++;
             PolyObject *obj = (PolyObject*)p;
             POLYUNSIGNED length = obj->Length();
+#ifdef POLYML32IN64
+            // We may have filler cells to get the alignment right.
+            // We mustn't try to print them.
+            if (((uintptr_t)obj & 4) != 0 && length == 0)
+                continue;
+#endif
             printObject(obj);
             p += length;
         }
@@ -343,78 +323,84 @@ Creates "permanent" address entries in the global memory table.
 class SpaceAlloc
 {
 public:
-    SpaceAlloc(bool isMut, POLYUNSIGNED def);
-    ~SpaceAlloc();
+    SpaceAlloc(unsigned *indexCtr, unsigned perms, POLYUNSIGNED def);
     PolyObject *NewObj(POLYUNSIGNED objWords);
-    bool AddToTable(void);
 
-    POLYUNSIGNED defaultSize;
-    POLYUNSIGNED currentSize;
-    PolyWord *base;
-    POLYUNSIGNED used;
-    bool isMutable;
-    unsigned spaceIndex;
+    size_t defaultSize;
+    PermanentMemSpace *memSpace;
+    size_t used;
+    unsigned permissions;
+    unsigned *spaceIndexCtr;
+
 };
 
-SpaceAlloc::SpaceAlloc(bool isMut, POLYUNSIGNED def)
+SpaceAlloc::SpaceAlloc(unsigned *indexCtr, unsigned perms, POLYUNSIGNED def)
 {
-    isMutable = isMut;
+    permissions = perms;
     defaultSize = def;
-    base = 0;
-    currentSize = 0;
+    memSpace = 0;
     used = 0;
-    spaceIndex = 1;
-}
-
-SpaceAlloc::~SpaceAlloc()
-{
-    if (base)
-        osMemoryManager->Free(base, currentSize*sizeof(PolyWord));
-}
-
-bool SpaceAlloc::AddToTable(void)
-{
-    if (base != 0)
-    {
-        // Add the new space to the permanent memory table.
-        MemSpace* space = gMem.NewPermanentSpace(base, used, isMutable ? MTF_WRITEABLE : 0, spaceIndex++);
-        if (space == 0)
-        {
-            fprintf(stderr, "Insufficient memory\n");
-            return false;
-        }
-    }
-    base = 0;
-    return true;
+    spaceIndexCtr = indexCtr;
 }
 
 // Allocate a new object.  May create a new space and add the old one to the permanent
 // memory table if this is exhausted.
+#ifndef POLYML32IN64
 PolyObject *SpaceAlloc::NewObj(POLYUNSIGNED objWords)
 {
-    if (currentSize - used <= objWords)
+    if (memSpace == 0 || memSpace->spaceSize() - used <= objWords)
     {
         // Need some more space.
-        if (! AddToTable())
-            return 0;
-        POLYUNSIGNED size = defaultSize;
+        size_t size = defaultSize;
         if (size <= objWords)
             size = objWords+1;
-        size_t iSpace = size*sizeof(PolyWord);
-        base = (PolyWord*)osMemoryManager->Allocate(iSpace, PERMISSION_READ|PERMISSION_WRITE|PERMISSION_EXEC);
-        if (base == 0)
+        memSpace =
+            gMem.AllocateNewPermanentSpace(size * sizeof(PolyWord), permissions, *spaceIndexCtr);
+        (*spaceIndexCtr)++;
+        // The memory is writable until CompletePermanentSpaceAllocation is called
+        if (memSpace == 0)
+        {
+            fprintf(polyStderr, "Unable to allocate memory\n");
+            return 0;
+        }
+        used = 0;
+    }
+    ASSERT(memSpace->spaceSize() - used > objWords);
+    PolyObject *newObj = (PolyObject*)(memSpace->bottom + used+1);
+    used += objWords+1;
+    return newObj;
+}
+#else
+// With 32in64 we need to allocate on 8-byte boundaries. 
+PolyObject *SpaceAlloc::NewObj(POLYUNSIGNED objWords)
+{
+    size_t rounded = objWords;
+    if ((objWords & 1) == 0) rounded++;
+    if (memSpace == 0 || memSpace->spaceSize() - used <= rounded)
+    {
+        // Need some more space.
+        size_t size = defaultSize;
+        if (size <= rounded)
+            size = rounded + 1;
+        memSpace =
+            gMem.AllocateNewPermanentSpace(size * sizeof(PolyWord), permissions, *spaceIndexCtr);
+        (*spaceIndexCtr)++;
+        // The memory is writable until CompletePermanentSpaceAllocation is called
+        if (memSpace == 0)
         {
             fprintf(stderr, "Unable to allocate memory\n");
             return 0;
         }
-        currentSize = iSpace/sizeof(PolyWord);
-        used = 0;
+        memSpace->bottom[0] = PolyWord::FromUnsigned(0);
+        used = 1;
     }
-    ASSERT(currentSize - used > objWords);
-    PolyObject *newObj = (PolyObject*)(base+used+1);
-    used += objWords+1;
+    PolyObject *newObj = (PolyObject*)(memSpace->bottom + used + 1);
+    if (rounded != objWords) newObj->Set(objWords, PolyWord::FromUnsigned(0));
+    used += rounded + 1;
+    ASSERT(((uintptr_t)newObj & 0x7) == 0);
     return newObj;
 }
+#endif
 
 class PImport
 {
@@ -425,20 +411,25 @@ public:
     FILE *f;
     PolyObject *Root(void) { return objMap[nRoot]; }
 private:
-    PolyObject *NewObject(POLYUNSIGNED words, bool isMutable);
     bool ReadValue(PolyObject *p, POLYUNSIGNED i);
     bool GetValue(PolyWord *result);
     
     POLYUNSIGNED nObjects, nRoot;
     PolyObject **objMap;
 
-    SpaceAlloc mutSpace, immutSpace;
+    unsigned spaceIndex;
+
+    SpaceAlloc mutSpace, immutSpace, codeSpace;
 };
 
-PImport::PImport(): mutSpace(true, 1024*1024), immutSpace(false, 1024*1024)
+PImport::PImport():
+    mutSpace(&spaceIndex, MTF_WRITEABLE, 1024*1024),
+    immutSpace(&spaceIndex, 0, 1024*1024),
+    codeSpace(&spaceIndex, MTF_EXECUTABLE, 1024 * 1024)
 {
     f = NULL;
     objMap = 0;
+    spaceIndex = 1;
 }
 
 PImport::~PImport()
@@ -448,20 +439,6 @@ PImport::~PImport()
     free(objMap);
 }
 
-PolyObject *PImport::NewObject(POLYUNSIGNED words, bool isMutableObj)
-{
-    PolyObject *newObj = 0;
-    if (isMutableObj)
-        newObj = mutSpace.NewObj(words);
-    else
-        newObj = immutSpace.NewObj(words);
-    if (newObj == 0)
-        return 0;
-
-    return newObj;
-
-}
-
 bool PImport::GetValue(PolyWord *result)
 {
     int ch = getc(f);
@@ -473,16 +450,6 @@ bool PImport::GetValue(PolyWord *result)
         ASSERT(obj < nObjects);
         *result = objMap[obj];
     }
-    else if (ch == '$')
-    {
-        /* Code address. */
-        POLYUNSIGNED obj, offset;
-        fscanf(f, "%" POLYUFMT "+%" POLYUFMT, &obj, &offset);
-        ASSERT(obj < nObjects);
-        PolyObject *q = objMap[obj];
-        ASSERT(q->IsCodeObject());
-        *result = PolyWord::FromCodePtr((PolyWord(q)).AsCodePtr() + offset); /* The offset is in bytes. */
-    }
     else if ((ch >= '0' && ch <= '9') || ch == '-')
     {
         /* Tagged integer. */
@@ -494,25 +461,9 @@ bool PImport::GetValue(PolyWord *result)
         ASSERT(j >= -MAXTAGGED-1 && j <= MAXTAGGED);
         *result = TAGGED(j);
     }
-    else if (ch == 'I')
-    {
-        /* IO entry number. */
-        POLYUNSIGNED j;
-        fscanf(f, "%" POLYUFMT, &j);
-        // We may still have references to the old empty string value (j == 48).
-        if (j == 48)
-        {
-            // This is a bit of a hack but it's only temporary.
-            PolyObject  *p = NewObject(1, false);
-            p->SetLengthWord(1, F_BYTE_OBJ);
-            p->Set(0, PolyWord::FromUnsigned(0));
-            *result = p;
-        }
-        else ASSERT(0);
-    }
     else
     {
-        fprintf(stderr, "Unexpected character in stream");
+        fprintf(polyStderr, "Unexpected character in stream");
         return false;
     }
     return true;
@@ -539,8 +490,6 @@ bool PImport::DoImport()
     ASSERT(gMem.eSpaces.size() == 0);
 
     ch = getc(f);
-    /* Skip the "Mapping" line. */
-    if (ch == 'M') { while (getc(f) != '\n') ; ch = getc(f); }
     ASSERT(ch == 'O'); /* Number of objects. */
     while (getc(f) != '\t') ;
     fscanf(f, "%" POLYUFMT, &nObjects);
@@ -548,7 +497,7 @@ bool PImport::DoImport()
     objMap = (PolyObject**)calloc(nObjects, sizeof(PolyObject*));
     if (objMap == 0)
     {
-        fprintf(stderr, "Unable to allocate memory\n");
+        fprintf(polyStderr, "Unable to allocate memory\n");
         return false;
     }
 
@@ -563,7 +512,6 @@ bool PImport::DoImport()
     /* Now the objects themselves. */
     while (1)
     {
-        bool     isMutable = false;
         unsigned    objBits = 0;
         POLYUNSIGNED  nWords, nBytes;
         do
@@ -581,11 +529,11 @@ bool PImport::DoImport()
         do
         {
             ch = getc(f);
-            if (ch == 'M') { isMutable = true; objBits |= F_MUTABLE_BIT; }
+            if (ch == 'M') objBits |= F_MUTABLE_BIT;
             else if (ch == 'N') objBits |= F_NEGATIVE_BIT;
             if (ch == 'V') objBits |= F_NO_OVERWRITE;
             if (ch == 'W') objBits |= F_WEAK_BIT;
-        } while (ch == 'M' || ch == 'N' || ch == 'L' || ch == 'V' || ch == 'W');
+        } while (ch == 'M' || ch == 'N' || ch == 'V' || ch == 'W');
 
         /* Object type. */
         switch (ch)
@@ -621,12 +569,22 @@ bool PImport::DoImport()
             nWords += (nBytes + sizeof(PolyWord) -1) / sizeof(PolyWord);
             break;
 
+        case 'L': // Closure
+            objBits |= F_CLOSURE_OBJ;
+            fscanf(f, "%" POLYUFMT, &nWords);
+            break;
+
         default:
-            fprintf(stderr, "Invalid object type\n");
+            fprintf(polyStderr, "Invalid object type\n");
             return false;
         }
 
-        PolyObject  *p = NewObject(nWords, isMutable);
+        PolyObject  *p;
+        if (objBits & F_MUTABLE_BIT)
+            p = mutSpace.NewObj(nWords);
+        else if ((objBits & 3) == F_CODE_OBJ)
+            p = codeSpace.NewObj(nWords);
+        else p = immutSpace.NewObj(nWords);
         if (p == 0)
             return false;
         objMap[objNo] = p;
@@ -641,11 +599,6 @@ bool PImport::DoImport()
     fseek(f, 0, SEEK_SET);
     /* Skip the information at the start. */
     ch = getc(f);
-    if (ch == 'M')
-    {
-        while (getc(f) != '\n') ;
-        ch = getc(f);
-    }
     ASSERT(ch == 'O'); /* Number of objects. */
     while (getc(f) != '\n');
     ch = getc(f);
@@ -654,7 +607,6 @@ bool PImport::DoImport()
 
     while (1)
     {
-        POLYUNSIGNED  nWords, nBytes, i;
         if (feof(f))
             break;
         fscanf(f, "%" POLYUFMT, &objNo);
@@ -669,34 +621,55 @@ bool PImport::DoImport()
         do
         {
             ch = getc(f);
-        } while (ch == 'M' || ch == 'N' || ch == 'L' || ch == 'V' || ch == 'W');
+        } while (ch == 'M' || ch == 'N' || ch == 'V' || ch == 'W');
 
         /* Object type. */
         switch (ch)
         {
         case 'O': /* Simple object. */
+        case 'L': // Closure
+        {
+            POLYUNSIGNED nWords;
+            bool isClosure = ch == 'L';
             fscanf(f, "%" POLYUFMT, &nWords);
             ch = getc(f);
             ASSERT(ch == '|');
             ASSERT(nWords == p->Length());
 
-            for (i = 0; i < nWords; i++)
+            POLYUNSIGNED i = 0;
+            if (isClosure)
             {
-                if (! ReadValue(p, i))
+                int ch = getc(f);
+                // This should be an address
+                if (ch != '@') return false; 
+                POLYUNSIGNED obj;
+                fscanf(f, "%" POLYUFMT, &obj);
+                ASSERT(obj < nObjects);
+                *(PolyObject**)p = objMap[obj];
+                ch = getc(f);
+                i = sizeof(PolyObject*) / sizeof(PolyWord);
+            }
+
+            while (i < nWords)
+            {
+                if (!ReadValue(p, i))
                     return false;
                 ch = getc(f);
-                ASSERT((ch == ',' && i < nWords-1) ||
-                       (ch == '\n' && i == nWords-1));
+                ASSERT((ch == ',' && i < nWords - 1) ||
+                    (ch == '\n' && i == nWords - 1));
+                i++;
             }
 
             break;
+        }
 
         case 'B': /* Byte segment. */
             {
                 byte *u = (byte*)p;
+                POLYUNSIGNED nBytes;
                 fscanf(f, "%" POLYUFMT, &nBytes);
                 ch = getc(f); ASSERT(ch == '|');
-                for (i = 0; i < nBytes; i++)
+                for (POLYUNSIGNED i = 0; i < nBytes; i++)
                 {
                     int n;
                     fscanf(f, "%02x", &n);
@@ -705,7 +678,8 @@ bool PImport::DoImport()
                 ch = getc(f);
                 ASSERT(ch == '\n');
                 // If this is an entry point object set its value.
-                if (p->IsMutable() && p->IsWeakRefObject())
+                //if (p->IsMutable() && p->IsWeakRefObject() && p->Length() > 2 && p->Get(2).AsUnsigned() != 0)
+                if (p->IsMutable() && p->IsWeakRefObject() && p->Length() > sizeof(uintptr_t)/sizeof(PolyWord))
                 {
                     bool loadEntryPt = setEntryPoint(p);
                     ASSERT(loadEntryPt);
@@ -717,10 +691,11 @@ bool PImport::DoImport()
             {
                 PolyStringObject * ps = (PolyStringObject *)p;
                 /* The length is the number of characters. */
+                POLYUNSIGNED nBytes;
                 fscanf(f, "%" POLYUFMT, &nBytes);
                 ch = getc(f); ASSERT(ch == '|');
                 ps->length = nBytes;
-                for (i = 0; i < nBytes; i++)
+                for (POLYUNSIGNED i = 0; i < nBytes; i++)
                 {
                     int n;
                     fscanf(f, "%02x", &n);
@@ -737,12 +712,13 @@ bool PImport::DoImport()
                 bool oldForm = ch == 'C';
                 byte *u = (byte*)p;
                 POLYUNSIGNED length = p->Length();
+                POLYUNSIGNED nWords, nBytes;
                 /* Read the number of bytes of code and the number of words
                    for constants. */
                 fscanf(f, "%" POLYUFMT ",%" POLYUFMT, &nWords, &nBytes);
                 /* Read the code. */
                 ch = getc(f); ASSERT(ch == '|');
-                for (i = 0; i < nBytes; i++)
+                for (POLYUNSIGNED i = 0; i < nBytes; i++)
                 {
                     int n;
                     fscanf(f, "%02x", &n);
@@ -762,7 +738,7 @@ bool PImport::DoImport()
                     ASSERT(nBytes == ((length-1-nWords-3)*sizeof(PolyWord)));
                 }
                 /* Read in the constants. */
-                for (i = 0; i < nWords; i++)
+                for (POLYUNSIGNED i = 0; i < nWords; i++)
                 {
                     if (! ReadValue(p, i+length-nWords-1))
                         return false;
@@ -782,24 +758,43 @@ bool PImport::DoImport()
                         fscanf(f, "%" POLYUFMT ",%d", &offset, &code);
                         ch = getc(f);
                         ASSERT(ch == ',');
-                        PolyWord constVal = TAGGED(0);
-                        if (! GetValue(&constVal))
-                            return false;
-                        byte *toPatch = (byte*)p + offset;
-                        ScanAddress::SetConstantValue(toPatch, constVal, (ScanRelocationKind)code);
-
+                        // This should be an address.
+                        ch = getc(f);
+                        if (ch == '@')
+                        {
+                            POLYUNSIGNED obj;
+                            fscanf(f, "%" POLYUFMT, &obj);
+                            ASSERT(obj < nObjects);
+                            PolyObject *addr = objMap[obj];
+                            byte *toPatch = (byte*)p + offset;
+                            ScanAddress::SetConstantValue(toPatch, addr, (ScanRelocationKind)code);
+                        }
+                        else
+                        {
+                            // Previously we also included tagged constants but they are
+                            // already in the code.
+                            ungetc(ch, f);
+                            PolyWord w;
+                            if (!GetValue(&w))
+                                return false;
+                        }
                         do ch = getc(f); while (ch == ' ');
                     }
                 }
+                // Clear the mutable bit
+                p->SetLengthWord(p->Length(), F_CODE_OBJ);
                 break;
             }
 
         default:
-            fprintf(stderr, "Invalid object type\n");
+            fprintf(polyStderr, "Invalid object type\n");
             return false;
         }
     }
-    return mutSpace.AddToTable() && immutSpace.AddToTable();
+    // Now remove write access from immutable spaces.
+    for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
+        gMem.CompletePermanentSpaceAllocation(*i);
+    return true;
 }
 
 // Import a file in the portable format and return a pointer to the root object.
@@ -810,14 +805,14 @@ PolyObject *ImportPortable(const TCHAR *fileName)
     pImport.f = _wfopen(fileName, L"r");
     if (pImport.f == 0)
     {
-        fprintf(stderr, "Unable to open file: %S\n", fileName);
+        fprintf(polyStderr, "Unable to open file: %S\n", fileName);
         return 0;
     }
 #else
     pImport.f = fopen(fileName, "r");
     if (pImport.f == 0)
     {
-        fprintf(stderr, "Unable to open file: %s\n", fileName);
+        fprintf(polyStderr, "Unable to open file: %s\n", fileName);
         return 0;
     }
 #endif
