@@ -1,7 +1,7 @@
 /*
     Title:  poly_specific.cpp - Poly/ML specific RTS calls.
 
-    Copyright (c) 2006, 2015-17 David C. J. Matthews
+    Copyright (c) 2006, 2015-17, 2019 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -50,14 +50,9 @@
 #include "run_time.h"
 #include "version.h"
 #include "save_vec.h"
-#include "exporter.h"
 #include "version.h"
-#include "sharedata.h"
 #include "memmgr.h"
 #include "processes.h"
-#include "savestate.h"
-#include "statistics.h"
-#include "../polystatistics.h"
 #include "gc.h"
 #include "rtsentry.h"
 
@@ -65,14 +60,17 @@ extern "C" {
     POLYEXTERNALSYMBOL POLYUNSIGNED PolySpecificGeneral(PolyObject *threadId, PolyWord code, PolyWord arg);
     POLYEXTERNALSYMBOL POLYUNSIGNED PolyGetABI();
     POLYEXTERNALSYMBOL POLYUNSIGNED PolyLockMutableCode(PolyObject * threadId, PolyWord byteSeg);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyLockMutableClosure(PolyObject * threadId, PolyWord closure);
     POLYEXTERNALSYMBOL POLYUNSIGNED PolyCopyByteVecToCode(PolyObject *threadId, PolyWord byteVec);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyCopyByteVecToClosure(PolyObject *threadId, PolyWord byteVec, PolyWord closure);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolySetCodeConstant(PolyWord closure, PolyWord offset, PolyWord c, PolyWord flags);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolySetCodeByte(PolyWord closure, PolyWord offset, PolyWord c);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyGetCodeByte(PolyWord closure, PolyWord offset);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolySortArrayOfAddresses(PolyWord array);
 }
 
 #define SAVE(x) taskData->saveVec.push(x)
 
-static const char *poly_runtime_system_copyright =
-    "Copyright (c) 2002-17 David C.J. Matthews, CUTS and contributors.";
-
 #ifndef GIT_VERSION
 #define GIT_VERSION             ""
 #endif
@@ -80,7 +78,7 @@ static const char *poly_runtime_system_copyright =
 
 Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
-    unsigned c = get_C_unsigned(taskData, DEREFWORDHANDLE(code));
+    unsigned c = get_C_unsigned(taskData, DEREFWORD(code));
     switch (c)
     {
     case 9: // Return the GIT version if appropriate
@@ -101,12 +99,8 @@ Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
             return SAVE(C_string_to_Poly(taskData, version));
         }
 
-    case 11: // Return the RTS copyright string
-        // Is this used??
-        return SAVE(C_string_to_Poly(taskData, poly_runtime_system_copyright));
-
     case 12: // Return the architecture
-        // Is this used??
+        // Used in InitialPolyML.ML for PolyML.architecture
         {
             const char *arch;
             switch (machineDependent->MachineArchitecture())
@@ -114,6 +108,7 @@ Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
             case MA_Interpreted:    arch = "Interpreted"; break;
             case MA_I386:           arch = "I386"; break;
             case MA_X86_64:         arch = "X86_64"; break;
+            case MA_X86_64_32:      arch = "X86_64_32"; break;
             default:                arch = "Unknown"; break;
             }
             return SAVE(C_string_to_Poly(taskData, arch));
@@ -122,129 +117,6 @@ Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 19: // Return the RTS argument help string.
         return SAVE(C_string_to_Poly(taskData, RTSArgHelp()));
 
-    case 20: // Write a saved state file.
-        return SaveState(taskData, args);
-
-    case 21: // Load a saved state file and any ancestors.
-        return LoadState(taskData, false, args);
-
-    case 22: // Show the hierarchy.
-        return ShowHierarchy(taskData);
-
-    case 23: // Change the name of the immediate parent stored in a child
-        return RenameParent(taskData, args);
-
-    case 24: // Return the name of the immediate parent stored in a child
-        return ShowParent(taskData, args);
-
-    case 27: // Get number of user statistics available
-        return Make_arbitrary_precision(taskData, N_PS_USER);
-
-    case 28: // Set an entry in the user stats table.
-        {
-            unsigned index = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(0));
-            if (index >= N_PS_USER)
-                raise_exception0(taskData, EXC_subscript);
-            POLYSIGNED value = getPolySigned(taskData, DEREFHANDLE(args)->Get(1));
-            globalStats.setUserCounter(index, value);
-            Make_arbitrary_precision(taskData, 0);
-        }
-
-    case 29: // Get local statistics.
-        return globalStats.getLocalStatistics(taskData);
-
-    case 30: // Get remote statistics.  The argument is the process ID to get the statistics.
-        return globalStats.getRemoteStatistics(taskData, getPolyUnsigned(taskData, DEREFHANDLE(args)));
-
-    case 31: // Store a module
-        return StoreModule(taskData, args);
-
-    case 32: // Load a module
-        return LoadModule(taskData, args);
-
-    case 33: // Load hierarchy.  This provides a complete list of children and parents.
-        return LoadState(taskData, true, args);
-
-    case 34: // Return the system directory for modules.  This is configured differently
-        // in Unix and in Windows.
-#if (defined(MODULEDIR))
-    return SAVE(C_string_to_Poly(taskData, MODULEDIR));
-#elif (defined(_WIN32) && ! defined(__CYGWIN__))
-        {
-            // This registry key is configured when Poly/ML is installed using the installer.
-            // It gives the path to the Poly/ML installation directory.  We return the
-            // Modules subdirectory.
-            HKEY hk;
-            if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,
-                    _T("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\PolyML.exe"), 0,
-                    KEY_QUERY_VALUE, &hk) == ERROR_SUCCESS)
-            {
-                DWORD valSize;
-                if (RegQueryValueEx(hk, _T("Path"), 0, NULL, NULL, &valSize) == ERROR_SUCCESS)
-                {
-#define MODULEDIR _T("Modules")
-                    TempString buff((TCHAR*)malloc(valSize + (_tcslen(MODULEDIR) + 1)*sizeof(TCHAR)));
-                    DWORD dwType;
-                    if (RegQueryValueEx(hk, _T("Path"), 0, &dwType, (LPBYTE)(LPTSTR)buff, &valSize) == ERROR_SUCCESS)
-                    {
-                        RegCloseKey(hk);
-                        // The registry entry should end with a backslash.
-                        _tcscat(buff, MODULEDIR);
-                        return SAVE(C_string_to_Poly(taskData, buff));
-                    }
-                }
-                RegCloseKey(hk);
-            }
-            return SAVE(C_string_to_Poly(taskData, ""));
-        }
-#else
-        return SAVE(C_string_to_Poly(taskData, ""));
-#endif
-
-    case 106: // Lock a mutable code segment and return the executable address.
-            // Legacy - used by bootstrap code only
-        {
-            PolyObject *codeObj = args->WordP();
-            if (! codeObj->IsCodeObject() || ! codeObj->IsMutable())
-                raise_fail(taskData, "Not mutable code area");
-            POLYUNSIGNED segLength = codeObj->Length();
-            codeObj->SetLengthWord(segLength, F_CODE_OBJ);
-            machineDependent->FlushInstructionCache(codeObj, segLength * sizeof(PolyWord));
-            // In the future it may be necessary to return a different address here.
-            // N.B.  The code area should only have execute permission in the native
-            // code version, not the interpreted version.
-            return args; // Return the original address.
-        }
-
-    case 107: // Copy a byte segment into the code area and make it mutable code
-              // Legacy - used by bootstrap code only
-        {
-            if (! args->WordP()->IsByteObject())
-                raise_fail(taskData, "Not byte data area");
-            while (true)
-            {
-                PolyObject *result = gMem.AllocCodeSpace(args->WordP());
-                if (result != 0)
-                    return taskData->saveVec.push(result);
-                // Could not allocate - must GC.
-                if (! QuickGC(taskData, args->WordP()->Length()))
-                    raise_fail(taskData, "Insufficient memory");
-            }
-        }
-
-    case 108:
-        // Return the ABI.  For 64-bit we need to know if this is Windows.
-        // Legacy - used by bootstrap code only
-#if (SIZEOF_VOIDP == 8)
-#ifdef _WIN32
-        return taskData->saveVec.push(TAGGED(2));
-#else
-        return taskData->saveVec.push(TAGGED(1));
-#endif
-#else
-        return taskData->saveVec.push(TAGGED(0));
-#endif
-
     default:
         {
             char msg[100];
@@ -282,7 +154,7 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyGetABI()
 {
     // Return the ABI.  For 64-bit we need to know if this is Windows.
 #if (SIZEOF_VOIDP == 8)
-#ifdef _WIN32
+#if (defined(_WIN32) || defined(__CYGWIN__))
     return TAGGED(2).AsUnsigned(); // 64-bit Windows
 #else
     return TAGGED(1).AsUnsigned(); // 64-bit Unix
@@ -295,6 +167,7 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyGetABI()
 // Code generation - Code is initially allocated in a byte segment.  When all the
 // values have been set apart from any addresses the byte segment is copied into
 // a mutable code segment.
+// PolyCopyByteVecToCode is now replaced by PolyCopyByteVecToClosure
 POLYEXTERNALSYMBOL POLYUNSIGNED PolyCopyByteVecToCode(PolyObject * threadId, PolyWord byteVec)
 {
     TaskData *taskData = TaskData::FindTaskForId(threadId);
@@ -308,13 +181,16 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyCopyByteVecToCode(PolyObject * threadId, Pol
         if (!pushedArg->WordP()->IsByteObject())
             raise_fail(taskData, "Not byte data area");
         do {
-            result = gMem.AllocCodeSpace(pushedArg->WordP());
+            PolyObject *initCell = pushedArg->WordP();
+            POLYUNSIGNED requiredSize = initCell->Length();
+            result = gMem.AllocCodeSpace(requiredSize);
             if (result == 0)
             {
                 // Could not allocate - must GC.
                 if (!QuickGC(taskData, pushedArg->WordP()->Length()))
                     raise_fail(taskData, "Insufficient memory");
             }
+            else memcpy(result, initCell, requiredSize * sizeof(PolyWord));
         } while (result == 0);
     }
     catch (...) {} // If an ML exception is raised
@@ -324,6 +200,49 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyCopyByteVecToCode(PolyObject * threadId, Pol
     return ((PolyWord)result).AsUnsigned();
 }
 
+// Copy the byte vector into code space.
+POLYUNSIGNED PolyCopyByteVecToClosure(PolyObject *threadId, PolyWord byteVec, PolyWord closure)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedByteVec = taskData->saveVec.push(byteVec);
+    Handle pushedClosure = taskData->saveVec.push(closure);
+    PolyObject *result = 0;
+
+    try {
+        if (!pushedByteVec->WordP()->IsByteObject())
+            raise_fail(taskData, "Not byte data area");
+        if (pushedClosure->WordP()->Length() != sizeof(PolyObject*)/sizeof(PolyWord))
+            raise_fail(taskData, "Invalid closure size");
+        if (!pushedClosure->WordP()->IsMutable())
+            raise_fail(taskData, "Closure is not mutable");
+        do {
+            PolyObject *initCell = pushedByteVec->WordP();
+            POLYUNSIGNED requiredSize = initCell->Length();
+            result = gMem.AllocCodeSpace(requiredSize);
+            if (result == 0)
+            {
+                // Could not allocate - must GC.
+                if (!QuickGC(taskData, pushedByteVec->WordP()->Length()))
+                    raise_fail(taskData, "Insufficient memory");
+            }
+            else memcpy(result, initCell, requiredSize * sizeof(PolyWord));
+        } while (result == 0);
+    }
+    catch (...) {} // If an ML exception is raised
+
+    // Store the code address in the closure.
+    *((PolyObject**)pushedClosure->WordP()) = result;
+    // Lock the closure.
+    pushedClosure->WordP()->SetLengthWord(pushedClosure->WordP()->LengthWord() & ~_OBJ_MUTABLE_BIT);
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    return TAGGED(0).AsUnsigned();
+}
+
 // Code generation - Lock a mutable code segment and return the original address.
 // Currently this does not allocate so other than the exception it could
 // be a fast call.
@@ -357,12 +276,145 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyLockMutableCode(PolyObject * threadId, PolyW
     else return result->Word().AsUnsigned();
 }
 
+// Replacement for above
+POLYEXTERNALSYMBOL POLYUNSIGNED PolyLockMutableClosure(PolyObject * threadId, PolyWord closure)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    PolyObject *codeObj = *(PolyObject**)(closure.AsObjPtr());
+
+    try {
+        if (!codeObj->IsCodeObject() || !codeObj->IsMutable())
+            raise_fail(taskData, "Not mutable code area");
+        POLYUNSIGNED segLength = codeObj->Length();
+        codeObj->SetLengthWord(segLength, F_CODE_OBJ);
+        // This is really a legacy of the PPC code-generator.
+        machineDependent->FlushInstructionCache(codeObj, segLength * sizeof(PolyWord));
+        // In the future it may be necessary to return a different address here.
+        // N.B.  The code area should only have execute permission in the native
+        // code version, not the interpreted version.
+    }
+    catch (...) {} // If an ML exception is raised
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    return TAGGED(0).AsUnsigned();
+}
+
+// Set code constant.  This can be a fast call.
+// This is in the RTS both because we pass a closure in here and cannot have
+// code addresses in 32-in-64 and also because we need to ensure there is no
+// possibility of a GC while the code is an inconsistent state.
+POLYUNSIGNED PolySetCodeConstant(PolyWord closure, PolyWord offset, PolyWord cWord, PolyWord flags)
+{
+    byte *pointer;
+    // Previously we passed the code address in here and we need to
+    // retain that for legacy code.  This is now the closure.
+    if (closure.AsObjPtr()->IsCodeObject())
+        pointer = closure.AsCodePtr();
+    else pointer = *(POLYCODEPTR*)(closure.AsObjPtr());
+    // pointer is the start of the code segment.
+    // c will usually be an address.
+    // offset is a byte offset
+    pointer += offset.UnTaggedUnsigned();
+    switch (UNTAGGED(flags))
+    {
+        case 0: // Absolute constant - size PolyWord
+        {
+            POLYUNSIGNED c = cWord.AsUnsigned();
+#ifdef WORDS_BIGENDIAN
+            // This is used to store constants in the constant area
+            // on the interpreted version. 
+            for (unsigned i = sizeof(PolyWord); i > 0; i--)
+            {
+                pointer[i-1] = (byte)(c & 255);
+                c >>= 8;
+            }
+#else
+            for (unsigned i = 0; i < sizeof(PolyWord); i++)
+            {
+                pointer[i] = (byte)(c & 255);
+                c >>= 8;
+            }
+#endif
+            break;
+        }
+        case 1: // Relative constant - X86 - size 4 bytes
+        {
+            // The offset is relative to the END of the constant.
+            byte *target;
+            // In 32-in-64 we pass in the closure address here
+            // rather than the code address.
+            if (cWord.AsObjPtr()->IsCodeObject())
+                target = cWord.AsCodePtr();
+            else target = *(POLYCODEPTR*)(cWord.AsObjPtr());
+            size_t c = target - pointer - 4;
+            for (unsigned i = 0; i < sizeof(PolyWord); i++)
+            {
+                pointer[i] = (byte)(c & 255);
+                c >>= 8;
+            }
+            break;
+        }
+    }
+    return TAGGED(0).AsUnsigned();
+}
+
+// Set a code byte.  This needs to be in the RTS because it uses the closure
+POLYEXTERNALSYMBOL POLYUNSIGNED PolySetCodeByte(PolyWord closure, PolyWord offset, PolyWord cWord)
+{
+    byte *pointer = *(POLYCODEPTR*)(closure.AsObjPtr());
+    pointer[UNTAGGED_UNSIGNED(offset)] = (byte)UNTAGGED_UNSIGNED(cWord);
+    return TAGGED(0).AsUnsigned();
+}
+
+POLYEXTERNALSYMBOL POLYUNSIGNED PolyGetCodeByte(PolyWord closure, PolyWord offset)
+{
+    byte *pointer = *(POLYCODEPTR*)(closure.AsObjPtr());
+    return TAGGED(pointer[UNTAGGED_UNSIGNED(offset)]).AsUnsigned();
+}
+
+static int compare(const void *a, const void *b)
+{
+    PolyWord *av = (PolyWord*)a;
+    PolyWord *bv = (PolyWord*)b;
+    if ((*av).IsTagged() || (*bv).IsTagged()) return 0; // Shouldn't happen
+    PolyObject *ao = (*av).AsObjPtr(), *bo = (*bv).AsObjPtr();
+    if (ao->Length() < 1 || bo->Length() < 1) return 0; // Shouldn't happen
+    if (ao->Get(0).AsUnsigned() < bo->Get(0).AsUnsigned())
+        return -1;
+    if (ao->Get(0).AsUnsigned() > bo->Get(0).AsUnsigned())
+        return 1;
+    return 0;
+}
+
+// Sort an array of addresses.  This is used in the code-generator to search for
+// duplicates in the address area.  The argument is an array of pairs.  The first
+// item of each pair is an address, the second is an identifier of some kind.
+POLYEXTERNALSYMBOL POLYUNSIGNED PolySortArrayOfAddresses(PolyWord array)
+{
+    if (!array.IsDataPtr()) return(TAGGED(0)).AsUnsigned();
+    PolyObject *arrayP = array.AsObjPtr();
+    POLYUNSIGNED numberOfItems = arrayP->Length();
+    if (!arrayP->IsMutable()) return(TAGGED(0)).AsUnsigned();
+    qsort(arrayP, numberOfItems, sizeof(PolyWord), compare);
+    return (TAGGED(1)).AsUnsigned();
+}
+
 struct _entrypts polySpecificEPT[] =
 {
     { "PolySpecificGeneral",            (polyRTSFunction)&PolySpecificGeneral},
     { "PolyGetABI",                     (polyRTSFunction)&PolyGetABI },
     { "PolyCopyByteVecToCode",          (polyRTSFunction)&PolyCopyByteVecToCode },
+    { "PolyCopyByteVecToClosure",       (polyRTSFunction)&PolyCopyByteVecToClosure },
     { "PolyLockMutableCode",            (polyRTSFunction)&PolyLockMutableCode },
+    { "PolyLockMutableClosure",         (polyRTSFunction)&PolyLockMutableClosure },
+    { "PolySetCodeConstant",            (polyRTSFunction)&PolySetCodeConstant },
+    { "PolySetCodeByte",                (polyRTSFunction)&PolySetCodeByte },
+    { "PolyGetCodeByte",                (polyRTSFunction)&PolyGetCodeByte },
+    { "PolySortArrayOfAddresses",       (polyRTSFunction)&PolySortArrayOfAddresses },
 
     { NULL, NULL} // End of list.
 };
