@@ -1,7 +1,7 @@
 /*
     Title:  New Foreign Function Interface
 
-    Copyright (c) 2015  David C.J. Matthews
+    Copyright (c) 2015, 2018, 2019  David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -75,9 +75,9 @@
 #include "processes.h"
 #include "polystring.h"
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #include <windows.h>
-#include "Console.h" /* For hApplicationInstance. */
+#include "winstartup.h" /* For hApplicationInstance. */
 #endif
 
 #include "scanaddrs.h"
@@ -94,6 +94,8 @@ extern "C" {
     POLYEXTERNALSYMBOL POLYUNSIGNED PolySizeDouble();
     POLYEXTERNALSYMBOL POLYUNSIGNED PolyFFIGetError(PolyWord addr);
     POLYEXTERNALSYMBOL POLYUNSIGNED PolyFFISetError(PolyWord err);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyFFICreateExtFn(PolyObject *threadId, PolyWord arg);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyFFICreateExtData(PolyObject *threadId, PolyWord arg);
 }
 
 static struct _abiTable { const char *abiName; ffi_abi abiCode; } abiTable[] =
@@ -108,9 +110,10 @@ static struct _abiTable { const char *abiName; ffi_abi abiCode; } abiTable[] =
     {"ms_cdecl", FFI_MS_CDECL},
 #elif defined(X86_WIN64)
     {"win64", FFI_WIN64},
+#elif defined(X86_64) || (defined (__x86_64__) && defined (X86_DARWIN))
+    {"unix64", FFI_UNIX64},
 #elif defined(X86_ANY)
     {"sysv", FFI_SYSV},
-    {"unix64", FFI_UNIX64},
 #endif
     { "default", FFI_DEFAULT_ABI}
 };
@@ -199,7 +202,7 @@ Handle poly_ffi(TaskData *taskData, Handle args, Handle code)
     case 2: // Load library
         {
             TempString libName(args->Word());
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             HINSTANCE lib = LoadLibrary(libName);
             if (lib == NULL)
             {
@@ -227,7 +230,7 @@ Handle poly_ffi(TaskData *taskData, Handle args, Handle code)
 
     case 3: // Load address of executable.
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             HINSTANCE lib = hApplicationInstance;
 #else
             void *lib = dlopen(NULL, RTLD_LAZY);
@@ -243,7 +246,7 @@ Handle poly_ffi(TaskData *taskData, Handle args, Handle code)
         }
     case 4: // Unload library - Is this actually going to be used?
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             HMODULE hMod = *(HMODULE*)(args->WordP());
             if (! FreeLibrary(hMod))
                 raise_syscall(taskData, "FreeLibrary failed", GetLastError());
@@ -262,7 +265,7 @@ Handle poly_ffi(TaskData *taskData, Handle args, Handle code)
     case 5: // Load the address of a symbol from a library.
         {
             TempCString symName(args->WordP()->Get(1));
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             HMODULE hMod = *(HMODULE*)(args->WordP()->Get(0).AsAddress());
             void *sym = (void*)GetProcAddress(hMod, symName);
             if (sym == NULL)
@@ -594,18 +597,18 @@ POLYUNSIGNED PolyFFIGeneral(PolyObject *threadId, PolyWord code, PolyWord arg)
 // These functions are needed in the compiler
 POLYUNSIGNED PolySizeFloat()
 {
-    return TAGGED(ffi_type_float.size).AsUnsigned();
+    return TAGGED((POLYSIGNED)ffi_type_float.size).AsUnsigned();
 }
 
 POLYUNSIGNED PolySizeDouble()
 {
-    return TAGGED(ffi_type_double.size).AsUnsigned();
+    return TAGGED((POLYSIGNED)ffi_type_double.size).AsUnsigned();
 }
 
 // Get either errno or GetLastError
 POLYUNSIGNED PolyFFIGetError(PolyWord addr)
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     addr.AsObjPtr()->Set(0, PolyWord::FromUnsigned(GetLastError()));
 #else
     addr.AsObjPtr()->Set(0, PolyWord::FromUnsigned((POLYUNSIGNED)errno));
@@ -616,7 +619,7 @@ POLYUNSIGNED PolyFFIGetError(PolyWord addr)
 // The argument is a SysWord.word value i.e. the address of a byte cell.
 POLYUNSIGNED PolyFFISetError(PolyWord err)
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     SetLastError((DWORD)(err.AsObjPtr()->Get(0).AsUnsigned()));
 #else
     errno = err.AsObjPtr()->Get(0).AsSigned();
@@ -624,6 +627,52 @@ POLYUNSIGNED PolyFFISetError(PolyWord err)
     return 0;
 }
 
+// Create an external function reference.  The value returned has space for
+// an address followed by the name of the external symbol.  Because the
+// address comes at the beginning it can be used in the same way as the
+// SysWord value returned by the get-symbol call from a library.
+POLYUNSIGNED PolyFFICreateExtFn(PolyObject *threadId, PolyWord arg)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedArg = taskData->saveVec.push(arg);
+    Handle result = 0;
+
+    try {
+        result = creatEntryPointObject(taskData, pushedArg, true);
+    }
+    catch (...) {} // If an ML exception is raised
+
+    taskData->saveVec.reset(reset); // Ensure the save vec is reset
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
+// Create an external reference to data.  On a small number of platforms
+// different forms of relocation are needed for data and for functions.
+POLYUNSIGNED PolyFFICreateExtData(PolyObject *threadId, PolyWord arg)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedArg = taskData->saveVec.push(arg);
+    Handle result = 0;
+
+    try {
+        result = creatEntryPointObject(taskData, pushedArg, false);
+    }
+    catch (...) {} // If an ML exception is raised
+
+    taskData->saveVec.reset(reset); // Ensure the save vec is reset
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
 struct _entrypts polyFFIEPT[] =
 {
     { "PolyFFIGeneral",                 (polyRTSFunction)&PolyFFIGeneral},
@@ -631,6 +680,8 @@ struct _entrypts polyFFIEPT[] =
     { "PolySizeDouble",                 (polyRTSFunction)&PolySizeDouble},
     { "PolyFFIGetError",                (polyRTSFunction)&PolyFFIGetError},
     { "PolyFFISetError",                (polyRTSFunction)&PolyFFISetError},
+    { "PolyFFICreateExtFn",             (polyRTSFunction)&PolyFFICreateExtFn},
+    { "PolyFFICreateExtData",           (polyRTSFunction)&PolyFFICreateExtData },
 
     { NULL, NULL} // End of list.
 };
