@@ -65,7 +65,7 @@
 #endif
 
 // Include this next before errors.h since in WinCE at least the winsock errors are defined there.
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #include <winsock2.h>
 #include <tchar.h>
 #define NOMEMORY        ERROR_NOT_ENOUGH_MEMORY
@@ -111,7 +111,7 @@ extern "C" {
 #define SAVE(x) mdTaskData->saveVec.push(x)
 #define ALLOC(n) alloc_and_save(mdTaskData, n)
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 #define ISPATHSEPARATOR(c)  ((c) == '\\' || (c) == '/')
 #define DEFAULTSEPARATOR    "\\"
 #else
@@ -166,7 +166,7 @@ void CygwinSpawnRequest::Perform()
 
 static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 {
-    unsigned c = get_C_unsigned(mdTaskData, DEREFWORDHANDLE(code));
+    unsigned c = get_C_unsigned(mdTaskData, DEREFWORD(code));
     switch (c)
     {
     case 0: /* Return the program name. */
@@ -203,7 +203,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
             TempString buff(args->Word());
             if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", NOMEMORY);
             int res = -1;
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             // Windows.
             TCHAR *argv[4];
             argv[0] = _tgetenv(_T("COMSPEC")); // Default CLI.
@@ -256,7 +256,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
                 try
                 {
                 // Test to see if the child has returned.
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
                     switch (WaitForSingleObject((HANDLE)pid, 0))
                     {
                     case WAIT_OBJECT_0:
@@ -294,7 +294,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
                     // Either IOException or KillException.
                     // We're abandoning the wait.  This will leave
                     // a zombie in Unix.
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
                     CloseHandle((HANDLE)pid);
 #endif
                     throw;
@@ -310,7 +310,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
             {
                 PolyObject *cell = alloc(mdTaskData, 2);
                 cell->Set(0, at_exit_list);
-                cell->Set(1, DEREFWORD(args));
+                cell->Set(1, args->Word());
                 at_exit_list = cell;
             }
             return Make_fixed_precision(mdTaskData, 0);
@@ -334,7 +334,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
         {
             /* I don't like terminating without some sort of clean up
                but we'll do it this way for the moment. */
-            int i = get_C_int(mdTaskData, DEREFWORDHANDLE(args));
+            int i = get_C_int(mdTaskData, args->Word());
             _exit(i);
         }
 
@@ -356,14 +356,14 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
 
     case 8: /* Test the character to see if it matches a separator. */
         {
-            int e = get_C_int(mdTaskData, DEREFWORDHANDLE(args));
+            int e = get_C_int(mdTaskData, args->Word());
             if (ISPATHSEPARATOR(e))
                 return Make_fixed_precision(mdTaskData, 1);
             else return Make_fixed_precision(mdTaskData, 0);
         }
 
     case 9: /* Are names case-sensitive? */
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
         /* Windows - no. */
         return Make_fixed_precision(mdTaskData, 0);
 #else
@@ -381,7 +381,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
             const TCHAR *volName = NULL;
             int  isAbs = 0;
             int  toRemove = 0;
-            PolyWord path = DEREFHANDLE(args);
+            PolyWord path = args->Word();
             /* This examines the start of a string and determines
                how much of it represents the volume name and returns
                the number of characters to remove, the volume name
@@ -391,7 +391,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
                (e.g. A:b\c) which means the file b\c relative to the
                currently selected directory on the volume A.
             */
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             TempString buff(path);
             if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", NOMEMORY);
             size_t length = _tcslen(buff);
@@ -446,7 +446,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
                 Handle sVol = SAVE(C_string_to_Poly(mdTaskData, volName));
                 Handle sRes = ALLOC(3);
                 DEREFWORDHANDLE(sRes)->Set(0, TAGGED(toRemove));
-                DEREFHANDLE(sRes)->Set(1, DEREFWORDHANDLE(sVol));
+                DEREFHANDLE(sRes)->Set(1, sVol->Word());
                 DEREFWORDHANDLE(sRes)->Set(2, TAGGED(isAbs));
                 return sRes;
             }
@@ -484,7 +484,7 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
                         return Make_fixed_precision(mdTaskData, 0);
                 }
             }
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             // We need to look for certain invalid characters but only after
             // we've converted it to Unicode if necessary.
             TempString name(volName);
@@ -511,14 +511,6 @@ static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle c
 #endif
         }
 
-        // These were supposed to have been moved to poly-specific but don't seem to have been.
-    case 100: /* Return the maximum word segment size. */ // Legacy - used in bootstrap
-            return mdTaskData->saveVec.push(TAGGED(MAX_OBJECT_SIZE));
-    case 101: /* Return the maximum string size (in bytes).
-                 It is the maximum number of bytes in a segment
-                 less one word for the length field. */  // Legacy - used in bootstrap
-            return mdTaskData->saveVec.push(TAGGED((MAX_OBJECT_SIZE)*sizeof(PolyWord) - sizeof(PolyWord)));
-
     case 104: return Make_arbitrary_precision(mdTaskData, POLY_version_number);
  
     case 105: /* Get the name of the function. */
@@ -580,8 +572,8 @@ void PolyFinish(PolyObject *threadId, PolyWord arg)
     ASSERT(taskData != 0);
     taskData->PreRTSCall();
     int i = get_C_int(taskData, arg);
-    // Cause the other threads to exit.
-    processes->Exit(i);
+    // Cause the other threads to exit and set the result code.
+    processes->RequestProcessExit(i);
     // Exit this thread
     processes->ThreadExit(taskData); // Doesn't return.
 }
@@ -705,7 +697,15 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyGetFunctionName(PolyObject *threadId, PolyWo
     Handle result = 0;
 
     try {
+        if (fnAddr.IsTagged()) raise_fail(taskData, "Not a code pointer");
         PolyObject *pt = fnAddr.AsObjPtr();
+        // In 32-in-64 this may be a closure and the first word is the absolute address of the code.
+        if (pt->IsClosureObject())
+        {
+            // It may not be set yet.
+            pt = *(PolyObject**)pt;
+            if (((uintptr_t)pt & 1) == 1) raise_fail(taskData, "Not a code pointer");
+        }
         if (pt->IsCodeObject()) /* Should now be a code object. */
         {
             /* Compiled code.  This is the first constant in the constant area. */
@@ -716,7 +716,7 @@ POLYEXTERNALSYMBOL POLYUNSIGNED PolyGetFunctionName(PolyObject *threadId, PolyWo
                 result = taskData->saveVec.push(C_string_to_Poly(taskData, ""));
             else result = taskData->saveVec.push(name);
         }
-        else raise_syscall(taskData, "Not a code pointer", 0);
+        else raise_fail(taskData, "Not a code pointer");
     }
     catch (...) {} // If an ML exception is raised
 
