@@ -2,7 +2,7 @@
     Title:      Lightweight process library
     Author:     David C.J. Matthews
 
-    Copyright (c) 2007-8, 2012, 2015, 2017 David C.J. Matthews
+    Copyright (c) 2007-8, 2012, 2015, 2017, 2019 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -82,8 +82,7 @@ typedef void *HANDLE;
 // are read and set by the ML code.
 class ThreadObject: public PolyObject {
 public:
-    PolyWord    index;  // Tagged integer with the index into the taskArray
-                        // Not used by ML
+    PolyWord    threadRef;  // Weak ref containing the address of the thread data. Not used by ML
     PolyWord    flags;  // Tagged integer containing flags indicating how interrupts
                         // are handled.  Set by ML but only by the thread itself
     PolyWord    threadLocal; // Head of a list of thread-local store items.
@@ -131,32 +130,37 @@ public:
     // atomic increment and decrement instructions.
     virtual void AtomicReset(Handle mutexp) = 0;
 
-    virtual void CopyStackFrame(StackObject *old_stack, POLYUNSIGNED old_length,
-                                StackObject *new_stack, POLYUNSIGNED new_length) = 0;
+    virtual void CopyStackFrame(StackObject *old_stack, uintptr_t old_length,
+                                StackObject *new_stack, uintptr_t new_length) = 0;
 
 
-    virtual POLYUNSIGNED currentStackSpace(void) const = 0;
+    virtual uintptr_t currentStackSpace(void) const = 0;
     // Add a count to the local function if we are using store profiling.
     virtual void addProfileCount(POLYUNSIGNED words) = 0;
 
     // Functions called before and after an RTS call.
-    virtual void PreRTSCall(void) {}
-    virtual void PostRTSCall(void) {}
+    virtual void PreRTSCall(void) { inML = false; }
+    virtual void PostRTSCall(void) { inML = true; }
 
     SaveVec     saveVec;
     PolyWord    *allocPointer;  // Allocation pointer - decremented towards...
     PolyWord    *allocLimit;    // ... lower limit of allocation
-    POLYUNSIGNED allocSize;     // The preferred heap segment size
+    uintptr_t   allocSize;     // The preferred heap segment size
     unsigned    allocCount;     // The number of allocations since the last GC
     StackSpace  *stack;
     ThreadObject *threadObject;  // Pointer to the thread object.
     int         lastError;      // Last error from foreign code.
     void        *signalStack;  // Stack to handle interrupts (Unix only)
-    PolyWord    foreignStack;   // Stack of saved data used in call_sym_and_convert
     bool        inML;          // True when this is in ML, false in the RTS
 
     // Get a TaskData pointer given the ML taskId.
-    static TaskData *FindTaskForId(PolyObject *taskId);
+    // This is called at the start of every RTS function that may allocate memory.
+    // It is can be called safely to get the thread's own TaskData object without
+    // a lock but any call to get the TaskData for another thread must take the
+    // schedLock first in case the thread is exiting.
+    static TaskData *FindTaskForId(PolyObject *taskId) {
+        return *(TaskData**)(((ThreadObject*)taskId)->threadRef.AsObjPtr());
+    }
 
 private:
     // If a thread has to block it will block on this.
@@ -253,12 +257,9 @@ public:
     virtual ~Waiter() {}
     virtual void Wait(unsigned maxMillisecs);
     static Waiter *defaultWaiter;
-#ifdef HAVE_WINDOWS_H
-    static HANDLE hWakeupEvent;
-#endif
 };
 
-#ifdef HAVE_WINDOWS_H
+#ifdef _WIN32
 class WaitHandle: public Waiter
 {
 public:
@@ -267,9 +268,9 @@ public:
 private:
     HANDLE m_Handle;
 };
-#endif
 
-#if (! defined(_WIN32) || defined(__CYGWIN__))
+#else
+
 // Unix: Wait until a file descriptor is available for input
 class WaitInputFD: public Waiter
 {
@@ -294,7 +295,7 @@ public:
                            Handle args, PolyWord flags) = 0;
     // Request all ML threads to exit and set the result code.  Does not cause
     // the calling thread itself to exit since this may be called on the GUI thread.
-    virtual void Exit(int n) = 0;
+    virtual void RequestProcessExit(int n) = 0;
     // Exit from this thread.
     virtual NORETURNFN(void ThreadExit(TaskData *taskData)) = 0;
 
