@@ -32,10 +32,11 @@ typedef enum {
     kProfileOff = 0,
     kProfileTime,
     kProfileStoreAllocation,
-    kProfileEmulation,
+    kProfileEmulation,   // No longer used
     kProfileLiveData,
     kProfileLiveMutables,
-    kProfileTimeThread
+    kProfileTimeThread,
+    kProfileMutexContention
 } ProfileMode;
 
 extern ProfileMode profileMode;
@@ -43,8 +44,9 @@ extern ProfileMode profileMode;
 #include "processes.h" // For SIGNALCONTEXT
 
 extern void handleProfileTrap(TaskData *taskData, SIGNALCONTEXT *context);
-extern Handle profilerc(TaskData *taskData, Handle mode_handle);
-extern void add_count(TaskData *taskData, POLYCODEPTR pc, PolyWord *sp, POLYUNSIGNED incr);
+extern void add_count(TaskData *taskData, POLYCODEPTR pc,POLYUNSIGNED incr);
 extern void AddObjectProfile(PolyObject *obj);
 
+extern struct _entrypts profilingEPT[];
+
 #endif /* _PROFILING_H_DEFINED */
