@@ -1,7 +1,7 @@
 /*
     Title:      Quick copying garbage collector
 
-    Copyright (c) 2011-12, 2016 David C. J. Matthews
+    Copyright (c) 2011-12, 2016-17 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -135,21 +135,27 @@ public:
 #   endif
 #endif
 
-static bool atomiclySetForwarding(LocalMemSpace *space, POLYUNSIGNED *pt,
-                                  POLYUNSIGNED testVal, POLYUNSIGNED update)
+#ifdef POLYML32IN64
+typedef uint32_t ptrasint;
+#else
+typedef uintptr_t ptrasint;
+#endif
+
+
+static bool atomiclySetForwarding(LocalMemSpace *space, ptrasint *pt, ptrasint testVal, ptrasint update)
 {
 #ifdef _MSC_VER
-# if (SIZEOF_VOIDP == 8)
+# if (SIZEOF_POLYWORD == 8)
     LONGLONG *address = (LONGLONG*)(pt-1);
-    POLYUNSIGNED result = InterlockedCompareExchange64(address, update, testVal);
+    uintptr_t result = InterlockedCompareExchange64(address, update, testVal);
     return result == testVal;
 # else
     LONG *address = (LONG*)(pt-1);
-    POLYUNSIGNED result = InterlockedCompareExchange(address, update, testVal);
+    uintptr_t result = InterlockedCompareExchange(address, update, testVal);
     return result == testVal;
 # endif
-#elif((defined(HOSTARCHITECTURE_X86) || defined(HOSTARCHITECTURE_X32)) && defined(__GNUC__))
-    POLYUNSIGNED result;
+#elif((defined(HOSTARCHITECTURE_X86) || defined(HOSTARCHITECTURE_X32) || defined(POLYML32IN64)) && defined(__GNUC__))
+    uintptr_t result;
     __asm__ __volatile__ (
         "lock; cmpxchgl %1,%2"
         :"=a"(result)
@@ -158,7 +164,7 @@ static bool atomiclySetForwarding(LocalMemSpace *space, POLYUNSIGNED *pt,
     );
     return result == testVal;
 #elif(defined(HOSTARCHITECTURE_X86_64) && defined(__GNUC__))
-    POLYUNSIGNED result;
+    uintptr_t result;
     __asm__ __volatile__ (
         "lock; cmpxchgq %1,%2"
         :"=a"(result)
@@ -201,7 +207,7 @@ PolyObject *QuickGCScanner::FindNewAddress(PolyObject *obj, POLYUNSIGNED L, Loca
     // be worth-while.
     if (isMutable || OBJ_IS_CODE_OBJECT(L))
     {
-        if (! atomiclySetForwarding(srcSpace, (POLYUNSIGNED*)obj, L, OBJ_SET_POINTER(newObject)))
+        if (! atomiclySetForwarding(srcSpace, (ptrasint*)obj, L, OBJ_SET_POINTER(newObject)))
         {
             newObject = obj->GetForwardingPtr();
             if (debugOptions & DEBUG_GC_DETAIL)
@@ -224,6 +230,14 @@ PolyObject *QuickGCScanner::FindNewAddress(PolyObject *obj, POLYUNSIGNED L, Loca
     }
 
     lSpace->lowerAllocPtr += n+1;
+#ifdef POLYML32IN64
+    // Maintain the odd-word alignment of lowerAllocPtr
+    if ((n & 1) == 0 && lSpace->lowerAllocPtr < lSpace->upperAllocPtr)
+    {
+        *lSpace->lowerAllocPtr = PolyWord::FromUnsigned(0);
+        lSpace->lowerAllocPtr++;
+    }
+#endif
     CopyObjectToNewAddress(obj, newObject, L);
     objectCopied = true;
     return newObject;
@@ -393,6 +407,14 @@ POLYUNSIGNED QuickGCScanner::ScanAddressAt(PolyWord *pt)
 // in objects that can't be handled by ScanAddressAt.
 PolyObject *QuickGCScanner::ScanObjectAddress(PolyObject *base)
 {
+#ifdef POLYML32IN64
+    // If this is a code address we can't turn it into a PolyWord.
+    // Check that it's a local address.
+    MemSpace *space = gMem.SpaceForAddress((PolyWord*)base - 1);
+    ASSERT(space != 0);
+    if (space->spaceType != ST_LOCAL)
+        return base;
+#endif
     PolyWord val = base;
     // Scan this as an address.
     (void)QuickGCScanner::ScanAddressAt(&val);
@@ -457,6 +479,13 @@ void ThreadScanner::ScanOwnedAreas()
                     PolyWord *p = space->partialGCScan;
                     while (p < mid)
                     {
+#ifdef POLYML32IN64
+                        if ((((uintptr_t)p) & 4) == 0)
+                        {
+                            p++; // Should be on an odd-word boundary
+                            continue;
+                        }
+#endif
                         PolyObject *o = (PolyObject*)(p+1);
                         ASSERT(o->ContainsNormalLengthWord());
                         p += o->Length()+1;
@@ -473,6 +502,13 @@ void ThreadScanner::ScanOwnedAreas()
                     }
                 }
                 PolyObject *obj = (PolyObject*)(space->partialGCScan+1);
+#ifdef POLYML32IN64
+                if ((((uintptr_t)obj) & 4) != 0)  // Should be on an even-word boundary
+                {
+                    space->partialGCScan++;
+                    continue;
+                }
+#endif
                 ASSERT(obj->ContainsNormalLengthWord());
                 POLYUNSIGNED length = obj->Length();
                 ASSERT(space->partialGCScan+length+1 <= space->lowerAllocPtr);
@@ -511,7 +547,7 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
     if (debugOptions & DEBUG_HEAPSIZE)
         gMem.ReportHeapSizes("Minor GC (before)");
 
-    POLYUNSIGNED spaceBeforeGC = 0;
+    uintptr_t spaceBeforeGC = 0;
 
     for(std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
@@ -618,7 +654,7 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
 
     gpTaskFarm->WaitForCompletion();
 
-    POLYUNSIGNED spaceAfterGC = 0;
+    uintptr_t spaceAfterGC = 0;
 
     if (succeeded)
     {
@@ -629,10 +665,15 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
         for(std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
         {
             LocalMemSpace *lSpace = *i;
-            POLYUNSIGNED free;
+            uintptr_t free;
             if (lSpace->allocationSpace)
             {
+#ifdef POLYML32IN64
+                lSpace->lowerAllocPtr = lSpace->bottom + 1;
+                lSpace->lowerAllocPtr[-1] = PolyWord::FromUnsigned(0);
+#else
                 lSpace->lowerAllocPtr = lSpace->bottom;
+#endif
                 free = lSpace->freeSpace();
 #ifdef FILL_UNUSED_MEMORY
                 // This provides extra checking if we have dangling pointers
@@ -644,7 +685,7 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
             else free = lSpace->freeSpace();
 
             if (debugOptions & DEBUG_GC_ENHANCED)
-                Log("GC: %s space %p %d free in %d words %2.1f%% full\n", lSpace->spaceTypeString(),
+                Log("GC: %s space %p %" PRI_SIZET " free in %" PRI_SIZET " words %2.1f%% full\n", lSpace->spaceTypeString(),
                     lSpace, lSpace->freeSpace(), lSpace->spaceSize(),
                     ((float)lSpace->allocatedSpace()) * 100 / (float)lSpace->spaceSize());
             globalStats.incSize(PSS_AFTER_LAST_GC, free*sizeof(PolyWord));
