@@ -5,7 +5,7 @@
     Copyright (c) 2000
         Cambridge University Technical Services Limited
 
-    Further work copyright David C. J. Matthews 2009, 2012, 2015-16
+    Further work copyright David C. J. Matthews 2009, 2012, 2015-18
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -68,17 +68,15 @@ extern "C" {
 #define SIZEOF(x) (sizeof(x)/sizeof(PolyWord))
 
 
-/******************************************************************************/
-/*                                                                            */
-/*      STORAGE ALLOCATION                                                    */
-/*                                                                            */
-/******************************************************************************/
-
 // This is the storage allocator for allocating heap objects in the RTS.
-PolyObject *alloc(TaskData *taskData, POLYUNSIGNED data_words, unsigned flags)
+PolyObject *alloc(TaskData *taskData, uintptr_t data_words, unsigned flags)
 /* Allocate a number of words. */
 {
-    POLYUNSIGNED words = data_words + 1;
+    // Check the size.  This might possibly happen with a long string.
+    if (data_words > MAX_OBJECT_SIZE)
+        raise_exception0(taskData, EXC_size);
+
+    POLYUNSIGNED words = (POLYUNSIGNED)data_words + 1;
     
     if (profileMode == kProfileStoreAllocation)
         taskData->addProfileCount(words);
@@ -91,7 +89,7 @@ PolyObject *alloc(TaskData *taskData, POLYUNSIGNED data_words, unsigned flags)
     }
 
     PolyObject *pObj = (PolyObject*)(foundSpace + 1);
-    pObj->SetLengthWord(data_words, flags);
+    pObj->SetLengthWord((POLYUNSIGNED)data_words, flags);
     
     // Must initialise object here, because GC doesn't clean store.
     // Is this necessary any more?  This used to be necessary when we used
@@ -101,15 +99,10 @@ PolyObject *alloc(TaskData *taskData, POLYUNSIGNED data_words, unsigned flags)
     return pObj;
 }
 
-/******************************************************************************/
-/*                                                                            */
-/*      alloc_and_save - called by run-time system                            */
-/*                                                                            */
-/******************************************************************************/
-Handle alloc_and_save(TaskData *taskData, POLYUNSIGNED size, unsigned flags)
+Handle alloc_and_save(TaskData *taskData, uintptr_t size, unsigned flags)
 /* Allocate and save the result on the vector. */
 {
-    return SAVE(alloc(taskData, size, flags));
+    return taskData->saveVec.push(alloc(taskData, size, flags));
 }
 
 POLYUNSIGNED PolyFullGC(PolyObject *threadId)
@@ -140,7 +133,7 @@ POLYUNSIGNED PolyFullGC(PolyObject *threadId)
 // anything.
 Handle errorMsg(TaskData *taskData, int err)
 {
-#ifdef _WIN32
+#if (defined(_WIN32))
     LPTSTR lpMsg = NULL;
     TCHAR *p;
     if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
@@ -203,8 +196,8 @@ static Handle make_exn(TaskData *taskData, int id, Handle arg, const char *fileN
     }
     
     DEREFEXNHANDLE(exnHandle)->ex_id   = TAGGED(id);
-    DEREFEXNHANDLE(exnHandle)->ex_name = DEREFWORD(pushed_name);
-    DEREFEXNHANDLE(exnHandle)->arg     = DEREFWORDHANDLE(arg);
+    DEREFEXNHANDLE(exnHandle)->ex_name = pushed_name->Word();
+    DEREFEXNHANDLE(exnHandle)->arg     = arg->Word();
     DEREFEXNHANDLE(exnHandle)->ex_location = location->Word();
 
     return exnHandle;
@@ -252,8 +245,8 @@ void raiseSycallWithLocation(TaskData *taskData, const char *errmsg, int err, co
         Handle pushed_option = SAVE(NONE_VALUE); /* NONE */
         Handle pushed_name = SAVE(C_string_to_Poly(taskData, errmsg));
         Handle pair = alloc_and_save(taskData, 2);
-        DEREFHANDLE(pair)->Set(0, DEREFWORDHANDLE(pushed_name));
-        DEREFHANDLE(pair)->Set(1, DEREFWORDHANDLE(pushed_option));
+        DEREFHANDLE(pair)->Set(0, pushed_name->Word());
+        DEREFHANDLE(pair)->Set(1, pushed_option->Word());
 
         raise_exception(taskData, EXC_syserr, pair, file, line);
     }
@@ -261,11 +254,11 @@ void raiseSycallWithLocation(TaskData *taskData, const char *errmsg, int err, co
     {
         Handle errornum = Make_sysword(taskData, err);
         Handle pushed_option = alloc_and_save(taskData, 1);
-        DEREFHANDLE(pushed_option)->Set(0, DEREFWORDHANDLE(errornum)); /* SOME err */
+        DEREFHANDLE(pushed_option)->Set(0, errornum->Word()); /* SOME err */
         Handle pushed_name = errorMsg(taskData, err); // Generate the string.
         Handle pair = alloc_and_save(taskData, 2);
-        DEREFHANDLE(pair)->Set(0, DEREFWORDHANDLE(pushed_name));
-        DEREFHANDLE(pair)->Set(1, DEREFWORDHANDLE(pushed_option));
+        DEREFHANDLE(pair)->Set(0, pushed_name->Word());
+        DEREFHANDLE(pair)->Set(1, pushed_option->Word());
 
         raise_exception(taskData, EXC_syserr, pair, file, line);
     }
@@ -291,31 +284,31 @@ Handle makeList(TaskData *taskData, int count, char *p, int size, void *arg,
         value = mkEntry(taskData, arg, p);
         next  = alloc_and_save(taskData, SIZEOF(ML_Cons_Cell));
 
-        DEREFLISTHANDLE(next)->h = DEREFWORDHANDLE(value); 
-        DEREFLISTHANDLE(next)->t = DEREFLISTHANDLE(list);
+        DEREFLISTHANDLE(next)->h = value->Word();
+        DEREFLISTHANDLE(next)->t = list->Word();
 
         taskData->saveVec.reset(saved);
-        list = SAVE(DEREFHANDLE(next));
+        list = SAVE(next->Word());
         count--;
     }
     return list;
 }
 
-void CheckAndGrowStack(TaskData *taskData, POLYUNSIGNED minSize)
+void CheckAndGrowStack(TaskData *taskData, uintptr_t minSize)
 /* Expands the current stack if it has grown. We cannot shrink a stack segment
    when it grows smaller because the frame is checked only at the beginning of
    a function to ensure that there is enough space for the maximum that can
    be allocated. */
 {
     /* Get current size of new stack segment. */
-    POLYUNSIGNED old_len = taskData->stack->spaceSize();
+    uintptr_t old_len = taskData->stack->spaceSize();
 
     if (old_len >= minSize) return; /* Ok with present size. */
 
     // If it is too small double its size.
-    POLYUNSIGNED new_len; /* New size */
+    uintptr_t new_len; /* New size */
     for (new_len = old_len; new_len < minSize; new_len *= 2);
-    POLYUNSIGNED limitSize = getPolyUnsigned(taskData, taskData->threadObject->mlStackSize);
+    uintptr_t limitSize = getPolyUnsigned(taskData, taskData->threadObject->mlStackSize);
 
     // Do not grow the stack if its size is already too big.
     if ((limitSize != 0 && old_len >= limitSize) || ! gMem.GrowOrShrinkStack(taskData, new_len))
@@ -369,24 +362,39 @@ Handle Make_fixed_precision(TaskData *taskData, long long val)
 {
     if (val > MAXTAGGED || val < -MAXTAGGED-1)
         raise_exception0(taskData, EXC_overflow);
-    return taskData->saveVec.push(TAGGED(val));
+    return taskData->saveVec.push(TAGGED((POLYSIGNED)val));
 }
 
 Handle Make_fixed_precision(TaskData *taskData, unsigned long long uval)
 {
     if (uval > MAXTAGGED)
         raise_exception0(taskData, EXC_overflow);
-    return taskData->saveVec.push(TAGGED(uval));
+    return taskData->saveVec.push(TAGGED((POLYUNSIGNED)uval));
 }
 #endif
 
 Handle Make_sysword(TaskData *taskData, uintptr_t p)
 {
-    Handle result = alloc_and_save(taskData, 1, F_BYTE_OBJ);
+    Handle result = alloc_and_save(taskData, sizeof(uintptr_t)/sizeof(PolyWord), F_BYTE_OBJ);
     *(uintptr_t*)(result->Word().AsCodePtr()) = p;
     return result;
 }
 
+// A volatile ref is used for data that is not valid in a different session.
+// When loaded from a saved state it is cleared to zero.
+Handle MakeVolatileWord(TaskData *taskData, void *p)
+{
+    Handle result = alloc_and_save(taskData,
+            WORDS(SIZEOF_VOIDP), F_BYTE_OBJ | F_WEAK_BIT | F_MUTABLE_BIT | F_NO_OVERWRITE);
+    *(void**)(result->Word().AsCodePtr()) = p;
+    return result;
+}
+
+Handle MakeVolatileWord(TaskData *taskData, uintptr_t p)
+{
+    return MakeVolatileWord(taskData, (void*)p);
+}
+
 // This is used to determine the endian-ness that Poly/ML is running under.
 // It's really only needed for the interpreter.  In particular the pre-built
 // compiler may be running under either byte order and has to check at
