@@ -53,8 +53,8 @@ public:
 };
 
 /* storage allocation functions */
-extern PolyObject *alloc(TaskData *taskData, POLYUNSIGNED words, unsigned flags = 0);
-extern Handle alloc_and_save(TaskData *taskData, POLYUNSIGNED words, unsigned flags = 0);
+extern PolyObject *alloc(TaskData *taskData, uintptr_t words, unsigned flags = 0);
+extern Handle alloc_and_save(TaskData *taskData, uintptr_t words, unsigned flags = 0);
 
 extern Handle makeList(TaskData *taskData, int count, char *p, int size, void *arg,
                        Handle (mkEntry)(TaskData *, void*, char*));
@@ -79,7 +79,7 @@ NORETURNFN(extern void raiseSycallWithLocation(TaskData *taskData, const char *e
 poly_exn *makeExceptionPacket(TaskData *taskData, int id);
 
 // Check to see that there is space in the stack.  May GC and may raise a C++ exception.
-extern void CheckAndGrowStack(TaskData *mdTaskData, POLYUNSIGNED minSize);
+extern void CheckAndGrowStack(TaskData *mdTaskData, uintptr_t minSize);
 
 extern Handle errorMsg(TaskData *taskData, int err);
 
@@ -96,6 +96,8 @@ extern Handle Make_fixed_precision(TaskData *taskData, unsigned long long);
 #endif
 
 extern Handle Make_sysword(TaskData *taskData, uintptr_t p);
+extern Handle MakeVolatileWord(TaskData *taskData, void *p);
+extern Handle MakeVolatileWord(TaskData *taskData, uintptr_t p);
 
 extern struct _entrypts runTimeEPT[];
 
