@@ -1,7 +1,7 @@
 /*
     Title:      Address scanner
 
-    Copyright (c) 2006-8, 2012 David C.J. Matthews
+    Copyright (c) 2006-8, 2012, 2019 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -84,7 +84,21 @@ void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord
             // Skip to the constants and get ready to scan them.
             obj->GetConstSegmentForCode(length, baseAddr, length);
 
-        } // else it's a normal object,
+        }
+
+        else if (OBJ_IS_CLOSURE_OBJECT(lengthWord))
+        {
+            // The first word is a code pointer so we need to treat it specially
+            // but it is possible it hasn't yet been set.
+            if ((*(uintptr_t*)baseAddr & 1) == 0)
+            {
+                POLYUNSIGNED lengthWord = ScanCodeAddressAt((PolyObject**)baseAddr); // N.B.  This could side-effect *baseAddr
+                if (lengthWord != 0)
+                    ScanAddressesInObject(*(PolyObject**)baseAddr, lengthWord);
+            }
+            baseAddr += sizeof(PolyObject*) / sizeof(PolyWord);
+            length -= sizeof(PolyObject*) / sizeof(PolyWord);
+        }
 
         PolyWord *endWord = baseAddr + length;
 
@@ -149,6 +163,14 @@ void ScanAddress::ScanAddressesInRegion(PolyWord *region, PolyWord *end)
     PolyWord *pt = region;
     while (pt < end)
     {
+#ifdef POLYML32IN64
+        if ((((uintptr_t)pt) & 4) == 0)
+        {
+            // Skip any padding.  The length word should be on an odd-word boundary.
+            pt++;
+            continue;
+        }
+#endif
         pt++; // Skip length word.
         // pt actually points AT the object here.
         PolyObject *obj = (PolyObject*)pt;
@@ -176,7 +198,7 @@ void ScanAddress::ScanAddressesInRegion(PolyWord *region, PolyWord *end)
 }
 
 // Extract a constant from the code.
-PolyWord ScanAddress::GetConstantValue(byte *addressOfConstant, ScanRelocationKind code)
+PolyObject *ScanAddress::GetConstantValue(byte *addressOfConstant, ScanRelocationKind code, PolyWord *base)
 {
     switch (code)
     {
@@ -188,8 +210,9 @@ PolyWord ScanAddress::GetConstantValue(byte *addressOfConstant, ScanRelocationKi
             if (pt[3] & 0x80) valu = 0-1; else valu = 0;
             for (i = sizeof(PolyWord); i > 0; i--)
                 valu = (valu << 8) | pt[i-1];
-
-            return PolyWord::FromUnsigned(valu);
+            if (valu == 0 || PolyWord::FromUnsigned(valu).IsTagged())
+                return 0;
+            else return PolyWord::FromUnsigned(valu).AsObjPtr(base);
         }
     case PROCESS_RELOC_I386RELATIVE:         // 32 bit relative address
         {
@@ -198,27 +221,24 @@ PolyWord ScanAddress::GetConstantValue(byte *addressOfConstant, ScanRelocationKi
             // Get the displacement. This is signed.
             if (pt[3] & 0x80) disp = -1; else disp = 0; // Set the sign just in case.
             for(unsigned i = 4; i > 0; i--) disp = (disp << 8) | pt[i-1];
-
             byte *absAddr = pt + disp + 4; // The address is relative to AFTER the constant
-
-            return PolyWord::FromCodePtr(absAddr);
+            return (PolyObject*)absAddr;
         }
     default:
         ASSERT(false);
-        return TAGGED(0);
+        return 0;
     }
 }
 
 // Store a constant value.  Also used with a patch table when importing a saved heap which has
 // been exported using the C exporter.
-void ScanAddress::SetConstantValue(byte *addressOfConstant, PolyWord p, ScanRelocationKind code)
+void ScanAddress::SetConstantValue(byte *addressOfConstant, PolyObject *p, ScanRelocationKind code)
 {
-
     switch (code)
     {
     case PROCESS_RELOC_DIRECT: // 32 or 64 bit address of target
         {
-            POLYUNSIGNED valu = p.AsUnsigned();
+            POLYUNSIGNED valu = ((PolyWord)p).AsUnsigned();
             for (unsigned i = 0; i < sizeof(PolyWord); i++)
             {
                 addressOfConstant[i] = (byte)(valu & 255); 
@@ -228,9 +248,10 @@ void ScanAddress::SetConstantValue(byte *addressOfConstant, PolyWord p, ScanRelo
         break;
     case PROCESS_RELOC_I386RELATIVE:         // 32 bit relative address
         {
-            POLYSIGNED newDisp = p.AsCodePtr() - addressOfConstant - 4;
+            // This offset may be positive or negative
+            intptr_t newDisp = (byte*)p - addressOfConstant - 4;
 #if (SIZEOF_VOIDP != 4)
-            ASSERT(newDisp < 0x80000000 && newDisp >= -(POLYSIGNED)0x80000000);
+            ASSERT(newDisp < (intptr_t)0x80000000 && newDisp >= -(intptr_t)0x80000000);
 #endif
             for (unsigned i = 0; i < 4; i++) {
                 addressOfConstant[i] = (byte)(newDisp & 0xff);
@@ -241,16 +262,17 @@ void ScanAddress::SetConstantValue(byte *addressOfConstant, PolyWord p, ScanRelo
     }
 }
 
-// The default action is to call the DEFAULT ScanAddressAt NOT the virtual which means that it calls
-// ScanObjectAddress for the base address of the object referred to.
 void ScanAddress::ScanConstant(PolyObject *base, byte *addressOfConstant, ScanRelocationKind code)
 {
-    PolyWord p = GetConstantValue(addressOfConstant, code);
+    PolyObject *p = GetConstantValue(addressOfConstant, code);
 
-    if (! IS_INT(p))
+    if (p != 0)
     {
-        PolyWord oldValue = p;
-        ScanAddress::ScanAddressAt(&p);
+        PolyObject *oldValue = p;
+        // If this was a relative address we must have a code address.
+        if (code == PROCESS_RELOC_I386RELATIVE)
+            ScanCodeAddressAt(&p);
+        else p = ScanObjectAddress(p);
         if (p != oldValue) // Update it if it has changed.
             SetConstantValue(addressOfConstant, p, code);
     }
@@ -298,6 +320,7 @@ PolyObject *RecursiveScan::ScanObjectAddress(PolyObject *obj)
 // This is called via ScanAddressesInRegion to process the permanent mutables.  It is
 // also called from ScanObjectAddress to process root addresses.
 // It processes all the addresses reachable from the object.
+// This is almost the same as MTGCProcessMarkPointers::ScanAddressesInObject. 
 void RecursiveScan::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord)
 {
     if (OBJ_IS_BYTE_OBJECT(lengthWord))
@@ -313,9 +336,10 @@ void RecursiveScan::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWo
         // these will be side-effected by GetConstSegmentForCode.
         POLYUNSIGNED length = OBJ_OBJECT_LENGTH(lengthWord);
 
-        if (OBJ_IS_CODE_OBJECT(lengthWord))
+        if (OBJ_IS_CODE_OBJECT(lengthWord) || OBJ_IS_CLOSURE_OBJECT(lengthWord))
         {
             // It's better to process the whole code object in one go.
+            // For the moment do that for closure objects as well.
             ScanAddress::ScanAddressesInObject(obj, lengthWord);
             length = 0; // Finished
         }
@@ -373,6 +397,9 @@ void RecursiveScan::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWo
         if (baseAddr == endWord)
         {
             // We have done everything except possibly firstWord and secondWord.
+            // Note: Unfortunately the way that ScanAddressesInRegion works means that
+            // we call Completed on the addresses of cells in the permanent areas without
+            // having called TestForScan.
             Completed(obj);
             if (secondWord != 0)
             {
