@@ -43,6 +43,8 @@
 #include <string.h>
 #endif
 
+#include <vector>
+
 #include "globals.h"
 #include "save_vec.h"
 #include "machine_dep.h"
@@ -121,6 +123,15 @@ extern "C" {
     POLYEXTERNALSYMBOL POLYUNSIGNED PolyShareCommonData(PolyObject *threadId, PolyWord root);
 }
 
+// The depth is stored in the length field.  If the Weak bit is set but the Mutable bit
+// is clear the value in the length word is a depth rather than a real length.
+// The tombstone bit is zero.
+// Previously "depth" values were encoded with the tombstone bit set but that isn't
+// possible in 32-in-64 because we need 31 bits in a forwarding pointer.
+inline bool OBJ_IS_DEPTH(POLYUNSIGNED L) { return (L & (_OBJ_WEAK_BIT| _OBJ_MUTABLE_BIT)) == _OBJ_WEAK_BIT; }
+inline POLYUNSIGNED OBJ_GET_DEPTH(POLYUNSIGNED L) { return OBJ_OBJECT_LENGTH(L); }
+inline POLYUNSIGNED OBJ_SET_DEPTH(POLYUNSIGNED n) { return n | _OBJ_WEAK_BIT; }
+
 // The DepthVector type contains all the items of a particular depth.
 // This is the abstract class.  There are variants for the case where all
 // the cells have the same size and where they may vary.
@@ -205,10 +216,7 @@ public:
     void AddToVector(POLYUNSIGNED depth, POLYUNSIGNED length, PolyObject *pt);
 
 private:
-    struct _depthVector {
-        DepthVector **vector;
-        POLYUNSIGNED vectorSize;
-    } depthVectorArray[FIXEDLENGTHSIZE];
+	std::vector<DepthVector*> depthVectorArray[FIXEDLENGTHSIZE];
 
     POLYUNSIGNED maxVectorSize;
 };
@@ -216,11 +224,6 @@ private:
 ShareDataClass::ShareDataClass()
 {
     maxVectorSize = 0;
-    for (unsigned i = 0; i < FIXEDLENGTHSIZE; i++)
-    {
-        depthVectorArray[i].vector = 0;
-        depthVectorArray[i].vectorSize = 0;
-    }
 }
 
 ShareDataClass::~ShareDataClass()
@@ -229,13 +232,12 @@ ShareDataClass::~ShareDataClass()
     for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
        (*i)->shareBitmap.Destroy();
 
-    // Free the depth vectors and vector of vectors.
+    // Free the depth vectors.
     for (unsigned i = 0; i < FIXEDLENGTHSIZE; i++)
     {
-        for (unsigned j = 0; j < depthVectorArray[i].vectorSize; j++)
-            delete(depthVectorArray[i].vector[j]);
-        free(depthVectorArray[i].vector);
-    }
+		for (std::vector <DepthVector*>::iterator j = depthVectorArray[i].begin(); j < depthVectorArray[i].end(); j++)
+			delete(*j);
+	}
 }
 
 // Grow the appropriate depth vector if necessary and add the item to it.
@@ -243,34 +245,23 @@ void ShareDataClass::AddToVector(POLYUNSIGNED depth, POLYUNSIGNED length, PolyOb
 {
     // Select the appropriate vector.  Element zero is the variable length vector and is
     // also used for the, rare, zero length objects.
-    struct _depthVector *vectorToUse = &(depthVectorArray[length < FIXEDLENGTHSIZE ? length : 0]);
+    std::vector<DepthVector*> *vectorToUse = &(depthVectorArray[length < FIXEDLENGTHSIZE ? length : 0]);
 
     if (depth >= maxVectorSize) maxVectorSize = depth+1;
 
-    if (depth >= vectorToUse->vectorSize) {
-        POLYUNSIGNED newDepth = depth+1;
-        DepthVector **newVec = (DepthVector **)realloc(vectorToUse->vector, sizeof(DepthVector*)*newDepth);
-        if (newVec == 0) throw MemoryException();
-        vectorToUse->vector = newVec;
-        // Clear new entries first
-        for (POLYUNSIGNED d = vectorToUse->vectorSize; d < newDepth; d++)
-            vectorToUse->vector[d] = 0;
+	while (vectorToUse->size() <= depth)
+	{
+		try {
+			if (length != 0 && length < FIXEDLENGTHSIZE)
+				vectorToUse->push_back(new DepthVectorWithFixedLength(length));
+			else vectorToUse->push_back(new DepthVectorWithVariableLength);
+		}
+		catch (std::bad_alloc&) {
+			throw MemoryException();
+		}
+	}
 
-        for (POLYUNSIGNED d = vectorToUse->vectorSize; d < newDepth; d++)
-        {
-            try {
-                if (length != 0 && length < FIXEDLENGTHSIZE)
-                    vectorToUse->vector[d] = new DepthVectorWithFixedLength(length);
-                else vectorToUse->vector[d] = new DepthVectorWithVariableLength;
-            }
-            catch (std::bad_alloc &) {
-                throw MemoryException();
-            }
-        }
-        vectorToUse->vectorSize = newDepth;
-    }
-
-    vectorToUse->vector[depth]->AddToVector(length, pt);
+    (*vectorToUse)[depth]->AddToVector(length, pt);
 }
 
 // Add an object to a depth vector
@@ -405,7 +396,7 @@ POLYUNSIGNED DepthVector::MergeSameItems()
             // If we can't find a permanent space choose a space that isn't
             // an allocation space.  Otherwise we could break the invariant
             // that immutable areas never point into the allocation area.
-            MemSpace *space = gMem.SpaceForAddress(ptrVector[j]-1);
+            MemSpace *space = gMem.SpaceForAddress((PolyWord*)ptrVector[j]-1);
             if (bestSpace == 0)
             {
                 bestShare = ptrVector[j];
@@ -598,6 +589,7 @@ class ProcessFixupAddress: public ScanAddress
 {
 protected:
     virtual POLYUNSIGNED ScanAddressAt(PolyWord *pt);
+    virtual POLYUNSIGNED ScanCodeAddressAt(PolyObject **pt);
     virtual PolyObject *ScanObjectAddress(PolyObject *base)
         { return GetNewAddress(base).AsObjPtr(); }
     PolyWord GetNewAddress(PolyWord old);
@@ -609,6 +601,12 @@ POLYUNSIGNED ProcessFixupAddress::ScanAddressAt(PolyWord *pt)
     return 0;
 }
 
+// Don't have to do anything for code since it isn't moved.
+POLYUNSIGNED ProcessFixupAddress::ScanCodeAddressAt(PolyObject **pt)
+{
+    return 0;
+}
+
 // Returns the new address if the argument is the address of an object that
 // has moved, otherwise returns the original.
 PolyWord ProcessFixupAddress::GetNewAddress(PolyWord old)
@@ -621,6 +619,13 @@ PolyWord ProcessFixupAddress::GetNewAddress(PolyWord old)
     PolyObject *obj = old.AsObjPtr();
     POLYUNSIGNED L = obj->LengthWord();
 
+    if (obj->ContainsForwardingPtr()) // tombstone is a pointer to a shared object
+    {
+        PolyObject *newp = obj->GetForwardingPtr();
+//        ASSERT (newp->ContainsNormalLengthWord());
+        return newp;
+    }
+
     // Generally each address will point to an object processed at a lower depth.
     // The exception is if we have a cycle and have assigned the rest of the
     // structure to a higher depth.
@@ -629,20 +634,13 @@ PolyWord ProcessFixupAddress::GetNewAddress(PolyWord old)
     if (OBJ_IS_DEPTH(L))
         return old;
 
-    if (obj->ContainsForwardingPtr()) // tombstone is a pointer to a shared object
-    {
-        PolyObject *newp = obj->GetForwardingPtr();
-//        ASSERT (newp->ContainsNormalLengthWord());
-        return newp;
-    }
-
     ASSERT (obj->ContainsNormalLengthWord()); // object is not shared
     return old;
 }
 
 // This class is used to set up the depth vectors for sorting.  It subclasses ScanAddress
 // in order to be able to use that for code objects since they are complicated but it
-// handles most object types itself.  It scans them depth-first using an explicit stack.
+// handles all the other object types itself.  It scans them depth-first using an explicit stack.
 class ProcessAddToVector: public ScanAddress
 {
 public:
@@ -650,15 +648,19 @@ public:
 
     ~ProcessAddToVector();
 
+    // These are used when scanning code areas.  They return either
+    // a length or a possibly updated address.
     virtual POLYUNSIGNED ScanAddressAt(PolyWord *pt)
-        { (void)AddObjectsToDepthVectors(*pt); return 0; }
+        { (void)AddPolyWordToDepthVectors(*pt); return 0; }
     virtual PolyObject *ScanObjectAddress(PolyObject *base)
-        { (void)AddObjectsToDepthVectors(base); return base; }
+        { (void)AddObjectToDepthVector(base); return base; }
 
     void ProcessRoot(PolyObject *root);
 
 protected:
-    POLYUNSIGNED AddObjectsToDepthVectors(PolyWord old);
+    // Process an address and return the "depth".
+    POLYUNSIGNED AddPolyWordToDepthVectors(PolyWord old);
+    POLYUNSIGNED AddObjectToDepthVector(PolyObject *obj);
 
     void PushToStack(PolyObject *obj);
 
@@ -684,6 +686,14 @@ ProcessAddToVector::~ProcessAddToVector()
     free(addStack); // Now free the stack
 }
 
+POLYUNSIGNED ProcessAddToVector::AddPolyWordToDepthVectors(PolyWord old)
+{
+    // If this is a tagged integer or an IO pointer that's simply a constant.
+    if (old.IsTagged() || old == PolyWord::FromUnsigned(0))
+        return 0;
+    return AddObjectToDepthVector(old.AsObjPtr());
+}
+
 // Either adds an object to the stack or, if its depth is known, adds it
 // to the depth vector and returns the depth.
 // We use _OBJ_GC_MARK to detect when we have visited a cell but not yet
@@ -691,17 +701,12 @@ ProcessAddToVector::~ProcessAddToVector()
 // before we finish in the case that we run out of memory and throw an
 // exception.  PushToStack may throw the exception if the stack needs to
 // grow.
-POLYUNSIGNED ProcessAddToVector::AddObjectsToDepthVectors(PolyWord old)
+POLYUNSIGNED ProcessAddToVector::AddObjectToDepthVector(PolyObject *obj)
 {
-    // If this is a tagged integer or an IO pointer that's simply a constant.
-    if (old.IsTagged() || old == PolyWord::FromUnsigned(0))
-        return 0;
-
-    MemSpace *space = gMem.SpaceForAddress(old.AsStackAddr()-1);
+    MemSpace *space = gMem.SpaceForAddress(((PolyWord*)obj)-1);
     if (space == 0)
         return 0;
 
-    PolyObject *obj = old.AsObjPtr();
     POLYUNSIGNED L = obj->LengthWord();
 
     if (OBJ_IS_DEPTH(L)) // tombstone contains genuine depth or 0.
@@ -725,7 +730,7 @@ POLYUNSIGNED ProcessAddToVector::AddObjectsToDepthVectors(PolyWord old)
             if (containsAddress)
             {
                 // Add it to the vector so we will update any addresses it contains.
-                m_parent->AddToVector(0, L, old.AsObjPtr());
+                m_parent->AddToVector(0, L, obj);
                 // and follow any addresses to try to merge those.
                 PushToStack(obj);
                 obj->SetLengthWord(L | _OBJ_GC_MARK); // To prevent rescan
@@ -761,7 +766,7 @@ POLYUNSIGNED ProcessAddToVector::AddObjectsToDepthVectors(PolyWord old)
     if (obj->IsCodeObject())
     {
         // We want to update addresses in the code segment.
-        m_parent->AddToVector(0, L, old.AsObjPtr());
+        m_parent->AddToVector(0, L, obj);
         PushToStack(obj);
         obj->SetLengthWord(L | _OBJ_GC_MARK); // To prevent rescan
 
@@ -771,12 +776,12 @@ POLYUNSIGNED ProcessAddToVector::AddObjectsToDepthVectors(PolyWord old)
     // Byte objects always have depth 1 and can't contain addresses.
     if (obj->IsByteObject())
     {
-        m_parent->AddToVector (1, L, old.AsObjPtr());// add to vector at correct depth
+        m_parent->AddToVector (1, L, obj);// add to vector at correct depth
         obj->SetLengthWord(OBJ_SET_DEPTH(1));
         return 1;
     }
 
-    ASSERT(OBJ_IS_WORD_OBJECT(L)); // That leaves immutable data objects.
+    ASSERT(OBJ_IS_WORD_OBJECT(L) || OBJ_IS_CLOSURE_OBJECT(L)); // That leaves immutable data objects.
     PushToStack(obj);
     obj->SetLengthWord(L | _OBJ_GC_MARK); // To prevent rescan
 
@@ -815,7 +820,7 @@ void ProcessAddToVector::PushToStack(PolyObject *obj)
 void ProcessAddToVector::ProcessRoot(PolyObject *root)
 {
     // Mark the initial object
-    AddObjectsToDepthVectors(root);
+    AddObjectToDepthVector(root);
 
     // Process the stack until it's empty.
     while (asp != 0)
@@ -825,6 +830,7 @@ void ProcessAddToVector::ProcessRoot(PolyObject *root)
 
         if (obj->IsCodeObject())
         {
+            // Code cells are now only found in the code area.
             /* There's a problem sharing code objects if they have relative calls/jumps
                in them to other code.  The code of two functions may be identical (e.g.
                they both call functions 100 bytes ahead) and so they will appear the
@@ -838,69 +844,76 @@ void ProcessAddToVector::ProcessRoot(PolyObject *root)
                 obj->SetLengthWord(OBJ_SET_DEPTH(0)); // Now scanned
         }
 
-        // Immutable local objects.  These can be shared.  We need to compute the
-        // depth by computing the maximum of the depth of all the addresses in it.
-        else if ((obj->LengthWord() & _OBJ_GC_MARK) && ! obj->IsMutable())
-        {
-            POLYUNSIGNED depth = 0;
-            POLYUNSIGNED length = obj->Length();
-            PolyWord *pt = (PolyWord*)obj;
-            unsigned osp = asp;
-
-            while (length != 0 && osp == asp)
-            {
-                POLYUNSIGNED d = AddObjectsToDepthVectors(*pt);
-                if (d > depth) depth = d;
-                pt++;
-                length--;
-            }
-
-            if (osp == asp)
-            {
-                // We've finished it
-                asp--; // Pop this item.
-                depth++; // One more for this object
-                obj->SetLengthWord(obj->LengthWord() & (~_OBJ_GC_MARK));
-                m_parent->AddToVector(depth, obj->LengthWord() & (~_OBJ_GC_MARK), obj);
-                obj->SetLengthWord(OBJ_SET_DEPTH(depth));
-            }
-        }
-
-        // Mutable or non-local objects.  These have depth zero.  Local objects have
-        // _OBJ_GC_MARK in their header.  Immutable permanent objects cannot be
-        // modified so we don't set the depth.  Mutable objects are added to the
-        // depth vectors even though they aren't shared so that they will be
-        // updated if they point to immutables that have been shared.
         else
         {
             POLYUNSIGNED length = obj->Length();
             PolyWord *pt = (PolyWord*)obj;
             unsigned osp = asp;
 
-            while (length != 0)
+            if (obj->IsClosureObject())
             {
-                if (! (*pt).IsTagged())
-                {
-                    // If we've already pushed an address break now
-                    if (osp != asp) break;
-                    // Process the address and possibly push it
-                    AddObjectsToDepthVectors(*pt);
-                }
-                pt++;
-                length--;
+                // The first word of a closure is a code pointer.  We don't share code but
+                // we do want to share anything reachable from the constants.
+                AddObjectToDepthVector(*(PolyObject**)pt);
+                pt += sizeof(PolyObject*) / sizeof(PolyWord);
+                length -= sizeof(PolyObject*) / sizeof(PolyWord);
             }
 
-            if (length == 0)
+            if (((obj->LengthWord() & _OBJ_GC_MARK) && !obj->IsMutable()))
             {
-                // We've finished it
-                if (osp != asp)
+                // Immutable local objects.  These can be shared.  We need to compute the
+                // depth by computing the maximum of the depth of all the addresses in it.
+                POLYUNSIGNED depth = 0;
+                while (length != 0 && osp == asp)
                 {
-                    ASSERT(osp == asp-1);
-                    addStack[osp-1] = addStack[osp];
+                    POLYUNSIGNED d = AddPolyWordToDepthVectors(*pt);
+                    if (d > depth) depth = d;
+                    pt++;
+                    length--;
+                }
+
+                if (osp == asp)
+                {
+                    // We've finished it
+                    asp--; // Pop this item.
+                    depth++; // One more for this object
+                    obj->SetLengthWord(obj->LengthWord() & (~_OBJ_GC_MARK));
+                    m_parent->AddToVector(depth, obj->LengthWord() & (~_OBJ_GC_MARK), obj);
+                    obj->SetLengthWord(OBJ_SET_DEPTH(depth));
+                }
+            }
+            else
+            {
+                // Mutable or non-local objects.  These have depth zero.  Local objects have
+                // _OBJ_GC_MARK in their header.  Immutable permanent objects cannot be
+                // modified so we don't set the depth.  Mutable objects are added to the
+                // depth vectors even though they aren't shared so that they will be
+                // updated if they point to immutables that have been shared.
+                while (length != 0)
+                {
+                    if (!(*pt).IsTagged())
+                    {
+                        // If we've already pushed an address break now
+                        if (osp != asp) break;
+                        // Process the address and possibly push it
+                        AddPolyWordToDepthVectors(*pt);
+                    }
+                    pt++;
+                    length--;
+                }
+
+                if (length == 0)
+                {
+                    // We've finished it
+                    if (osp != asp)
+                    {
+                        ASSERT(osp == asp - 1);
+                        addStack[osp - 1] = addStack[osp];
+                    }
+                    asp--; // Pop this item.
+                    if (obj->LengthWord() & _OBJ_GC_MARK)
+                        obj->SetLengthWord(OBJ_SET_DEPTH(0));
                 }
-                asp--; // Pop this item.
-                if (obj->LengthWord() & _OBJ_GC_MARK)
-                    obj->SetLengthWord(OBJ_SET_DEPTH(0));
             }
         }
     }
@@ -944,9 +957,9 @@ bool ShareDataClass::RunShareData(PolyObject *root)
     {
         for (unsigned j = 0; j < FIXEDLENGTHSIZE; j++)
         {
-            if (depth < depthVectorArray[j].vectorSize)
+            if (depth < depthVectorArray[j].size())
             {
-                DepthVector *vec = depthVectorArray[j].vector[depth];
+                DepthVector *vec = depthVectorArray[j][depth];
                 // Set the length word and update all addresses.
                 vec->FixLengthAndAddresses(&fixup);
 
@@ -997,12 +1010,12 @@ bool ShareDataClass::RunShareData(PolyObject *root)
        (mutables and code). */
     for (unsigned j = 0; j < FIXEDLENGTHSIZE; j++)
     {
-        if (depthVectorArray[j].vectorSize > 0)
+        if (! depthVectorArray[j].empty())
         {
-            DepthVector *v = depthVectorArray[j].vector[0];
+            DepthVector *v = depthVectorArray[j][0];
             // Log this because it could be very large.
             if (debugOptions & DEBUG_SHARING)
-                Log("Sharing: Level %4" POLYUFMT ", size %3u, Objects %6" POLYUFMT "\n", 0, j, v->ItemCount());
+                Log("Sharing: Level %4" POLYUFMT ", size %3u, Objects %6" POLYUFMT "\n", 0ul, j, v->ItemCount());
             v->FixLengthAndAddresses(&fixup);
         }
     }
@@ -1015,9 +1028,9 @@ bool ShareDataClass::RunShareData(PolyObject *root)
     {
         for (unsigned j = 0; j < FIXEDLENGTHSIZE; j++)
         {
-            if (d < depthVectorArray[j].vectorSize)
+            if (d < depthVectorArray[j].size())
             {
-                DepthVector *v = depthVectorArray[j].vector[d];
+                DepthVector *v = depthVectorArray[j][d];
                 v->RestoreForwardingPointers();
             }
         }
