@@ -122,12 +122,13 @@
 #include "processes.h"
 #include "heapsizing.h"
 #include "rtsentry.h"
+#include "mpoly.h" // For polyStderr
 
 extern "C" {
     POLYEXTERNALSYMBOL POLYUNSIGNED PolyTimingGeneral(PolyObject *threadId, PolyWord code, PolyWord arg);
 }
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 /* Windows file times are 64-bit numbers representing times in
    tenths of a microsecond. */
 #define TICKS_PER_MICROSECOND 10
@@ -157,7 +158,7 @@ extern "C" {
     operating system provides it.
 */
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 static FILETIME startTime;
 #define StrToLL _strtoi64
 #else
@@ -176,14 +177,14 @@ static PLock timeLock("Timing");
 
 static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
-    unsigned c = get_C_unsigned(taskData, DEREFWORDHANDLE(code));
+    unsigned c = get_C_unsigned(taskData, code->Word());
     switch (c)
     {
     case 0: /* Get ticks per microsecond. */
         return Make_arbitrary_precision(taskData, TICKS_PER_MICROSECOND);
     case 1: /* Return time since the time base. */
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             FILETIME ft;
             GetSystemTimeAsFileTime(&ft);
             return Make_arb_from_Filetime(taskData, ft);
@@ -196,7 +197,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
         }
     case 2: /* Return the base year.  This is the year which corresponds to
                zero in the timing sequence. */
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
         return Make_arbitrary_precision(taskData, 1601);
 #else
         return Make_arbitrary_precision(taskData, 1970);
@@ -218,7 +219,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
 #if (defined(HAVE_GMTIME_R) || defined(HAVE_LOCALTIME_R))
             struct tm result;
 #endif
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             /* Although the offset is in seconds it is since 1601. */
             FILETIME ftSeconds; // Not really a file-time because it's a number of seconds.
             getFileTimeFromArb(taskData, args, &ftSeconds); /* May raise exception. */
@@ -227,7 +228,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
             liTime.LowPart = ftSeconds.dwLowDateTime;
             theTime = (long)(liTime.QuadPart - SECSSINCE1601);
 #else
-            theTime = get_C_long(taskData, DEREFWORDHANDLE(args)); /* May raise exception. */
+            theTime = get_C_long(taskData, DEREFWORD(args)); /* May raise exception. */
 #endif
 
             {
@@ -268,7 +269,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 5: /* Find out if Summer Time (daylight saving) was/will be in effect. */
         {
             time_t theTime;
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             FILETIME ftSeconds; // Not really a file-time because it's a number of seconds.
             getFileTimeFromArb(taskData, args, &ftSeconds); /* May raise exception. */
             ULARGE_INTEGER   liTime;
@@ -276,7 +277,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
             liTime.LowPart = ftSeconds.dwLowDateTime;
             theTime = (long)(liTime.QuadPart - SECSSINCE1601);
 #else
-            theTime = get_C_long(taskData, DEREFWORDHANDLE(args)); /* May raise exception. */
+            theTime = get_C_long(taskData, DEREFWORD(args)); /* May raise exception. */
 #endif
             int isDst = 0;
 #ifdef HAVE_LOCALTIME_R
@@ -313,7 +314,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
             time.tm_wday = get_C_int(taskData, DEREFHANDLE(args)->Get(7));
             time.tm_yday = get_C_int(taskData, DEREFHANDLE(args)->Get(8));
             time.tm_isdst = get_C_int(taskData, DEREFHANDLE(args)->Get(9));
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             _tzset(); /* Make sure we set the current locale. */
 #else
             setlocale(LC_TIME, "");
@@ -334,7 +335,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 7: /* Return User CPU time since the start. */
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             FILETIME ut, ct, et, kt;
             if (! GetProcessTimes(GetCurrentProcess(), &ct, &et, &kt, &ut))
                 raise_syscall(taskData, "GetProcessTimes failed", GetLastError());
@@ -350,7 +351,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 8: /* Return System CPU time since the start. */
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             FILETIME ct, et, kt, ut;
             if (! GetProcessTimes(GetCurrentProcess(), &ct, &et, &kt, &ut))
                 raise_syscall(taskData, "GetProcessTimes failed", GetLastError());
@@ -369,7 +370,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 10: /* Return real time since the start. */
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             FILETIME ft;
             GetSystemTimeAsFileTime(&ft);
             subFiletimes(&ft, &startTime);
@@ -386,7 +387,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
         /* These next two are used only in the Posix structure. */
     case 11: /* Return User CPU time used by child processes. */
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             return Make_arbitrary_precision(taskData, 0);
 #else
             struct rusage rusage;
@@ -399,7 +400,7 @@ static Handle timing_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 12: /* Return System CPU time used by child processes. */
         {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
             return Make_arbitrary_precision(taskData, 0);
 #else
             struct rusage rusage;
@@ -445,7 +446,7 @@ POLYUNSIGNED PolyTimingGeneral(PolyObject *threadId, PolyWord code, PolyWord arg
     else return result->Word().AsUnsigned();
 }
 
-#ifdef HAVE_WINDOWS_H
+#ifdef _WIN32
 void addFiletimes(FILETIME *result, const FILETIME *x)
 {
     ULARGE_INTEGER liA, liB;
@@ -556,7 +557,7 @@ static Timing timingModule;
 
 void Timing::Init(void)
 {
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
     // Record an initial time of day to use as the basis of real timing
     GetSystemTimeAsFileTime(&startTime);
 #else
@@ -572,23 +573,23 @@ time_t getBuildTime(void)
         char *endptr;
         long long epoch = StrToLL(source_date_epoch, &endptr, 10);
         if ((errno == ERANGE && (epoch == LLONG_MIN || epoch == LLONG_MAX)) || (errno != 0 && epoch == 0)) {
-            fprintf(stderr, "Environment variable $SOURCE_DATE_EPOCH: " XSTR(StrToLL) ": %s\n", strerror(errno));
+            fprintf(polyStderr, "Environment variable $SOURCE_DATE_EPOCH: " XSTR(StrToLL) ": %s\n", strerror(errno));
             goto err;
         }
         if (endptr == source_date_epoch) {
-            fprintf(stderr, "Environment variable $SOURCE_DATE_EPOCH: No digits were found: %s\n", endptr);
+            fprintf(polyStderr, "Environment variable $SOURCE_DATE_EPOCH: No digits were found: %s\n", endptr);
             goto err;
         }
         if (*endptr != '\0') {
-            fprintf(stderr, "Environment variable $SOURCE_DATE_EPOCH: Trailing garbage: %s\n", endptr);
+            fprintf(polyStderr, "Environment variable $SOURCE_DATE_EPOCH: Trailing garbage: %s\n", endptr);
             goto err;
         }
         if (epoch < (long long)std::numeric_limits<time_t>::min()) {
-            fprintf(stderr, "Environment variable $SOURCE_DATE_EPOCH: value must be greater than or equal to: %lld but was found to be: %lld\n", (long long)std::numeric_limits<time_t>::min(), epoch);
+            fprintf(polyStderr, "Environment variable $SOURCE_DATE_EPOCH: value must be greater than or equal to: %lld but was found to be: %lld\n", (long long)std::numeric_limits<time_t>::min(), epoch);
             goto err;
         }
         if (epoch > (long long)std::numeric_limits<time_t>::max()) {
-            fprintf(stderr, "Environment variable $SOURCE_DATE_EPOCH: value must be smaller than or equal to: %lld but was found to be: %lld\n", (long long)std::numeric_limits<time_t>::max(), epoch);
+            fprintf(polyStderr, "Environment variable $SOURCE_DATE_EPOCH: value must be smaller than or equal to: %lld but was found to be: %lld\n", (long long)std::numeric_limits<time_t>::max(), epoch);
             goto err;
         }
         return (time_t) epoch;
