@@ -1,7 +1,7 @@
 /*
     Title:  polyimport.c 
 
-    Copyright (c) 2006, 2015 David C.J. Matthews
+    Copyright (c) 2006, 2015, 2019 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -29,7 +29,7 @@
 
 #include "polyexports.h"
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#if (defined(_WIN32))
 int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
 {
     return PolyWinMain(hInstance, hPrevInstance, lpCmdLine, nCmdShow, 0);
